package platform.edit;

import java.awt.Component;
import java.awt.Cursor;
import java.awt.Toolkit;
import java.io.File;

import javax.swing.JOptionPane;

/**
 *	<< PLATFORM for DEGIGNING HIGH FUNCTIONAL MATERIALS >>
 *     GOURMET
 *
 * Copyright(c)1999-2005 The Japan Research Institute,Limited,
 * All rights reserved.
 *
 * @author Yoshiharu Inui
 * @version     2.00, 2001/05/26
 * @since       1.0
 */

/*
 *  Platform Application Information
 */
public class AppInfo {

	/**** History.
		Version 3.6.4 : Released 2004/2/1.
	 	Version 4.0.0 : Released 2006/8/8.
	 	Version 4.1.4 : Released 2011/8/8.
	 	Version 4.1.5 : Released 2013/4/1.
	 	Version 4.1.6 : Released 2013/6/1.
	 	Version 4.1.7 : Released 2013/10/1.
		Version 4.2.0 : Released 2015/3/1.
		Version 4.3.0 : Released 2016/6/8.
		Version 4.3.1 : Released 2016/8/8.
		Version 4.3.3 : Released 2017/11/8.
		Version 4.4.0 : Released 2020/08/08.
	 ********************************/

     // for Debug
    private final static boolean DEBUG = false;

    /** true for no python sub-interpreter. */
    public static final boolean PYTHON_NO_SUBINTERPRETER = true;
    /** true for confirmation of termination */
    public static final boolean CONFIRMATION_OF_TERMINATION = false;

    public static final String APPNAME = "GOURMET";

    public static final String SYSTEMNAME = "Graphical Open User interface foR Multiscale-analysis EnvironmenT";
    /*
    public static final String AUTHOR = "DOI PROJECT";
    public static final String COPYRIGHT = "Copyright(c) 1999-2017, OCTA User's Group,\n"
                                         + "All rights reserved.";
    */
    public static final String AUTHOR = "OCTA User's Group and Research Association of ADMAT";
    public static final String COPYRIGHT = "Copyright(c) OCTA licensing committee (1999-2020) and ADMAT (2020) All rights reserved.";
    //public static final String COPYRIGHT = "Copyright(c) 1999-2020, OCTA licensing committee All rights reserved.";
    //public static final String COPYRIGHT = "Copyright(c) 2020, Research Association of ADMAT All rights reserved.";
    public static final String COMMENT = "Released 2020/08/08";
    //public static final String COMMENT = "Developer's ver.";
    static final String RCSNAME = "@(#)$Name:  $";
    static final String RCSDATE = "@(#)$Date: 2020/08/08 $";
    static final int major = 4;
    static final int minor = 4;
    static final int mainte = 1;
	//static final String subversion = " "; // ""," alpha "," beta "," rc-1 "
	static final String subversion = " ";
    static final String versionNickName = "";  //"Shaker","beta8";

    public static boolean version410(){
    	return (major >= 4 && minor >= 1);
    }

    public static String getVersionString() {
        String version = Integer.toString(major) + '.' +
                         Integer.toString(minor) + '.' +
                         Integer.toString(mainte) + subversion;
        return version;
    }

    public static String getVersion() {
        String version = Integer.toString(major) + '.' +
                         Integer.toString(minor) + '.' +
                         Integer.toString(mainte) + subversion;

        if (versionNickName.length() > 0)
            return version + '(' + versionNickName + ')';
        else
            return version;
    }

    public static void underConstruction(Component comp) {
        JOptionPane.showMessageDialog(comp, "Sorry... under Construction",
                 "Application Infomation", JOptionPane.INFORMATION_MESSAGE);
    }

    public static void about(Component comp) {
    	String app_info = "";
    	String java_name = getJavaModule();
    	String arch_name = getArchitecture();
    	if (java_name.length() > 3) {
    		app_info = "Java : " + java_name;
    	}
    	if (arch_name.length() > 1) {
    		if (app_info.length() > 1) {
    			//XXX Mac
    			//app_info = app_info + ",  ";
    			app_info = app_info + "\n";
    		}
    		app_info = app_info + "Edition : " + arch_name;
    	}

        String msg = APPNAME + " " + getVersion() + "\n"
                    + SYSTEMNAME + "\n"
                    + AUTHOR + "\n"
                    + COPYRIGHT + "\n"
                    + COMMENT + "\n"
                    + app_info + "\n";
        // about dialog
        //JOptionPane.showMessageDialog(comp, msg,
        //         APPNAME + " - About", JOptionPane.INFORMATION_MESSAGE);
      	// about dialog(Picture)
    	platform.show.JarFile jf = new platform.show.JarFile();
        JOptionPane.showMessageDialog(comp, msg,
                 APPNAME + " - About", JOptionPane.INFORMATION_MESSAGE,
    	         new javax.swing.ImageIcon(jf.getImage(null,"/platform/manager/icon/doiproj.gif")));
    }

    public static String aboutString() {
        String msg = APPNAME + " " + getVersion() + "\n"
                    + SYSTEMNAME + "\n"
                    + AUTHOR + "\n"
                    + COPYRIGHT + "\n"
                    + COMMENT + "\n";
       return msg;
    }

	public static void imformation(Component comp, String msg){
		try {
			// Beep
			Toolkit.getDefaultToolkit().beep();
			if(comp!=null)
				comp.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));

			// Show imformation dialog
			JOptionPane.showMessageDialog(comp, msg,
						APPNAME + " - Imformation", JOptionPane.INFORMATION_MESSAGE);

		} catch (Exception ex) {
		    System.out.println("Error!! Cannot show Information Dialog: " + ex);
		}
	}

	static public String getJavaVersion(){
		String version = System.getProperty("java.version");
		if(version == null){
			version = "0.0";
		}
		return version;
	}

    static public String getJavaModule(){
    	String java_home = System.getProperty("java.home");
    	if(java_home != null){
    		//XXX Mac
    		//File path = new File(java_home);
    		//String name = path.getName();
    		//return name;
    		return java_home;
    	}
    	return "";
    }

    public static String getArchitecture(){
    	String pf_modules = System.getProperty("PF_MODULES");
    	if(pf_modules != null){
    		File path = new File(pf_modules);
    		String name = path.getName();
    		return name;
    	}
    	return "";
    }

	public static void warning(Component comp, String msg, Exception e){
		try {
	    	// Beep
	    	Toolkit.getDefaultToolkit().beep();
	    	if(comp!=null)
	    	    comp.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));

	    	// Show warning dialog
	    	String message;
	    	if(e==null && msg==null)
	    	    message = "System Warning Occurred!!";
            else if(e==null)
                message = msg;
            else if(msg==null)
                message = e.toString();
            else
                message = msg + "\n\n" + e;

	    	JOptionPane.showMessageDialog(comp, message,
	    	                APPNAME + " - Warning", JOptionPane.WARNING_MESSAGE);
		} catch (Exception ex) {
		    System.out.println("Error!! Cannot show Warning Dialog: " + ex);
		}
    }

	public static void error(Component comp, String msg, Exception e){
		try {
	    	// Beep
	    	Toolkit.getDefaultToolkit().beep();
	    	if(comp!=null)
	        	comp.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));

	    	// Show error dialog
	    	String message;
	    	if(e==null && msg==null)
	    	    message = "System Error Occurred!!";
            else if(e==null)
                message = msg;
            else if(msg==null)
                message = e.toString();
            else
                message = msg + "\n\n" + e;

            if(DEBUG){
                if(e!=null)
                    e.printStackTrace();
            }

	    	JOptionPane.showMessageDialog(comp, message,
	    	                APPNAME + " - Error", JOptionPane.ERROR_MESSAGE);
		} catch (Exception ex) {
		    System.out.println("Error!! Cannot show Error Dialog: " + ex);
		}
    }

    public static void errorExit(Component comp, String msg, Exception e){
        error(comp, msg, e);
        if(DEBUG==false) System.exit(0);   // close the VM
    }
}
