/*
 *  Material Modeling Platform Package
 *  Copyright(c) 2002, The Japan Research Institute, Ltd.
 *  All rights reserved.
 */
#include "BinaryDataCommon.h"
#include "BinaryFilesEndian.h"
#include "util_file_seek.h"
#include <stdio.h>
#include <string.h>
using namespace BinaryFiles;
using namespace std;

namespace binaryio {

char* IntegerToString( longint_t ivalue, char *s  )
{
	char *sst, *p, *temp;
	char buffer[64];
	sst = s;
	p = temp = buffer;
	if((ivalue) < 0){
		(ivalue) = -(ivalue);
		*s++ = '-';
	}
	do{
		*p++ = (char)((ivalue)%10 + (int)'0');
		(ivalue) /= 10;
	} while((ivalue));
	while(p != temp) *s++ = *--p;
	*s++ = 0x0;
	return sst;
}
void debugINT64( longint_t ivalue )
{
	char buffer[64];
	IntegerToString(ivalue,buffer);
	printf("%s\n",buffer);
}

// size of a string and its length data.
unsigned int lengthBinarySTRINGL8( const char* text )
{
	unsigned int len = 0;
	if(text) len = strlen(text);
	if(len > BINARY_READ_STRING_SIZE) len = BINARY_READ_STRING_SIZE;
	return len + 1;
}
unsigned int lengthBinarySTRINGL16( unsigned int len )
{
	return len + 2;
}
unsigned int lengthBinarySTRINGL16( const char* text )
{
	unsigned int len = 0;
	if(text) len = strlen(text);
	return lengthBinarySTRINGL16(len);
}
unsigned int lengthBinarySTRINGNL( unsigned int len )
{
	int loop = len / BINARY_READ_STRING_SIZE;
	// int modo = len % BINARY_READ_STRING_SIZE;
	return ( len + loop + 1 );
}
unsigned int lengthBinarySTRINGNL( const char* text )
{
	unsigned int len = 0;
	if(text) len = strlen(text);
	return lengthBinarySTRINGNL( len );
}

bool copyBinarySTRINGNL( FILE* destination, FILE* source, unsigned int* optCopiedSize )
{
	unsigned int allSize = 0;
	unsigned int writingSize = 0;
	unsigned char readingSize = 0;
	char buffer[BINARY_MAX_STRING_SIZE];
	while(feof(source) == 0){
		if(fread((void*)&readingSize,sizeof(unsigned char),1,source) < 1) return false;
		if(readingSize <= 0){
			if(fwrite((void*)&readingSize,1,1,destination) < 1) return false;
			break;
		}
		//if(readingSize > BINARY_READ_STRING_SIZE) return false; //always false.
		writingSize = fread((void*)buffer,1,readingSize,source);
		if(writingSize < readingSize) return false;
		allSize += writingSize;
		if(fwrite((void*)&readingSize,1,1,destination) < 1) return false;
		if(fwrite((void*)buffer,1,writingSize,destination) < writingSize) return false;
		if(readingSize < BINARY_READ_STRING_SIZE) break;
	}
	if(optCopiedSize) *optCopiedSize = allSize;
	return true;
}

int skipBinarySTRINGNL( FILE* binaryOpened )
{
	unsigned int allSize = 0;
	unsigned char readingSize = 0;
	while(feof(binaryOpened) == 0){
		if(fread((void*)&readingSize,sizeof(unsigned char),1,binaryOpened) < 1) return -1;
		if(readingSize <= 0) break;
		//if(readingSize > BINARY_READ_STRING_SIZE) return -1; //always false.
		if(fseek(binaryOpened,(long)readingSize,SEEK_CUR)) return -1;
		allSize += readingSize;
		if(readingSize < BINARY_READ_STRING_SIZE) break;
	}
	return allSize;
}

// Checker.
static unsigned char binary_checker[] = {0x0,0x0d,0x0a,0x0, 0xff,0xff,0xff,0xff};
void copyBinaryChecker( unsigned char* bits )
{
	if(!bits) return;
	for(int i=0;i<BINARY_DATA_CHECKER_SIZE;i++){
		bits[i] = binary_checker[i];
	}
}
unsigned char* getBinaryChecker()
{
	return binary_checker;
}
unsigned char checkBinaryChecker( const unsigned char* bits )
{
	if(!bits) return 0x0;
	for(int i=0;i<BINARY_DATA_CHECKER_SIZE;i++){
		if( bits[i] != binary_checker[i] ) return (BINARY_DATA_CHECKER_SIZE+i);
	}
	return 0x1;
}
bool checkBinaryBit( const unsigned char bit )
{
	switch(bit){
	case 0x1:
	case 0x2:
	case 0x3:
	case 0x4:
		return true;
	}
	return false;
}

// According to the endian, the bytes of value is swapped.
void BinaryInterface::eswap( short& value ){
	if(_needSwap) Endian::binarySwap(value);
}
void BinaryInterface::eswap( unsigned short& value )
{
	if(_needSwap) Endian::binarySwap(value);
}
void BinaryInterface::eswap( int& value )
{
	if(_needSwap) Endian::binarySwap(value);
}
void BinaryInterface::eswap( unsigned int& value )
{
	if(_needSwap) Endian::binarySwap(value);
}
void BinaryInterface::eswap( longint_t& value )
{
	if(_needSwap) Endian::binarySwap(value);
}
void BinaryInterface::eswap( ulongint_t& value )
{
	if(_needSwap) Endian::binarySwap(value);
}
void BinaryInterface::eswap( float& value )
{
	if(_needSwap) Endian::binarySwap(value);
}
void BinaryInterface::eswap( double& value )
{
	if(_needSwap) Endian::binarySwap(value);
}
void BinaryInterface::eneswap( double& value )
{
	if(_crypt) value = _crypt->encode( value );
	if(_needSwap) Endian::binarySwap(value);
}
void BinaryInterface::deeswap( double& value )
{
	if(_crypt) value = _crypt->decode( value );
	if(_needSwap) Endian::binarySwap(value);
}

// Character fwrite/fread.
size_t fwrite2( const char* buffer, size_t size, size_t count, FILE *stream, crypt_udf* crypt)
{
	if(crypt){
		char pp;
		size_t i,j,nn = 0;
		for(i=0;i<count;i++){
			for(j=0;j<size;j++){
				pp = crypt->encode(buffer[nn]);
				if(fwrite(&pp,1,1,stream) != 1){
					return i;
				}
				nn++;
			}
		}
		return count;
	}
	else{
		return fwrite(buffer,size,count,stream);
	}
}
size_t fread2( char* buffer, size_t size, size_t count, FILE *stream, crypt_udf* crypt)
{
	if(crypt){
		char pp;
		size_t i,j,nn = 0;
		for(i=0;i<count;i++){
			for(j=0;j<size;j++){
				if(fread(&pp,1,1,stream) != 1){
					return i;
				}
				buffer[nn] = crypt->decode(pp);
				nn++;
			}
		}
		return count;
	}
	else{
		return fread(buffer,size,count,stream);
	}
}

//// writer ////
// INTEGER.
bool BinaryInterface::writeUCHAR( unsigned char number, FILE* binaryOpened )
{
	if(fwrite((void*)&number,1,1,binaryOpened) < 1) return false;
	return true;
}
bool BinaryInterface::writeUCHARARRAY( unsigned int number, unsigned char* array, FILE* binaryOpened )
{
	if(fwrite((void*)array,1,number,binaryOpened) < number) return false;
	return true;
}
bool BinaryInterface::writeINT16( short number,  FILE* binaryOpened )
{
	this->eswap(number);
	if(fwrite((void*)&number,sizeof(short),1,binaryOpened) < 1) return false;
	return true;
}
bool BinaryInterface::writeUINT16( unsigned short number,  FILE* binaryOpened )
{
	this->eswap(number);
	if(fwrite((void*)&number,sizeof(unsigned short),1,binaryOpened) < 1) return false;
	return true;
}
bool BinaryInterface::writeINT32( int number,  FILE* binaryOpened )
{
	this->eswap(number);
	if(fwrite((void*)&number,sizeof(int),1,binaryOpened) < 1) return false;
	return true;
}
bool BinaryInterface::writeUINT32( unsigned int number,  FILE* binaryOpened )
{
	this->eswap(number);
	if(fwrite((void*)&number,sizeof(unsigned int),1,binaryOpened) < 1) return false;
	return true;
}
bool BinaryInterface::writeINT64( longint_t number,  FILE* binaryOpened )
{
	this->eswap(number);
	if(fwrite((void*)&number,sizeof(longint_t),1,binaryOpened) < 1) return false;
	return true;
}
bool BinaryInterface::writeUINT64( ulongint_t number,  FILE* binaryOpened )
{
	this->eswap(number);
	if(fwrite((void*)&number,sizeof(ulongint_t),1,binaryOpened) < 1) return false;
	return true;
}
// REAL NUMBER
bool BinaryInterface::writeREAL32( float number, FILE* binaryOpened )
{
	this->eswap(number);
	if(fwrite((void*)&number,sizeof(float),1,binaryOpened) < 1) return false;
	return true;
}
bool BinaryInterface::writeREAL64( double number, FILE* binaryOpened )
{
	this->eneswap(number);
	if(fwrite((void*)&number,sizeof(double),1,binaryOpened) < 1) return false;
	return true;
}
// STRING.
bool BinaryInterface::writeSTRINGL8( const char* text, FILE* binaryOpened )
{
	int len = 0;
	unsigned char csize = 0;
	if(text) len = strlen(text);
	if(len > BINARY_READ_STRING_SIZE) csize = BINARY_READ_STRING_SIZE;
	else     csize = (unsigned char)len;
	if(!writeUCHAR(csize,binaryOpened)) return false;
	if(len > 0){
		if(fwrite((void*)text,1,(int)csize,binaryOpened) < csize) return false;
	}
	return true;
}
bool BinaryInterface::writeSTRINGL16( const char* text, FILE* binaryOpened )
{
	unsigned short len = 0;
	if(text) len = (unsigned short)strlen(text);
	if(!writeUINT16(len,binaryOpened)) return false;
	if(len > 0){
		if(fwrite((void*)text,1,(int)len,binaryOpened) < len) return false;
	}
	return true;
}
bool BinaryInterface::writeSTRINGL32( const char* text, FILE* binaryOpened )
{
	unsigned int len = 0;
	if(text) len = (unsigned int)strlen(text);
	if(!writeUINT32(len,binaryOpened)) return false;
	if(len > 0){
		if(fwrite((void*)text,1,(int)len,binaryOpened) < len) return false;
	}
	return true;
}
bool BinaryInterface::writeSTRINGL32(const std::string& text, FILE* binaryOpened)
{
	unsigned int len = (unsigned int)text.length();
	if (!writeUINT32(len, binaryOpened)) return false;
	if (len > 0) {
		if (fwrite((void*)text.c_str(), 1, (int)len, binaryOpened) < len) return false;
	}
	return true;
}
// UNLIMITED STRING.
bool BinaryInterface::writeSTRINGNL(
	const char* text,
	FILE* binaryOpened
)
{
	unsigned int length = 0;
	if(text) length = strlen(text);
	return writeSTRINGNL(length,text,binaryOpened,_crypt);
}
bool BinaryInterface::writeSTRINGNL(
	const char* text,
	FILE* binaryOpened,
	crypt_udf* crypt
)
{
	unsigned int length = 0;
	if(text) length = strlen(text);
	return writeSTRINGNL(length,text,binaryOpened,crypt);
}
bool BinaryInterface::writeSTRINGNL(
	const unsigned int length,
	const char* text,
	FILE* binaryOpened,
	crypt_udf* crypt
)
{
	if(length <= 0 || text == 0){
		return writeUCHAR(0,binaryOpened);
	}
	int location = 0;
	int loop = length / BINARY_READ_STRING_SIZE;
	unsigned char csize = BINARY_READ_STRING_SIZE;
	unsigned char modo = length % BINARY_READ_STRING_SIZE;
	for(int i=0;i<loop;i++){
		if(!writeUCHAR(csize,binaryOpened)) return false;
		if(fwrite2(&text[location],1,BINARY_READ_STRING_SIZE,binaryOpened,crypt) < BINARY_READ_STRING_SIZE){
			return false;
		}
		location += BINARY_READ_STRING_SIZE;
	}
	if(!writeUCHAR(modo,binaryOpened)) return false;
	if(modo > 0){
		if(fwrite2(&text[location],1,(int)modo,binaryOpened,crypt) < (int)modo) return false;
	}
	return true;
}
bool BinaryInterface::writeSTRINGNL( const string& text, FILE* binaryOpened )
{
	return writeSTRINGNL(text.length(),text.c_str(),binaryOpened);
}

bool BinaryInterface::writeSTRING_NOLENGTH(const char* text, FILE* binaryOpened)
{
	unsigned int len = 0;
	if(text) len = (unsigned int)strlen(text);
	if(len > 0){
		if(fwrite((void*)text,1,(int)len,binaryOpened) < len) return false;
	}
	return true;
}

//// Reader ////
bool BinaryInterface::readUCHAR( unsigned char& number, FILE* binaryOpened )
{
	if(fread((void*)&number,sizeof(unsigned char),1,binaryOpened) < 1) return false;
#ifdef BINARY_DEBUG
	printf("READ UCHAR:[%d]\n",number);
#endif
	return true;
}
bool BinaryInterface::readUCHARARRAY( unsigned int number, unsigned char* array, FILE* binaryOpened )
{
	if(fread((void*)array,sizeof(unsigned char),number,binaryOpened) < number) return false;
#ifdef BINARY_DEBUG
	printf("READ UCHAR:[%d]\n",number);
#endif
	return true;
}
bool BinaryInterface::readINT16( short& number,  FILE* binaryOpened )
{
	if(fread((void*)&number,sizeof(short),1,binaryOpened) < 1) return false;
	this->eswap(number);
#ifdef BINARY_DEBUG
	printf("READ INT16:[%d]\n",number);
#endif
	return true;
}
bool BinaryInterface::readUINT16( unsigned short& number,  FILE* binaryOpened )
{
	if(fread((void*)&number,sizeof(unsigned short),1,binaryOpened) < 1) return false;
	this->eswap(number);
#ifdef BINARY_DEBUG
	printf("READ UINT16:[%d]\n",number);
#endif
	return true;
}
bool BinaryInterface::readINT32( int& number,  FILE* binaryOpened )
{
	if(fread((void*)&number,sizeof(int),1,binaryOpened) < 1) return false;
	this->eswap(number);
#ifdef BINARY_DEBUG
	printf("READ INT32:[%d]\n",number);
#endif
	return true;
}
bool BinaryInterface::readUINT32( unsigned int& number,  FILE* binaryOpened )
{
	if(fread((void*)&number,sizeof(unsigned int),1,binaryOpened) < 1) return false;
	this->eswap(number);
#ifdef BINARY_DEBUG
	printf("READ UINT32:[%d]\n",number);
#endif
	return true;
}
bool BinaryInterface::readINT64( longint_t& number,  FILE* binaryOpened )
{
	if(fread((void*)&number,sizeof(longint_t),1,binaryOpened) < 1) return false;
	this->eswap(number);
#ifdef BINARY_DEBUG
	printf("READ INT64:"); debugINT64(number);
#endif
	return true;
}
bool BinaryInterface::readUINT64( ulongint_t& number,  FILE* binaryOpened )
{
	if(fread((void*)&number,sizeof(ulongint_t),1,binaryOpened) < 1) return false;
	this->eswap(number);
#ifdef BINARY_DEBUG
	printf("READ UINT64:"); debugINT64(number);
#endif
	return true;
}
bool BinaryInterface::readREAL32( float& number, FILE* binaryOpened )
{
	if(fread((void*)&number,sizeof(float),1,binaryOpened) < 1) return false;
	this->eswap(number);
#ifdef BINARY_DEBUG
	printf("READ REAL32:[%f]\n",number);
#endif
	return true;
}
bool BinaryInterface::readREAL64( double& number, FILE* binaryOpened )
{
	if(fread((void*)&number,sizeof(double),1,binaryOpened) < 1) return false;
	this->deeswap(number);
#ifdef BINARY_DEBUG
	printf("READ REAL64:[%f]\n",number);
#endif
	return true;
}
// STRING.
bool BinaryInterface::readSTRINGL8( char** text, FILE* binaryOpened )
{
	unsigned int len = 0;
	unsigned char csize = 0;
	char buffer[BINARY_MAX_STRING_SIZE];
	if(!readUCHAR(csize,binaryOpened)) return false;
	if(csize > 0){
		if(csize >= BINARY_READ_STRING_SIZE){
			csize = BINARY_READ_STRING_SIZE - 1;
		}
		len = fread((void*)buffer,sizeof(char),csize,binaryOpened);
		if(len != csize) return false;
		if(*text) delete[] *text;
		*text = new char[len+1];
		strncpy(*text,buffer,len);
		(*text)[len] = 0x0;
	}
	return true;
}
bool BinaryInterface::readSTRINGL32( char** text, FILE* binaryOpened )
{
	unsigned int len = 0;
	char *buffer;
	if(!readUINT32(len,binaryOpened)) return false;
	if(*text) delete[] *text;
	*text = 0x0;

	if(len > 0){
		buffer = new char[len+1];
		unsigned int result = fread((void*)buffer,sizeof(char),len,binaryOpened);
		if(len != result) return false;
		buffer[len] = 0x0;
		*text = buffer;
	}
	return true;
}
bool BinaryInterface::readSTRINGL32( std::string& text, FILE* binaryOpened )
{
	char* chartext = 0;
	if(readSTRINGL32( &chartext, binaryOpened )){
		if(chartext){
			text = chartext;
			delete[] chartext;
		}
		return true;
	}
	return false;
}

// Unlimited.
bool BinaryInterface::readSTRINGNL( string& text, FILE* binaryOpened )
{
	unsigned int num;
	unsigned char nsize = 0;
	num = fread((void*)&nsize,sizeof(unsigned char),1,binaryOpened);
	if(num <= 0) return false;
	text = "";
	char buffer[BINARY_MAX_STRING_SIZE];
	while(nsize >= BINARY_READ_STRING_SIZE){
		num = fread((void*)buffer,sizeof(char),nsize,binaryOpened);
		buffer[nsize] = 0x0;
		text += buffer;
		num = fread((void*)&nsize,sizeof(unsigned char),1,binaryOpened);
		if(num <= 0) return false;
	}
	if(nsize > 0){
		num = fread((void*)buffer,sizeof(char),nsize,binaryOpened);
		buffer[nsize] = 0x0;
		text += buffer;
	}
#ifdef BINARY_DEBUG
	printf("READ STRINGNL:[%s]\n",text.c_str());
#endif
	return true;
}

void BinaryInterface::setupCrypt( int code )
{
	if(code > 0){
		if(_crypt){
			_crypt->reset( code );
		}
		else{
			_crypt = new crypt_udf( code );
		}
	}
}
crypt_udf* BinaryInterface::getCrypt()
{
	return _crypt;
}
bool BinaryInterface::isCrypt()
{
	return _crypt != 0;
}

//// Skip ////
bool BinaryInterface::skipUCHAR( FILE* binaryOpened )
{
	if(fseek(binaryOpened,(long)1,SEEK_CUR)) return false;
	return true;
}
bool BinaryInterface::skipINT16( FILE* binaryOpened )
{
	if(fseek(binaryOpened,(long)2,SEEK_CUR)) return false;
	return true;
}
bool BinaryInterface::skipUINT16( FILE* binaryOpened )
{
	if(fseek(binaryOpened,(long)2,SEEK_CUR)) return false;
	return true;
}
bool BinaryInterface::skipINT32( FILE* binaryOpened )
{
	if(fseek(binaryOpened,(long)4,SEEK_CUR)) return false;
	return true;
}
bool BinaryInterface::skipUINT32( FILE* binaryOpened )
{
	if(fseek(binaryOpened,(long)4,SEEK_CUR)) return false;
	return true;
}
bool BinaryInterface::skipINT64( FILE* binaryOpened )
{
	if(fseek(binaryOpened,(long)sizeof(longint_t),SEEK_CUR)) return false;
	return true;
}
bool BinaryInterface::skipUINT64( FILE* binaryOpened )
{
	if(fseek(binaryOpened,(long)sizeof(ulongint_t),SEEK_CUR)) return false;
	return true;
}
bool BinaryInterface::skipREAL32( FILE* binaryOpened )
{
	if(fseek(binaryOpened,(long)4,SEEK_CUR)) return false;
	return true;
}
bool BinaryInterface::skipREAL64( FILE* binaryOpened )
{
	if(fseek(binaryOpened,(long)8,SEEK_CUR)) return false;
	return true;
}
void BinaryInterface::skipBYTE( FILE* binaryOpened, int byte, size_t number )
{
	if(number < 100000){
		unsigned char buffer[10];
		for(size_t i=0;i<number;i++){
			if(fread(buffer,byte,number,binaryOpened) < 1) return;
		}
	} else {
		longint_t skip = byte * number;
		if(fseek(binaryOpened,skip,SEEK_CUR)) return;
	}
}

} // End of namespace.
