/*
 *  Material Modeling Platform Package
 *  Copyright(c) 2002, The Japan Research Institute, Ltd.
 *  All rights reserved.
 */
#include "BinaryDataRead.h"
//#include "BinaryFilesEndian.h"
#include "StructureCacheInterface.h"
#include "UtilHugeFile.h"
#include "CacheEnviron.h"
using namespace std;
using namespace CacheUtilCore;
//using namespace BinaryFiles;

namespace binaryio {

#define BINARY_READ_MSG_ERR "Unexpected binary read error"
#define BINARY_READ_MSG_KEY "Binary data may be collapsed or reading was refused."
#define BINARY_READ_MSG_DEF1 " ["
#define BINARY_READ_MSG_DEF2 "] undeclared"

bool BinaryDataRead::checkBits( const unsigned char* bits )
{
	if(!_binaryInfo) return false;
	return _binaryInfo->checkBits(bits);
}

// Read string value.
int BinaryDataRead::readBinarySTRINGL8( char** text, FILE* binaryOpened )
{
	int num;
	unsigned char nsize;
	num = fread((void*)&nsize,sizeof(unsigned char),1,binaryOpened);
	if(num <= 0) return -1;
	num = (int)nsize;
	num = fread((void*)_temporary,sizeof(char),num,binaryOpened);
	_temporary[nsize] = 0x0;
	*text = _temporary;
#ifdef BINARY_DEBUG
	printf("READ STRINGL8:[%s]\n",*text);
#endif
	return num;
}
int BinaryDataRead::readBinarySTRINGL16( char** text, FILE* binaryOpened )
{
	int num;
	unsigned short nsize;
	num = fread((void*)&nsize,sizeof(unsigned short),1,binaryOpened);
	if(num <= 0) return -1;
	_binaryIO.eswap(nsize);
	if(nsize > BINARY_READ_STRING_SIZE){
		int loop = nsize / BINARY_READ_STRING_SIZE;
		int last = nsize % BINARY_READ_STRING_SIZE;
		_buffer.reset();
		for(int i=0;i<loop;i++){
			num = fread((void*)_temporary,sizeof(char),BINARY_READ_STRING_SIZE,binaryOpened);
			_temporary[BINARY_READ_STRING_SIZE] = 0x0;
			_buffer.cat(_temporary);
		}
		num = fread((void*)_temporary,sizeof(char),last,binaryOpened);
		_temporary[last] = 0x0;
		_buffer.cat(_temporary);
		*text = _buffer.get();
#ifdef BINARY_DEBUG
	printf("READ STRINGL16(1):[%s]\n",*text);
#endif
		return _buffer.len();
	}
	else{
		num = fread((void*)_temporary,sizeof(char),nsize,binaryOpened);
		_temporary[nsize] = 0x0;
		*text = _temporary;
#ifdef BINARY_DEBUG
	printf("READ STRINGL16(2):[%s]\n",*text);
#endif
		return num;
	}
}
// unlimited string.
int BinaryDataRead::readBinarySTRINGNL( char** text, FILE* binaryOpened )
{
	unsigned int num;
	unsigned char nsize = 0;
	num = fread((void*)&nsize,sizeof(unsigned char),1,binaryOpened);
	if(num <= 0) return -1;
	crypt_udf* pcrypt = _binaryIO.getCrypt();
	_buffer.reset();
	while(nsize >= BINARY_READ_STRING_SIZE){
		num = fread2(_temporary,sizeof(char),nsize,binaryOpened,pcrypt);
		_temporary[nsize] = 0x0;
		_buffer.cat(_temporary);
		num = fread((void*)&nsize,sizeof(unsigned char),1,binaryOpened);
		if(num <= 0) return -1;
	}
	if(nsize > 0){
		num = fread2(_temporary,sizeof(char),nsize,binaryOpened,pcrypt);
		_temporary[nsize] = 0x0;
		_buffer.cat(_temporary);
	}
	*text = _buffer.get();
#ifdef BINARY_DEBUG
	printf("READ STRINGNL:[%s]\n",*text);
#endif
	return _buffer.len();
}

int BinaryDataRead::skipBinarySTRINGNL( FILE* binaryOpened )
{
	unsigned int num;
	unsigned int result = 0;
	unsigned char nsize = 0;
	num = fread((void*)&nsize,sizeof(unsigned char),1,binaryOpened);
	if(num <= 0) return -1;
	while(nsize >= BINARY_READ_STRING_SIZE){
		num = fread2(_temporary,sizeof(char),nsize,binaryOpened, 0);
		result += num;
		num = fread((void*)&nsize,sizeof(unsigned char),1,binaryOpened);
		if(num <= 0) return -1;
	}
	if(nsize > 0){
		num = fread2(_temporary,sizeof(char),nsize,binaryOpened, 0);
		result += num;
	}
	return result;
}

int BinaryDataRead::skipBinarySTRINGNL( unsigned int skip, FILE* binaryOpened )
{
	unsigned int result = 0;
	for(unsigned int i=0;i<skip;i++){
		result += skipBinarySTRINGNL( binaryOpened );
	}
	return result;
}

int BinaryDataRead::skipArray(unsigned int skip, unsigned int size, FILE* binaryOpened)
{
	if(skip == 0) return 0;
	return UtilHugeFile::seek(_binaryOpened, (ulongint_t)(skip * size), SEEK_CUR);
}

void BinaryDataRead::setDataPosition(unsigned int id, ulongint_t position)
{
	if(_data_position_list){
		DataPositionType pos(id);
		DataPositionType* found = _data_position_list->search(&pos);
		if(found){
			found->setPosition(position);
		} else {
			_data_position_list->put(new DataPositionType(id, position));
		}
	}
}

void BinaryDataRead::resetAllDataPosition()
{
	if (_data_position_list) {
		resetDataPositionList(_data_position_list);
	}
}

 /**
  * A constructor.
  */
BinaryDataRead::BinaryDataRead(
	FILE* binaryOpened,
	BinaryDataInfo* binaryInfo,
	BinaryNameManager* nameManager,
	StructureCacheComponent* baseComponent,
	SortedDataPositionList* data_position_list
):_user_data(0)
{
	_binaryOpened = binaryOpened;
	_binaryInfo = binaryInfo;
	_nameManager = nameManager;
	_baseComponent = baseComponent;
	_data_position_list = data_position_list;
	memset((void*)_temporary,0,BINARY_MAX_STRING_SIZE);
	if(binaryInfo){
		if(binaryInfo->needEndianSwap()) _binaryIO.swap();
		if(binaryInfo->getKeyCode() > 0){
			_binaryIO.setupCrypt( binaryInfo->getKeyCode() );
		}
	}
	switch_partial_read_array = GetCacheEnvironInstance()->isPartialReadArray();
	limit_size_of_read_array = GetCacheEnvironInstance()->getLimitSizeOfArray();
}

BinaryDataRead::~BinaryDataRead() {
	if (_user_data) {
		delete _user_data;
	}
}

void BinaryDataRead::setErrorMessage( int eid, const char* name )
{
	switch(eid){
	case 0:
		_bin_error_message = "Binary data is collapsed";
		break;
	case 10:
		_bin_error_message = BINARY_READ_MSG_KEY;
		break;
	case 20:
		_bin_error_message = BINARY_READ_MSG_DEF1;
		if(name) _bin_error_message += name;
		_bin_error_message += BINARY_READ_MSG_DEF2;
		break;
	default:
		_bin_error_message = BINARY_READ_MSG_ERR;
	}
}

void BinaryDataRead::setErrorMessage( int eid, const int number )
{
	char buffer[20];
	sprintf(buffer,"%d",number);
	setErrorMessage( eid, buffer );
}

//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
// Read binary file.
//////////////////////////////////////////////////////
static const char* BINARY_FILE_KEYWORD[] = {
	BINARY_DATA_START_KEY,
	BINARY_DATA_NAMEP_KEY,
	BINARY_DATA_STEP0_KEY,
	BINARY_DATA_STEPX_KEY,
	BINARY_DATA_BEGIN_KEY,
	BINARY_DATA_DIVID_KEY
};
static int stringCompare(int n, const char* s1, const char* s2){
	for(int i=0;i<n;i++){
		if(s1[i] != s2[i]) return 1;
	}
	return 0;
}
static int keyCompare(int start, const char* ss){
	for(int i=0;i<BINARY_DATA_KEY_NUMBER;i++){
		if(stringCompare(2,&BINARY_FILE_KEYWORD[i][start],ss) == 0) return i;
	}
	return -1;
}
const char* BinaryDataRead::getKey( int id )
{
	if(id < 0 || id >= BINARY_DATA_KEY_NUMBER) return 0;
	return BINARY_FILE_KEYWORD[id];
}

// [1] KEY ('DATA')
// --++DATA----
// xxoooooxxxxx : this reading point('o') can give us the searched keyword.
int BinaryDataRead::searchKey( FILE* binaryOpened, int* shiftFineTuning )
{
	int n,start,nkey,shift;
	char readKey[BINARY_DATA_KEY_LENGTH*2];
	readKey[BINARY_DATA_KEY_LENGTH] = 0x0;
	start = nkey = -1;
	int loop = 0;
	for( ; loop <= 1 ; loop++ ){
		if(fread((void*)readKey,1,2,binaryOpened) < 2){
			if(feof(binaryOpened)) return BINARY_DATA_EOF_KEYID;
			return -1;
		}
		for(n=0;n<=2;n++){
			if((nkey = keyCompare(n,readKey)) >= 0){
				start = n;
				goto KEY_FOUND;
			}
		}
	}
	if(start < 0) return -1;
KEY_FOUND:
	shift = ((-2) * loop) + start;
	if(shiftFineTuning){
		*shiftFineTuning = shift;
	}
	if(shift != 0) fprintf(stderr,"Warning: shifted binary key location %d.\n",shift);
	int nrev = start + 2;
	start = 2 - start;
	if(start == 0) return nkey;
	else if(start > 0){
		if(fread((void*)&(readKey[nrev]),1,start,binaryOpened) < start){
			return -1;
		}
#ifdef BINARY_DEBUG
	printf("BINARY KEY:[%s]\n",readKey);
#endif
		if(stringCompare(start,&(readKey[nrev]),&(BINARY_FILE_KEYWORD[nkey][nrev])) == 0){
			return nkey;
		}
	}
	return -1;
}

// [2] VERSION ('00.00.00')
bool BinaryDataRead::readVersion()
{
	if(!_binaryInfo) return false;
	char buff[16];
	if(fread((void*)buff,1,BINATY_FORMAT_VERSION_LEN,_binaryOpened) <= 0){
		setErrorMessage(10);
		return false;
	}
	buff[BINATY_FORMAT_VERSION_LEN] = 0x0;
#ifdef BINARY_DEBUG
	printf("VERSION:[%s]\n",buff);
#endif
	if(!_binaryInfo->setVersion(buff)){
		setErrorMessage(10);
		return false;
	}
	return true;
}

// [3] CHECKER (000D0A00FFFFFFFF)
bool BinaryDataRead::checkBinaryByte()
{
	unsigned int nsize = BINARY_DATA_CHECKER_SIZE;
	unsigned char bits[BINARY_DATA_CHECKER_SIZE];
	if(fread((void*)bits,1,nsize,_binaryOpened) < nsize) return false;
	if(!_binaryInfo) return false;
	if(!_binaryInfo->checkBits(bits)){
		setErrorMessage(0);
		return false;
	}
	return true;
}

// [4] ENDIAN ('L' or 'B')
bool BinaryDataRead::readEndian()
{
	if(!_binaryInfo) return false;
	char buff;
	if(fread((void*)&buff,1,1,_binaryOpened) <= 0) return false;
	if(!_binaryInfo->setEndian(buff)) return false;
	if(_binaryInfo->needEndianSwap()){
		_binaryIO.swap();
		// _binaryInfo->readInfo.needEndianSwap();
	}
#ifdef BINARY_DEBUG
	if(buff == BINATY_LITTLE_ENDIAN_KEY) printf("LITTLE ENDIAN\n");
	else if(buff == BINATY_BIG_ENDIAN_KEY)  printf("BIG ENDIAN\n");
#endif
	return true;
}
// [5] DATA TYPES INFORMATION (set (name, ID, size), Ex. ("byte", 1, 1)).
bool BinaryDataRead::readInfo( BinaryDataInfo* info, unsigned short* datasize )
{
	// BinaryDataInfo* info = _binaryInfo;
	// Byte size.
	if(datasize){
		*datasize = 0;
		if(!_binaryIO.readUINT16(*datasize,_binaryOpened)) return false;
		if(*datasize <= 0) return false;
	}
	// Number of type.
	unsigned char ucsize = 0;
	if(!_binaryIO.readUCHAR(ucsize,_binaryOpened)) return false;
	if(ucsize <= 0) return false;
	info->clearType();
	unsigned char temp;
	char* text;
	short id,size;
	for(unsigned char i=0; i<ucsize ; i++ ){
		// Name
		if(readBinarySTRINGL8(&text,_binaryOpened) <= 0) return false;
		// TypeID
		if(!_binaryIO.readUCHAR(temp,_binaryOpened)) return false;
		id = temp;
		// Size
		if(!_binaryIO.readUCHAR(temp,_binaryOpened)) return false;
		size = temp;
		//// set binary type ////
		info->setType(text,id,size);
	}
	// Optional flags.
	temp = 0;
	if(!_binaryIO.readUCHAR(temp,_binaryOpened)) return false;
	info->setNullFalg(temp);
	if(!_binaryIO.readUCHAR(temp,_binaryOpened)) return false;
	info->setArrayFlag(temp);
	if(!_binaryIO.readUCHAR(temp,_binaryOpened)) return false;
	info->setOptionFlag(0,temp);
	if(!_binaryIO.readUCHAR(temp,_binaryOpened)) return false;
	info->setOptionFlag(1,temp);
	// Option messages.
	if(info->getOptionFlag(0) > 0){
		info->readOptionMessage(0,_binaryOpened);
	}
	if(info->getOptionFlag(1) > 0){
		info->readOptionMessage(1,_binaryOpened);
	}
	return true;
}

// [6] NAME-ID
bool BinaryDataRead::readDataNameID( bool readKey )
{
	if(!_nameManager) return false;
	if(readKey){
		if(searchKey(_binaryOpened) != BINARY_DATA_NAMEP_KEYID) return false;
	}
	if(_binaryInfo->getKeyCode() > 0){
		_binaryIO.setupCrypt( _binaryInfo->getKeyCode() );
	}
	////////////////////////////////////////////////////
	BinaryAttributeMap* binName = _nameManager->getBinaryIdName();
	unsigned int number;
	int id,parentID;
	unsigned char typeID, partArrayDim;
	char* name;
	if(!_binaryIO.readUINT32(number,_binaryOpened)) return false;
	string nameWork;
	for(unsigned int i=0;i<number;i++){
		// Name
		if(readBinarySTRINGNL(&name,_binaryOpened) < 0) return false;
		if( name == 0 ) continue;
		if( *name == 0 ) continue;
		// ID
		if(!_binaryIO.readINT32(id,_binaryOpened)) return false;
		// parentID
		if(!_binaryIO.readINT32(parentID,_binaryOpened)) return false;
		// Part name -> Full name.
		if(parentID > 0){
			if(_nameManager->getBinaryName(parentID,nameWork)){
				nameWork += ".";
				nameWork += name;
			}
			else return false;
		}
		else{
			nameWork = name;
		}
		// typeID
		if(!_binaryIO.readUCHAR(typeID,_binaryOpened)) return false;
		// partArrayDimension
		if(!_binaryIO.readUCHAR(partArrayDim,_binaryOpened)) return false;
		// set
		_nameManager->setBinaryID(id,nameWork.c_str(),parentID,typeID,partArrayDim);
	}
	return true;
}

// [7] STEP DATA.
bool BinaryDataRead::readStep(
	int&             section,
	string&          label,
	longint_t&       datasize,
	int*             shiftFineTuning
)
{
	unsigned int header_size = 0;
	if(!readStepHeader(section,label,header_size,shiftFineTuning)) return false;
	if(header_size <= 0){
		// End of file.
		datasize = 0;
		return true;
	}
	if(!readStepData( section, datasize )) return false;
	datasize += header_size;
	return true;
}

bool BinaryDataRead::readSeries(
	int&             section,
	string&          label,
	longint_t&       datasize
)
{
	unsigned int header_size = 0;
	SeriesPositions series;
	if(!readStepHeader(section,label,header_size,0,&series)) return false;
	if(header_size <= 0){
		// End of file.
		datasize = 0;
		return true;
	}
	if(!skipStepData( datasize )) return false;
	datasize += header_size;
	return true;
}

bool BinaryDataRead::skipStep(
	int&             section,
	string&          label,
	longint_t&       datasize,
	int*             shiftFineTuning
)
{
	unsigned int header_size = 0;
	if(!readStepHeader(section,label,header_size,shiftFineTuning)) return false;
	if(header_size <= 0){
		// End of file.
		datasize = 0;
		return true;
	}
	if(!skipStepData( datasize )) return false;
	datasize += header_size;
	return true;
}

// SKIP
bool BinaryDataRead::skipStepData(longint_t& datasize)
{
	// KEY
	// if(searchKey(_binaryOpened) != BINARY_DATA_BEGIN_KEYID) return false;
	// SIZE
	ulongint_t udatasize = 0;
	if(!_binaryIO.readUINT64(udatasize,_binaryOpened)) return false;
	datasize = (longint_t)udatasize;
	// SEEK
	if(UtilHugeFile::seek(_binaryOpened,datasize,SEEK_CUR)){
		setErrorMessage(10);
		return false;
	}
	return true;
}

bool BinaryDataRead::skipSeriesInfo( unsigned int& datasize )
{
	datasize = 0;
	_binaryIO.readUINT32(datasize,_binaryOpened);
	if(datasize > 0){
		if(UtilHugeFile::seek(_binaryOpened,datasize,SEEK_CUR))
			return false;
	}
	return true;
}

// [7-1] STEP HEADER.
// (1) KEY ('COMN' or 'STEP')
// (2) STEP LABEL
// (3) Series data
bool BinaryDataRead::readStepHeader(
	int&              section,
	string&           label,
	unsigned int&     datasize,
	int*              shiftFineTuning,
	SeriesPositions*  series
)
{
	// KEY -> section.
	int num = searchKey(_binaryOpened,shiftFineTuning);
	switch(num){
	case BINARY_DATA_STEP0_KEYID: // COMN
		section = 0;
		break;
	case BINARY_DATA_STEPX_KEYID: // STEP
		section = 1;
		break;
	case BINARY_DATA_DIVID_KEYID: // DIVIDE
		section = BINARY_READ_DIVIDE;
		break;
	case BINARY_DATA_EOF_KEYID:
		section = -1;
		datasize = 0;
		return true;
	default:
		return false;
	}
	// label
	char* text;
	num = readBinarySTRINGNL(&text,_binaryOpened);
	if(num > 0) label = text;
	else        label = "";
	// series.
	if(series){
		if(!readSeriesInfo(datasize,*series)) return false;
	}
	else{
		if(!skipSeriesInfo(datasize)) return false;
	}
	// serise + label
	datasize += lengthBinarySTRINGNL((unsigned int)num);
	// DIVIDE DATA
	if(section == BINARY_READ_DIVIDE){
		datasize = 0;
		return readDivideHeader();
	}
	return true;
}

int BinaryDataRead::readBinDivideKey(
	int* shiftFineTuning
)
{
	// KEY -> section.
	int num = searchKey(_binaryOpened,shiftFineTuning);
	switch(num){
	case BINARY_DATA_STEP0_KEYID: // COMN
		return 0;
	case BINARY_DATA_STEPX_KEYID: // STEP
		return 1;
	case BINARY_DATA_DIVID_KEYID: // DIVIDE
		if( readDivideHeader() ){
			return -1;
		}
		break;
	case BINARY_DATA_EOF_KEYID:
		return -9;
	default:
		break;
	}
	return -99;
}

bool BinaryDataRead::readDivideHeader()
{
	// SIZE
	ulongint_t udatasize = 0;
	if(!_binaryIO.readUINT64(udatasize,_binaryOpened)) return false;
	// NUMBER
	unsigned int counter = 0;
	if(!_binaryIO.readUINT32(counter,_binaryOpened)) return false;
	for(unsigned int i=0;i<counter;i++){
		// ID
		int id;
		if(!_binaryIO.readINT32(id,_binaryOpened)) return false;
		// Number of data.
		unsigned int number_of_data = 0;
		if (!_binaryIO.readUINT32(number_of_data, _binaryOpened)) return false;
		if (number_of_data > 0) {
			// flag1 x 4
			unsigned char flag1[4];
			if (!_binaryIO.readUCHARARRAY((unsigned int)4, flag1, _binaryOpened)) return false;
			// flag2 x 4
			unsigned char flag2[4];
			if (!_binaryIO.readUCHARARRAY((unsigned int)4, flag2, _binaryOpened)) return false;
			// initial record
			ulongint_t initial_rec = 0;
			if (!_binaryIO.readUINT64(initial_rec, _binaryOpened)) return false;

			if (!_user_data) {
				_user_data = new UserBinaryUdfData();
			}
			_user_data->setDrawObjectPriority( (flag1[0] > 0) );
			_user_data->setInitialRecLast( (flag1[1] > 0) );
			_user_data->setInitialRec(initial_rec);
			
			// Number of string.
			unsigned int number_of_string = 0;
			if (!_binaryIO.readUINT32(number_of_string, _binaryOpened)) return false;
			for (unsigned int n = 0; n < number_of_string; n++) {
				string text;
				string buffer;
				string::size_type index_find;
				if (!_binaryIO.readSTRINGL32(buffer, _binaryOpened)) return false;
				// DIVID_DIRECTORY
				index_find = buffer.find(BINARY_DIVID_DIRECTORY);
				if (index_find != string::npos) {
					text = buffer.substr(BINARY_DIVID_DIRECTORY_LENGTH);
					_user_data->setUserHeaderDirectory(text.c_str());
					continue;
				}
				// DIVID_RECORD
				index_find = buffer.find(BINARY_DIVID_RECORD);
				if (index_find != string::npos) {
					text = buffer.substr(BINARY_DIVID_RECORD_LENGTH);
					_user_data->setUserHeaderRecord(text.c_str());
					continue;
				}
				// OBJEC
				index_find = buffer.find(BINARY_DIVID_OBJECT);
				if (index_find != string::npos) {
					text = buffer.substr(BINARY_DIVID_OBJECT_LENGTH);
					_user_data->setUserHeaderObjectFile(text.c_str());
					continue;
				}
			}
		}
	}
	return true;
}
// Sub reader.
bool BinaryDataRead::readSeriesInfo(unsigned int& datasize,SeriesPositions& series)
{
	series.clear();
	datasize = 0;
	unsigned int number = 0;
	_binaryIO.readUINT32(datasize,_binaryOpened);
	if(datasize <= 0) return true;
	_binaryIO.readUINT32(number,_binaryOpened);
	return true;
}
// [7-2] STEP CONTENTS
// (1) KEY ('BGND')
// (2) EACH SECTION DATA.
bool BinaryDataRead::readStepData(
	int section,
	longint_t& datasize
)
{
	// KEY
	// if(searchKey(_binaryOpened) != BINARY_DATA_BEGIN_KEYID) return false;
	// SIZE
	ulongint_t udatasize = 0;
	if(!_binaryIO.readUINT64(udatasize,_binaryOpened)) return false;
	datasize = (longint_t)udatasize;
	// SECTIONS
	return readSections(section);
}
bool BinaryDataRead::readSections( int section )
{
#ifdef BINARY_DEBUG
	printf("-------- READ STEP VALUES --------\n");
#endif
	// NUMBER
	unsigned int counter = 0;
	if(!_binaryIO.readUINT32(counter,_binaryOpened)) return false;
	//////////////////////////////////////////////////////////////
	const char* name = 0;
	unsigned char typeID = 0;
	unsigned char partArrayDim = 0;
	int id;
	int parentID = 0;
	StructureCacheComponent* component;
	// Reset positions.
	resetAllDataPosition();
	for(unsigned int i=0;i<counter;i++){
		// NAME
		// if(readBinarySTRINGNL(&name,_binaryOpened) <= 0) return false;
		// TypeID
		// if(!_binaryIO.readUCHAR(typeID,_binaryOpened)) return false;

		// ID
		if(!_binaryIO.readINT32(id,_binaryOpened)) return false;
		if(id > 0){
			// Set position data.
			setDataPosition(id, ftell(_binaryOpened));
			// ParentID
			if(!_nameManager->getBinaryAttribute(id,&name,&parentID,&typeID,&partArrayDim)){
				setErrorMessage( 10 ); // Collapsed.
				return false;
			}
		}
		// Component.
		component = _nameManager->findComponentFromFullName(name);
		if(!component){
			return false;
			//// No error occurs.
			//unsigned short fullArrayDim = _nameManager->getFullArrayDim(id);
			//if(!skipSection(typeID, partArrayDim, fullArrayDim)) return false;
			//continue;
			// An error occurs at the other location.
			if (!skipSection(typeID, partArrayDim, partArrayDim)) return false;
			continue;
			/*** As an error. ***
			string binname;
			if(_nameManager->getBinaryName(id,binname)){
				setErrorMessage( 20, binname.c_str() );
				return false;
			}
			setErrorMessage( 10 ); // Collapsed.
			return false;
			**********************/
		}
		// section.
		if(!readSection(typeID,section,component)) return false;
	}
#ifdef BINARY_DEBUG
	printf("-------- END  STEP VALUES --------\n");
#endif
	return true;
}
bool BinaryDataRead::readSection(
	unsigned char typeID,
	int section,
	StructureCacheComponent* component
)
{
	if(!component) return false;
	CacheSection* csection = component->section(section);
	if(!csection) return false;
	///////////////////////////////////////////////////////
	BaseDataType baseType = component->attribute()->baseType();
	int dataSize = 0;
	if(typeID > 0){
		const char* typeName = _binaryInfo->getTypeName(typeID);
		if(!_binaryInfo->dataTypeNameComp(baseType,typeName)) return false;
		dataSize = _binaryInfo->getTypeByteFromID(typeID);
	}
	///////////////////////////////////////////////////////
	// Array dimension
	unsigned char adim = (unsigned char)component->attribute()->arrayDim();
	// if(!_binaryIO.readUCHAR(adim,_binaryOpened)) return false;
	// Data size
	unsigned int number = 0;
	if(adim > 0){
		// Array index data.
		NumberArraySize* nasize = csection->getArraySize();
		if(!nasize) return false;
		if(adim != nasize->partDimension()) return false;
		AsizeArrayVector* avector = nasize->getSizeVector();
		unsigned int asize = 0;
		for(unsigned short i=0;i<adim;i++){
			if(!_binaryIO.readUINT32(asize,_binaryOpened)) return false;
			avector->at(i).resize(asize);
			for(unsigned int j=0;j<asize;j++){
				if(!_binaryIO.readUINT32(number,_binaryOpened)) return false;
				avector->at(i).at(j) = number;
			}
		}
	}
	// Values.
	return readValue(number,typeID,baseType,csection);
}
bool BinaryDataRead::readValue(
	unsigned int number,
	unsigned char typeID,
	BaseDataType baseType,
	CacheSection* csection
)
{
	if(!entityType(baseType)) return true;
	if(csection->isArray()){
		unsigned int nsize = number;
		// Number of data.
		if(!_binaryIO.readUINT32(nsize,_binaryOpened)) return false;
		number = nsize;
		LimitArraySection* arsection = (LimitArraySection*) csection;
		nsize = arsection->getRealSizeOfRead(nsize, this->isPartialReadArray(), this->getLimitSizeOfArray());
		unsigned int num_skip = number - nsize;
		/* Special operation for multi-array is needed.
		if(csection->partArrayDimension() >= 2){
			nsize = 1;
			num_skip = number - 1;
			arsection->setSizeOfRead(1);
		}
		*/
		// STRING.
		if(stringType(baseType)){
			//- Read all string array elements.
			if (_binaryIO.isCrypt()) {
				nsize = number;
				num_skip = 0;
				arsection->setSizeOfRead(0);
			}
			//-----End-----
			char* value;
			for(unsigned int i=0;i<nsize;i++){
				if(readBinarySTRINGNL(&value,_binaryOpened) < 0) return false;
				if(!csection->setData(i,value,true)) return false;
			}
			skipBinarySTRINGNL( num_skip, _binaryOpened );
		}
		// INTEGER.
		else if(integerType(baseType)){
			unsigned char nflag;
			switch(_binaryInfo->getTypeByteFromID(typeID)){
			case 1:
				{
					BYTE_TYPE value;
					for(unsigned int i=0;i<nsize;i++){
						if(_binaryInfo->existNullFlag()) _binaryIO.readUCHAR(nflag,_binaryOpened);
						_binaryIO.readUCHAR(value,_binaryOpened);
						if(!csection->setData(i,value,true)) return false;
					}
					skipArray(num_skip, 1, _binaryOpened);
				}
				break;
			case 2:
				{
					SHORT_TYPE value;
					for(unsigned int i=0;i<nsize;i++){
						if(_binaryInfo->existNullFlag()) _binaryIO.readUCHAR(nflag,_binaryOpened);
						_binaryIO.readINT16(value,_binaryOpened);
						if(!csection->setData(i,value,true)) return false;
					}
					skipArray(num_skip, 2, _binaryOpened);
				}
				break;
			case 4:
				{
					INT_TYPE value;
					for(unsigned int i=0;i<nsize;i++){
						if(_binaryInfo->existNullFlag()) _binaryIO.readUCHAR(nflag,_binaryOpened);
						_binaryIO.readINT32(value,_binaryOpened);
						if(!csection->setData(i,value,true)) return false;
					}
					skipArray(num_skip, 4, _binaryOpened);
				}
				break;
			default:
				return false;
			}
		}
		// FLOAT.
		else if(realType(baseType)){
			unsigned char nflag;
			switch(_binaryInfo->getTypeByteFromID(typeID)){
			case 4:
				{
					float value;
					for(unsigned int i=0;i<nsize;i++){
						if(_binaryInfo->existNullFlag()) _binaryIO.readUCHAR(nflag,_binaryOpened);
						_binaryIO.readREAL32(value,_binaryOpened);
						if(!csection->setData(i,value,true)) return false;
					}
					skipArray(num_skip, 4, _binaryOpened);
				}
				break;
			case 8:
				{
					DOUBLE_TYPE value;
					for(unsigned int i=0;i<nsize;i++){
						if(_binaryInfo->existNullFlag()) _binaryIO.readUCHAR(nflag,_binaryOpened);
						_binaryIO.readREAL64(value,_binaryOpened);
						if(!csection->setData(i,value,true)) return false;
					}
					skipArray(num_skip, 8, _binaryOpened);
				}
				break;
			default:
				return false;
			}
		} // End of TYPE.
	} // End of FULL ARRAY.
	else{ // SCALAR.
		unsigned int nsize = 0;
		if(!_binaryIO.readUINT32(nsize,_binaryOpened)) return false;
		if(nsize <= 0){
			csection->setNull(0,true);
			return true;
		}
		// STRING.
		if(stringType(baseType)){
			char* value;
			if(readBinarySTRINGNL(&value,_binaryOpened) < 0) return false;
			if(!csection->setData(0,value,true)) return false;
		}
		// INTEGER.
		else if(integerType(baseType)){
			switch(_binaryInfo->getTypeByteFromID(typeID)){
			case 1:
				{
					unsigned char flag = 1;
					BYTE_TYPE value;
					if(_binaryInfo->existNullFlag()){
						_binaryIO.readUCHAR(flag,_binaryOpened);
						if(!flag) csection->setNull(0,true);
					}
					_binaryIO.readUCHAR(value,_binaryOpened);
					if(flag){
						if(!csection->setData(0,value,true)) return false;
					}
				}
				break;
			case 2:
				{
					unsigned char flag = 1;
					SHORT_TYPE value;
					if(_binaryInfo->existNullFlag()){
						_binaryIO.readUCHAR(flag,_binaryOpened);
						if(!flag) csection->setNull(0,true);
					}
					_binaryIO.readINT16(value,_binaryOpened);
					if(flag){
						if(!csection->setData(0,value,true)) return false;
					}
				}
				break;
			case 4:
				{
					unsigned char flag = 1;
					INT_TYPE value;
					if(_binaryInfo->existNullFlag()){
						_binaryIO.readUCHAR(flag,_binaryOpened);
						if(!flag) csection->setNull(0,true);
					}
					_binaryIO.readINT32(value,_binaryOpened);
					if(flag){
						if(!csection->setData(0,value,true)) return false;
					}
				}
				break;
			default:
				return false;
			}
		}
		// FLOAT.
		else if(realType(baseType)){
			switch(_binaryInfo->getTypeByteFromID(typeID)){
			case 4:
				{
					unsigned char flag = 1;
					float value;
					if(_binaryInfo->existNullFlag()){
						_binaryIO.readUCHAR(flag,_binaryOpened);
						if(!flag) csection->setNull(0,true);
					}
					_binaryIO.readREAL32(value,_binaryOpened);
					if(flag){
						if(!csection->setData(0,value,true)) return false;
					}
				}
				break;
			case 8:
				{
					unsigned char flag = 1;
					DOUBLE_TYPE value;
					if(_binaryInfo->existNullFlag()){
						_binaryIO.readUCHAR(flag,_binaryOpened);
						if(!flag) csection->setNull(0,true);
					}
					_binaryIO.readREAL64(value,_binaryOpened);
					if(flag){
						if(!csection->setData(0,value,true)) return false;
					}
				}
				break;
			default:
				return false;
			}
		}
	} // END of SCALAR.
	return true;
}

// Skip.
bool BinaryDataRead::skipSection(
	unsigned char typeID,
	unsigned char partArrayDim,
	unsigned short fullArrayDim
)
{
	unsigned char typeByte = 0;
	if(typeID > 0){
		typeByte = _binaryInfo->getTypeByteFromID(typeID);
	}
	///////////////////////////////////////////////////////
	// Data size
	unsigned int number = 0;
	if(partArrayDim > 0){
		// Array index data.
		unsigned int asize = 0;
		for(unsigned char i=0;i<partArrayDim;i++){
			if(!_binaryIO.readUINT32(asize,_binaryOpened)) return false;
			for(unsigned int j=0;j<asize;j++){
				if(!_binaryIO.readUINT32(number,_binaryOpened)) return false;
			}
		}
	}
	// Values.
	return skipValue(number,typeID,typeByte,fullArrayDim);
}
bool BinaryDataRead::skipValue(
	unsigned int number,
	unsigned char typeID,
	unsigned char typeByte,
	unsigned short fullArrayDim
)
{
	if(typeByte <= 0) return true;
	if(fullArrayDim > 0){
		// Number of data.
		if(!_binaryIO.readUINT32(number,_binaryOpened)) return false;
		// STRING.
		switch(_binaryInfo->aboutTypeFromID(typeID)){
		case BinaryDataInfo::aboutStringType:
			{
				for(unsigned int i=0;i<number;i++){
					if(skipBinarySTRINGNL(_binaryOpened) < 0) return false;
				}
			}
			break;
		case BinaryDataInfo::aboutIntegerType: // INTEGER.
			{
				unsigned char nflag;
				switch(_binaryInfo->getTypeByteFromID(typeID)){
				case 1:
					{
						BYTE_TYPE value;
						for(unsigned int i=0;i<number;i++){
							if(_binaryInfo->existNullFlag()) _binaryIO.readUCHAR(nflag,_binaryOpened);
							_binaryIO.readUCHAR(value,_binaryOpened);
						}
					}
					break;
				case 2:
					{
						SHORT_TYPE value;
						for(unsigned int i=0;i<number;i++){
							if(_binaryInfo->existNullFlag()) _binaryIO.readUCHAR(nflag,_binaryOpened);
							_binaryIO.readINT16(value,_binaryOpened);
						}
					}
					break;
				case 4:
					{
						INT_TYPE value;
						for(unsigned int i=0;i<number;i++){
							if(_binaryInfo->existNullFlag()) _binaryIO.readUCHAR(nflag,_binaryOpened);
							_binaryIO.readINT32(value,_binaryOpened);
						}
					}
					break;
				default:
					return false;
				}
			}
			break;
		case BinaryDataInfo::aboutFloatType: // FLOAT.
			{
				unsigned char nflag;
				switch(_binaryInfo->getTypeByteFromID(typeID)){
				case 4:
					{
						float value;
						for(unsigned int i=0;i<number;i++){
							if(_binaryInfo->existNullFlag()) _binaryIO.readUCHAR(nflag,_binaryOpened);
							_binaryIO.readREAL32(value,_binaryOpened);
						}
					}
					break;
				case 8:
					{
						DOUBLE_TYPE value;
						for(unsigned int i=0;i<number;i++){
							if(_binaryInfo->existNullFlag()) _binaryIO.skipUCHAR(_binaryOpened);
							_binaryIO.readREAL64(value,_binaryOpened);
						}
					}
					break;
				default:
					return false;
				}
			}
		} // End of TYPE.
	} // End of FULL ARRAY.
	else{ // SCALAR.
		unsigned char flag;
		if(!_binaryIO.readUINT32(number,_binaryOpened)) return false;
		if(number <= 0) return true;
		switch(_binaryInfo->aboutTypeFromID(typeID)){
		case BinaryDataInfo::aboutStringType: // STRING.
			{
				if(skipBinarySTRINGNL(_binaryOpened) < 0) return false;
			}
			break;
		case BinaryDataInfo::aboutIntegerType: // INTEGER.
			switch(_binaryInfo->getTypeByteFromID(typeID)){
			case 1:
				{
					BYTE_TYPE value;
					if(_binaryInfo->existNullFlag()){
						_binaryIO.readUCHAR(flag,_binaryOpened);
					}
					_binaryIO.readUCHAR(value,_binaryOpened);
				}
				break;
			case 2:
				{
					SHORT_TYPE value;
					if(_binaryInfo->existNullFlag()){
						_binaryIO.readUCHAR(flag,_binaryOpened);
					}
					_binaryIO.readINT16(value,_binaryOpened);
				}
				break;
			case 4:
				{
					INT_TYPE value;
					if(_binaryInfo->existNullFlag()){
						_binaryIO.readUCHAR(flag,_binaryOpened);
					}
					_binaryIO.readINT32(value,_binaryOpened);
				}
				break;
			default:
				return false;
			}
			break;
		case BinaryDataInfo::aboutFloatType: // FLOAT.
			switch(_binaryInfo->getTypeByteFromID(typeID)){
			case 4:
				{
					float value;
					if(_binaryInfo->existNullFlag()){
						_binaryIO.readUCHAR(flag,_binaryOpened);
					}
					_binaryIO.readREAL32(value,_binaryOpened);
				}
				break;
			case 8:
				{
					DOUBLE_TYPE value;
					if(_binaryInfo->existNullFlag()){
						_binaryIO.skipUCHAR(_binaryOpened);
					}
					//_binaryIO.skipREAL64(_binaryOpened);
					_binaryIO.readREAL64(value,_binaryOpened);
				}
				break;
			default:
				return false;
			}
		}
	} // END of SCALAR.
	return true;
}

} // END of namespace.
