#ifndef DATA_POSITION_TYPE_H
#define DATA_POSITION_TYPE_H
#include "UtilNumberType.h"
#include "util_sorted_vector.h"

/**
 * @class DataPositionType
 * @brief Data id and start position pair.
 */
class DataPositionType {
	unsigned int id;
	ulongint_t   position;
public:
	DataPositionType():id(0), position(0){}
	DataPositionType(unsigned int id_):id(id_), position(0){}
	DataPositionType(unsigned int id_, ulongint_t position_):id(id_), position(position_){}
	DataPositionType(const DataPositionType& e):id(e.id), position(e.position){}
	void setId(unsigned int id){
		this->id = id;
	}
	unsigned int getId() const {
		return id;
	}
	void setPosition(ulongint_t position){
		this->position = position;
	}
	ulongint_t getPosition() const {
		return position;
	}
};

/**
 * @class CompareIdNumber
 * @brief Comparator of DataPositionType by id.
 */
class CompareIdNumber : public util_sorted::CompareClass<DataPositionType> {
public:
	bool operator()(
		const DataPositionType* a,
		const DataPositionType* b
	)
	{
		return ( a->getId() < b->getId() );
	}
};

/**
 * @class EqualIdNumber
 * @brief Equal function of DataPositionType by id.
 */
class EqualIdNumber : public util_sorted::EqualClass<DataPositionType> {
	unsigned int id;
public:
	EqualIdNumber( unsigned int id ){
		this->id = id;
	}
	bool operator()( const EqualIdNumber* a ) {
		if(!a) return false;
		return (id == a->id);
	}
};

/**
 * @class SortedDataPositionList
 * @brief Sorted vector of data positions.
 */
typedef util_sorted::util_sorted_vector<DataPositionType,CompareIdNumber,EqualIdNumber> SortedDataPositionList;

/**
 * @brief Clear all components.
 * @param[in] (position_list) list is not deleted.
 */
void resetDataPositionList(SortedDataPositionList* position_list);

/**
 * @brief Clear all components.
 * @param[in] (position_list) list is not deleted.
 */
void clearDataPositionList(SortedDataPositionList* position_list);

/**
 * @brief Delete all components and list
 * @param[in] (position_list) deleted list
 */
void deleteDataPositionList(SortedDataPositionList* position_list);

/**
 * @brief Copy all components and list
 * @param[in] (position_list) list to be copied
 */
SortedDataPositionList* copyDataPositionList(SortedDataPositionList* position_list);

#endif
