
/*
 *  Material Modeling Platform Package
 *  Copyright(c) 2002, The Japan Research Institute, Ltd.
 *  All rights reserved.
 *  
 *  $Id: FileStreamManager.h,v 1.9 2007/06/04 11:17:18 nishimoto Exp $
 */

#ifndef FILESTREAMMANAGER_H
#define FILESTREAMMANAGER_H
#include "file_stream_record.h"
#include "FileStreamShelter.h"
#include "CacheManagerCommon.h"
#include "CacheRecordOperation.h"

#include <time.h>
#include <string>
#include <vector>
#include <map>
using namespace std;
using namespace fsshelter;

//struct DataLocation {
//	DataLocation();
//	DataLocation(short part, short loc, const char* file);
//	~DataLocation();
//	void setLoc( short part, short loc, const char* file );
//	void  getLoc( short* part, short* loc, char** file );
//	short	_part;
//	short	_location;
//	char*	_file;
//};

//struct DataLocationInfo {
//	DataLocationInfo();
//	~DataLocationInfo();
//	int insertLocInfo(
//		const char* name,
//		short part,
//		short loc,
//		const char* file
//	);
//	int getLocInfo(
//		const char* name,	// IN
//		short part,			// IN
//		short* loc,			// OUT
//		char** file			// OUT
//	);
//	void clearLocInfo();
//	multimap<string,DataLocation*> _dtloc;
//};

////////////////////////////////////////////////////////
////////////////////////////////////////////////////////
//////////////// FileOperatingData /////////////////////
struct FileOperatingData {
	string        name;      // DATA IDENTIFICATION.
	string        path;      // DATA FILE/PATH.
	int           faccess;   // 0:READ-ONLY,1:WRITE-ALLOW
	vector<string> includes;          // INCLUDING FILES.
	time_t         file_time;         // LAST MODIFIED.
	UINT           rec_size;          // Number of Records.
public:
	REC_LABELPOS_DATA   rec_labelpos; // All Record Labels & Position.
	REC_LINENUMBER_TYPE read_lineno;  // Line number of a file.
	REC_POSITION_TYPE   last_position;
};

/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////
/////////////////// CacheVirtualRecord //////////////////
/////////////////////////////////////////////////////////
/**
// Data for edited records and appended records.
struct TemporaryCacheRecord {
	TemporaryCacheRecord():availability(true),
		record_no(0),record_label(0),filepath(0){}
	~TemporaryCacheRecord()
	{
		if(record_label) delete[] record_label;
		if(filepath) delete[] filepath;
	}
	void clear();
	void setRecordNo( const UINT number );
	void setFile( const char* path );
	void setLabel( const char* label );
	UINT getRecordNo() const { return record_no; }
	const char* getLabel()
	{
		return record_label;
	}
	const char* getFile()
	{
		return filepath;
	}
	bool available() const {
		return availability;
	}
	void unavail(){ availability = false; }
	void avail(){ availability = true; }
	//bool deleteRecord(){ return deletedRedord; }
	//void deletingRecord(){ deletedRedord = true; }
private:
	bool availability;
	//bool deletedRedord;
	UINT record_no;
	char* record_label;
	char* filepath;
};

class TemporaryRecordData {
private:
	typedef map<unsigned int,TemporaryCacheRecord*> data_type;
	typedef map<unsigned int,TemporaryCacheRecord*>::iterator data_iterator;
	data_type _tmprec;
public:
	TemporaryRecordData();
	~TemporaryRecordData();
	/////////////////////////
	bool insert( 
		UINT record_no, 
		const char* record_label, 
		const char* filepath = 0x0
	);
	bool get( 
		UINT record_no, 
		TemporaryCacheRecord** tmprec,
		bool forceGet    = false,
		bool eraseOption = false // true:get & erase.
	);
	bool get( 
		const char* record_label, 
		TemporaryCacheRecord** tmprec,
		bool forceGet    = false,
		bool eraseOption = false // true:get & erase.
	);
	UINT getAll(
		vector<TemporaryCacheRecord*>& tmprec_array,
		bool forceGet    = false
	);
	UINT getSize();
	UINT getUnconditionalSize(){ return _tmprec.size(); }
	TemporaryCacheRecord* getCacheRecord( unsigned int index ){
		if(index >= _tmprec.size()) return 0;
		data_iterator pp = _tmprec.begin();
		for(unsigned int i=0;i<index;i++) pp++;
		if(pp != _tmprec.end()) return pp->second;
		return 0;
	}
	void clear();
	void sortByRecord(); // no use for map.
	void annul( UINT record )
	{
		TemporaryCacheRecord *tmpRec;
		data_iterator pp = _tmprec.begin();
		while( pp != _tmprec.end() ){
			tmpRec = pp->second;
			if(tmpRec->getRecordNo() >= record){
				tmpRec->unavail();
			}
			pp++;
		}
	}
};
**/

// The types of virtual records.
/**
enum E_CACHE_VIRTUAL_RECORD_ACCESS {
	eVIRTUAL_ACCESS_BOTH = 0,
	eVIRTUAL_ACCESS_EDIT = 1,
	eVIRTUAL_ACCESS_APPEND = 2,
	eVIRTUAL_ACCESS_DELETE = 10
};
class CacheVirtualRecord_ {
private:
	bool _set_temporary_directory;
	string _temporary_directory;
	TemporaryRecordData	_edit_record;
	TemporaryRecordData _apnd_record;
	TemporaryRecordData _delt_record;
public:
	CacheVirtualRecord_();
	virtual ~CacheVirtualRecord_();

	bool insertVRec(
		E_CACHE_VIRTUAL_RECORD_ACCESS raccss,
		UINT record_no, 
		const char* record_label, 
		const char* filepath = 0x0
	);
	bool getVRec(
		E_CACHE_VIRTUAL_RECORD_ACCESS raccss,
		UINT record_no, 
		TemporaryCacheRecord** tmprec, 
		bool forceGet    = false,
		bool eraseOption = false // true:get & delete.
	);
	bool getVRec(
		E_CACHE_VIRTUAL_RECORD_ACCESS raccss,
		const char* record_label, 
		TemporaryCacheRecord** tmprec, 
		bool forceGet    = false,
		bool eraseOption = false // true:get & delete.
	);
	UINT getAllVRec(
		E_CACHE_VIRTUAL_RECORD_ACCESS raccss,
		vector<TemporaryCacheRecord*>& tmprec_array
	);
	UINT getAllNum(
		E_CACHE_VIRTUAL_RECORD_ACCESS raccss  = eVIRTUAL_ACCESS_EDIT
	);
	UINT getUnconditionalAllSize();
	void clearVirtualRecord();
	bool eraseFile();
	void sortingRecord();
	// Temporary directory.
	bool directory(){ return _set_temporary_directory; }
	void registDirectory(){ _set_temporary_directory = true; }
	void clearDirectory(){
		_set_temporary_directory = false;
		_temporary_directory = "";
	}
	string &getDirectory(){
		return _temporary_directory;
	}
	void setDirectory( string& dirpath ){
		_temporary_directory = dirpath;
	}
	// Deleted record.
	bool deleteRecord(){ return (_delt_record.getSize() > 0); }
	bool deleteRecord( UINT record ){
		if(_delt_record.getSize() <= 0) return false;
		TemporaryCacheRecord* tmpRec = _delt_record.getCacheRecord(0);
		return ( record == tmpRec->getRecordNo() );
	}
	UINT getDeleteRecNumber(){
		if(_delt_record.getSize() <= 0) return 0;
		TemporaryCacheRecord* tmpRec = _delt_record.getCacheRecord(0);
		return tmpRec->getRecordNo();
	}
	TemporaryCacheRecord* getDeleteRecord(){
		if(_delt_record.getSize() <= 0) return 0;
		return _delt_record.getCacheRecord(0);
	}
};
**/

class UserBinaryUdfData {
public:
	UserBinaryUdfData() :user_initial_record(0), user_initial_rec_last(false), user_draw_object_priority(false){}
	UserBinaryUdfData(const UserBinaryUdfData& o) {
		user_initial_record = o.user_initial_record;
		user_initial_rec_last = o.user_initial_rec_last;
		user_draw_object_priority = o.user_draw_object_priority;
		user_header_directory = o.user_header_directory;
		user_header_object_directory = o.user_header_object_directory;
		user_header_record = o.user_header_record;
		user_header_object_file = o.user_header_object_file;
	}
	void setInitialRec(longint_t rec) {
		user_initial_record = rec;
	}
	longint_t getInitialRec() {
		return user_initial_record;
	}
	void setInitialRecLast(bool v) {
		user_initial_rec_last = v;
	}
	bool isInitialRecLast() {
		return user_initial_rec_last;
	}
	void setDrawObjectPriority(bool v) {
		user_draw_object_priority = v;
	}
	bool getDrawObjectPriority() {
		return user_draw_object_priority;
	}
	const string& getUserHeaderDirectory() {
		return user_header_directory;
	}
	const string& getUserHeaderObjectDirectory() {
		return user_header_object_directory;
	}
	const string& getUserHeaderRecord() {
		return user_header_record;
	}
	const string& getUserHeaderObjectFile() {
		return user_header_object_file;
	}
	void setUserHeaderDirectory(const char* v) {
		if (v) user_header_directory = v;
		else user_header_directory.clear();
	}
	void setUserHeaderObjectDirectory(const char* v) {
		if (v) user_header_object_directory = v;
		else user_header_object_directory.clear();
	}
	void setUserHeaderRecord(const char* v) {
		if (v) user_header_record = v;
		else user_header_record.clear();
	}
	void setUserHeaderObjectFile(const char* v) {
		if (v) user_header_object_file = v;
		else user_header_object_file.clear();
	}
	void clearFile() {
		user_header_directory.clear();
	}
private:
	longint_t user_initial_record;
	bool user_initial_rec_last;
	bool user_draw_object_priority;
	string user_header_directory;
	string user_header_object_directory;
	string user_header_record;
	string user_header_object_file;
};

/////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////
////////////////////    FileStreamManager    ////////////////////////
/////////////////////////////////////////////////////////////////////
class FileStreamManager {
public:
	FileStreamManager(bool temporary_file = false);
	FileStreamManager( const char* path );
	~FileStreamManager();

	int setFile( const char* path = 0, int timeopt = 0 );
	bool exfile( const char* path );
	char* getFile();
	// Return udf-file-path or virtual record file path.
	/**
	 * @brief Get file path to read or reread
	 * @param[in] (record) record number
	 * @return file path string or 0
	 */
	//bool getReadFile(unsigned int record, string& read_file);
	bool getReadFile(unsigned int record, string& read_file, bool get_virtual = true);
	// Include files.
	void clearIncludeFile();
	unsigned int sizeIncludeFile();
	int setIncludeFile( const char* inc_file );
	const char* getLastIncludeFile();
	int getIncludeFiles( vector<string>& inc_files );
	// Related the temporary file managing. //
	bool isTemporary(){ return _temporary_file; }
	void releaseTemporary();
	int deleteTemporary();
	string& SaveFileName();
	//////////////////////////////////////////
	
	int openFile(
		const char* path = 0x0,
		FileStreamShelterMode fmode = fssm_non
	);
	int openTempFile(  // Open a temporary (edited record) file.
		const char* path = 0x0,
		FileStreamShelterMode fmode = fssm_non
	);
	const char* getOpenedFile(); // openFile or openTempFile.
	bool isOpenedVirtual();
	void* getStream();
	void closeFile();
	int good();
	bool setFileTime();
	int timeCompare( const char* path = 0x0 );
	// UINT	setRec( const char* label ){ return ++_file.rec_size; }
	// Current record size.
	void setRecSize( UINT rsize );
	UINT getRecSize();
	UINT getLastRecordNumber();
	// Read record size.
	UINT getFileRecSize(){
		return _file.rec_labelpos.getRecPosSize();
	}
	/////////////////////////////////////////////////////////////
	FileStreamShelter* getShelter(){ return _fss; }
	// about Locking.
	bool lock();
	bool unlock();
	bool islock();
	int getAccess(){ return _file.faccess; }
	///////////////////////////////////////////
	//////////// Record Information.//////////
	//void setCurRec(UINT no){ _rec_current = no; }
	//UINT getCurRec(){ return _rec_current; }
	//void setReadRec( UINT no ){ _rec_lastread = no; }
	//int getReadRec(){ return _rec_lastread; }
	///////////////////////////////////////////
	// Is quick output possible?
	bool quickOutputAppendRecord();
	// Specification to store the record information.
	void useRecTemporary(int size);
	void uselessRecTemporary();
	// Set record information.
	bool appendRecInfo(
		UINT recno,
		const char* label,
		const REC_LINENUMBER_TYPE lastLine,
		const REC_POSITION_TYPE   lastPosition,
		REC_LINENUMBER_TYPE       appendLine, // Appended file line no.
		REC_POSITION_TYPE         appendSize, // Appended file size.
		const char*               record_file
	);
	bool changeRecInfo(
		unsigned int recNumber,
		bool changeLabel,
		bool changeLine,
		bool changePosition,
		const char*         label,
		REC_LINENUMBER_TYPE line,
		REC_POSITION_TYPE   position
	);
	bool setRecInfo(
		UINT recno,
		const char* label,
		UINT line,              // Start line no.
		REC_POSITION_TYPE pos,  // Start position.
		const char* record_file = NULL, // Divided record file.
		bool binary_record_file = false
	);
	bool getRecInfo(
		UINT recno,		// IN
		UINT* line,		// OUT
		REC_POSITION_TYPE* pos  // OUT
	);
	bool getRecInfo(
		UINT recno,				// IN
		const char** label,		// OUT ( not delete. )
		UINT* line = 0, 		// OUT
		REC_POSITION_TYPE* pos = 0,  // OUT
		string* record_file = NULL, // Divided record file.
		bool* binary_record_file = 0
	);
	bool getRecInfo(
		const char* label,	// IN
		UINT* recno,		// OUT
		UINT* line,			// OUT
		REC_POSITION_TYPE* pos = 0,  // OUT
		string* record_file = NULL, // Divided record file.
		bool* binary_record_file = 0
	);
	bool getRecInfoExtend(
		UINT recno,			// IN
		const char** label 	// OUT ( not delete. )
	);
	bool getRecLabelAll(
		vector<const char*>& label		// OUT
	);
	bool getRecLabelExtend(
		vector<const char*>& label		// OUT
	);

	/**
	 * @brief Get binary record start position
	 * @param (recno) record number
	 * @param (pos) record start position
	 * @return true for normal
	 */
	bool getBinaryRecordInfo(
		UINT recno,
		REC_POSITION_TYPE& pos,
		bool get_virtual_info = true
	);

	/**
	 * @brief Get data position list
	 * @param (recNumber) record number
	 * @return data position list or 0.
	 */
	SortedDataPositionList* getBinaryDataPosition(
		unsigned int recNumber
	);
	/**
	 * @brief Set data position list
	 * @param (recNumber) record number
	 * @param (datapos) data position list
	 */
	bool setBinaryDataPosition(
		unsigned int recNumber,
		SortedDataPositionList* datapos
	);

	// Divided record files.
	bool compareRecFile(
		unsigned int recno,
		const char* rec_file
	);
	bool setRecFile(
		unsigned int recno,
		const char* rec_file,
		bool binary_record_file
	);
	const char* getRecFile(
		unsigned int recno,
		bool* binary_record_file
	);
	
	bool getSaveDivideMode();
	void setSaveDivideMode(bool divide);
	//unsigned char getConvertDivideMode();
	/** 0:RESET, 1:SINGLE, 2:DIVIDE */
	void setConvertDivideMode(unsigned char divide);

	// Always use.
	bool isDividedRecord();
	void setDividedRecord( bool divide );
	bool existDividedRecord();

#if defined(USING_NONDIVIDED_RECORD)
	bool existNondividedRecord();
	void setNondividedRecord();
#endif
	// In saving a divided udf.
	CacheDividedRecordOperation* makeDividedRecord(
		bool overwrite,
		const char* outputfile
	);

	//bool deleteRecAfter( UINT record );
	bool editRecLabel( UINT record, const char* label );
	bool editRecLabelExtend( UINT record, const char* label );
	void setLastLineNo( REC_LINENUMBER_TYPE line ){ _file.read_lineno = line; }
	REC_LINENUMBER_TYPE getLastLineNo(){ return _file.read_lineno; }
	void setLastPosition( REC_POSITION_TYPE pos ){ _file.last_position = pos; }
	REC_POSITION_TYPE getLastPosition(){ return _file.last_position; }
	// Location.
	// At the end of first parsing.
	void callAtEndOfFirstParse();
	void movePosSizeToRecSize();
	// Clear myself.
	void clear( bool reload = false );
	// EVENT
	void event( CACHE_EVENT_DATA& evdata );

	// Insert/Append/Delete Records.
	CacheRecordOperation* getRecordOperation(){
		return _rec_operation;
	}

	// Binary Record.
	void changeToBinaryMode();
	bool isBinaryMode();
	void startBinary();
	void clearBinary();
	void setBinaryStartRecord( unsigned int startRecord );
	unsigned int getBinaryStartRecord();
	bool isBinaryRecord( unsigned int record );
	void beginBinary( REC_POSITION_TYPE binaryPosition );
	REC_POSITION_TYPE beginBinaryPosition();
	void setBinaryInitialRecPos( REC_POSITION_TYPE binaryPosition );
	REC_POSITION_TYPE getBinaryInitialRecPos();

	// Utility.
	bool makepath(const char* path, const char* fname, string& fullpath);
	bool splitpath(const char* path, string& pathpart, string& filepart);
	static int pathCompare(const char* path1,const char* path2);

	// filepath
	bool makeDataNamePath( const char* post_fix, string& outpath );

	void setOutputFile( const char* file );
	const char* getOutputFile();

	// Divided record information.
	//void setIsUserDirectory(bool user);
	//void setIsUserRecord(bool user);
	//void setIsUserObjectFile(bool user);
	//bool isUserDirectory();
	//bool isUserRecord();
	//bool isUserObjectFile();
	// binary divide udf.
	void setInitialRec(longint_t rec);
	longint_t getInitialRec();
	void setInitialRecLast(bool v);
	bool isInitialRecLast();
	void setDrawObjectPriority(bool v);
	bool getDrawObjectPriority();
	void setDivideDirectory(const string& directory);
	void getDivideDirectory(string& directory, bool raw = false);
	//const char* getHeaderDirectory();
	void setObjectDirectory(const char* directory);
	const char* getObjectDirectory();
	void setRecordFileName(const char* file_name);
	const char* getRecordFileName();
	//const char* getHeaderRecord();
	void setDrawObjectFile(const char* file_pattern);
	const char* getDrawObjectFile();
	//const char* getHeaderObjectFile();
	void getUserBinaryUdfData(UserBinaryUdfData& user_data);
	void setUserBinaryUdfData(UserBinaryUdfData* user_data);
	// Last read file.
	void set_last_opened_file(const char* path, bool is_virtual = false);

private:
	time_t get_time_status( const char* path );
	//UINT  _rec_current;
	//int   _read_lastlineno; // Line number of a file.
	void initializeOperation();
	void initOperation();
	void clearOperation(bool reload);
	void createRecordOperation(); // At the end of first pasing a udf-file.
	// write divide mode.
	bool _save_divided_record;
	unsigned char _convert_divided_mode;
	// Releted the temporary file managing.(All UDF file)
	bool _temporary_file;
	string _temporary_file_name;
	string _save_file_name_when_temporary;
	// Object below.
	FileOperatingData  _file;
	FileStreamShelter *_fss;
	//DataLocationInfo  *_data_loc_info;
	// Record
	CacheRecordOperation *_rec_operation;

	char* _output_only_file;
	string _last_opend_file;
	bool _last_opened_is_virtual;
	/* Divided record infomation:(default or setting) */
	longint_t _initial_rec;
	bool _initial_rec_last;
	bool _object_priority;
	//bool _is_user_directory;
	//bool _is_user_record_name;
	//bool _is_user_object_file;
	char* _record_directory;
	char* _object_directory;
	char* _record_file_name;
	char* _draw_object_file; /* file pattern */
};

#endif

