
/*
 *  Material Modeling Platform Package
 *  Copyright(c) 2002, The Japan Research Institute, Ltd.
 *  All rights reserved.
 *  
 *  $Id: CacheExportManager.cpp,v 1.17 2007/10/13 05:55:42 nishimoto Exp $
 */

#ifdef WIN32
#pragma warning (disable:4786)
#endif

#ifdef WIN32
# include <stdio.h>
# include <io.h>
#else
#include <unistd.h>
#endif

#include "FileStreamManager.h"
#include "CacheEventController.h"
#include "CacheExportManager.h"
#include "StructureCacheOperation.h"
#include "DataAttrCache.h"
#include "cache_util_function.h"
#include "StructureCacheManager.h"
#include "binaryOutput.h"
#include "textudfOutput.h"
#include "util_file_seek.h"
#include "BinaryFileLib.h"
#include <string>
#include <vector>
#include <algorithm>
using namespace std;

// class CacheExportRecord.
CacheExportRecord::CacheExportRecord()//CacheVirtualRecord* virtualRecord)
	:edit_idx(0),apnd_idx(0)
{
	//edit_num = virtualRecord->getAllVRec(eVIRTUAL_ACCESS_EDIT,edit_rec);
	//apnd_num = virtualRecord->getAllVRec(eVIRTUAL_ACCESS_APPEND,apnd_rec);
	//delet_rec = virtualRecord->getDeleteRecord();
}

TemporaryCacheRecord* 
CacheExportRecord::getPreviousEdited()
{
	TemporaryCacheRecord* ret_rec = 0x0;
	if(edit_idx > 0) edit_idx--;
	if(edit_idx < edit_num){
		ret_rec = edit_rec[edit_idx];
		edit_idx++;
	}
	return ret_rec;
}

TemporaryCacheRecord* 
CacheExportRecord::getNextEdited()
{
	TemporaryCacheRecord* ret_rec = 0x0;
	if(edit_idx < edit_num){
		ret_rec = edit_rec[edit_idx];
		edit_idx++;
	}
	return ret_rec;
}

TemporaryCacheRecord* 
CacheExportRecord::getAppend( int index )
{
	if(index < 0 || index >= (int)apnd_num) return 0;
	return apnd_rec[index];
}

TemporaryCacheRecord* 
CacheExportRecord::getNextAppend()
{
	TemporaryCacheRecord* ret_rec = 0x0;
	if(apnd_idx < apnd_num){
		ret_rec = edit_rec[apnd_idx];
		apnd_idx++;
	}
	return ret_rec;
}

TemporaryCacheRecord* 
CacheExportRecord::getDeleteRecord()
{
	return delet_rec;
}

// class CacheExportManager.
CacheExportManager::CacheExportManager(
	CacheEventController* controller,
	FileStreamManager*  fsmgr,
	DefineCache*        define,
	DataAttrCache*		attr,
	StructureCacheOperation* ope
)
{
	_controller = controller;
	_fsmgr = fsmgr;
	_define = define;
	_attr = attr;
	_operation = ope;
}

CacheExportManager::~CacheExportManager()
{
	for(unsigned int i=0;i<_searched.size();i++){
		if(_searched[i]) delete _searched[i];
	}
	// if(_fsmgr) delete _fsmgr;
}

//
// A temporary-file name is generated.
//
bool
CacheExportManager::getNewName(UINT rec,string& filename)
{
	if(!_fsmgr) return false;
	CacheRecordOperation* recop = _fsmgr->getRecordOperation();
	//CacheVirtualRecord_* vrec = _fsmgr->getVirtualRecord();
	//////////////////////////////////////////////////////
	string newfilename,tmpfilename;
	char bf[20];
#if 1
	if(recop->directory()){
		tmpfilename = recop->getDirectory();
		/////////////////////////////////////////
		if(fcache_directory( tmpfilename.c_str() ) < 2) return false;
		/////////////////////////////////////////
		sprintf(bf,"rec%d",rec);
		tmpfilename = tmpfilename + fcache_get_delimiter();
		newfilename = tmpfilename + bf;
		return fudf_create_extfile_name(newfilename,filename);
	}
	newfilename = _fsmgr->getFile();
	if(fudf_create_extdir_name(newfilename,tmpfilename)){
		// A temporary directory is registered.
		recop->registDirectory();
		recop->setDirectory(tmpfilename);
		/////////////////////////////////////////
		sprintf(bf,"rec%d",rec);
		tmpfilename = tmpfilename + fcache_get_delimiter();
		newfilename = tmpfilename + bf;
		return fudf_create_extfile_name(newfilename,filename);
	}
	else{
		// OLD Version operation.
		sprintf(bf,".%d",rec);
		newfilename = newfilename + bf;
		return fudf_create_extfile_name(newfilename,filename);
	}
#else
	newfilename = _fsmgr->getFile();
	sprintf(bf,".%d",rec);
	newfilename = newfilename + bf;
	return fudf_create_extfile_name(newfilename,filename);
#endif
	return false;
}

// This function generates the temporary file name of the record edited or added.
string& 
CacheExportManager::createOutputFileName( const char* curfilename )
{
	if(curfilename){
		_output_file_name = curfilename;
	}
	else{
		// Create a new temporary file name.
		const char* label = 0x0;
		unsigned int rec = _controller->getCurrentRec();
		// Here, it investigates whether
		// the file is already registered into temporary file.
		CacheRecordOperation* recop = _fsmgr->getRecordOperation();
		//-CacheVirtualRecord* vrec = _fsmgr->getVirtualRecord();
		//-TemporaryCacheRecord *tvrec = 0x0;
		//-bool rb = vrec->getVRec(eVIRTUAL_ACCESS_EDIT,(UINT)rec,&tvrec,false); // available = false:erased record.
		if(recop->isEditedRecord(rec)){
			char* file = recop->getEditedRecordFilePath( rec );
			if(file){
				_output_file_name = file;
			}
			else{ // A file name is generated and registered.
				if(!getNewName(rec,_output_file_name)){
					throw("No permission to write in or no temporary directory.\n");
				}
				recop->setEditedRecordFilePath(rec,_output_file_name.c_str());
			}
		}
		else{ // A file name is generated and registered.
			// In Append, the file name should be registered.
			_fsmgr->getRecInfoExtend(rec,&label);
			if(!getNewName((UINT)rec,_output_file_name)){
				throw("No permission to write in or no temporary directory.\n");
			}
			recop->setEditRecord(rec,label,_output_file_name.c_str());
		}
		//-else{
		//-	// Here, a file name is registered newly.
		//-	// rb = vrec->getVRec(eVIRTUAL_ACCESS_APPEND,(UINT)rec,&tvrec,true);
		//-	rb = vrec->getVRec(eVIRTUAL_ACCESS_APPEND,(UINT)rec,&tvrec,false);
		//-	if(rb){
		//-		if(tvrec->getFile()){
		//-			_output_file_name = tvrec->getFile();
		//-		}
		//-		else{ // A file name is generated and registered.
		//-			if(!getNewName((UINT)rec,_output_file_name)){
		//-				throw("No permission to write in or no temporary directory.\n");
		//-			}
		//-			tvrec->setFile(_output_file_name.c_str());
		//-		}
		//-	}
		//-	else{ // A file name is generated and registered.
		//-		// In Append, the file name should be registered.
		//-		_fsmgr->getRecInfoExtend((UINT)rec,&label);
		//-		if(!getNewName((UINT)rec,_output_file_name)){
		//-			throw("No permission to write in or no temporary directory.\n");
		//-		}
		//-		vrec->insertVRec(eVIRTUAL_ACCESS_EDIT,(UINT)rec,label,_output_file_name.c_str());
		//-	}
		//-}
	}
	return _output_file_name;
}

// The data of the record edited or added is written to a temporary file.
bool
CacheExportManager::outputCurrentRecMultiways(
	const char* path,
	short create_temporary_path, // 0:use path,1:create path.
	short writemode,             // 0:overwrite,1:append
	short filemode,              // binary/text:0:current,-1:binary,1:text.
	short divide_record_mode,    // 0:normal(no-divide),1:depend on state,2:1 + with record-line.
	short* useMode,              // (Optional) OUTPUT, -1:binary,1:text,2:xml.
	REC_POSITION_TYPE* beforePosition, // (Optional) OUTPUT
	REC_POSITION_TYPE* afterPosition,  // (Optional) OUTPUT
	REC_LINENUMBER_TYPE* outputLine,   // (Optional) OUTPUT (no use for binary output).
	string* divide_record_link_name    // (Optional) OUTPUT Link name of divided record.
)
{
	if(filemode == 0){
		if(_fsmgr->isBinaryRecord(_controller->getCurrentRec())){
			filemode = -1;
		}
		else{
			filemode = 1;
		}
	}
	if(filemode < 0){
		// Binary.
		if(useMode) *useMode = -1;
		// Check and reread.
		UINT cur_rec = _controller->getCurrentRec();
		int section = StructureCacheManager::recordToSection(cur_rec);
		_operation->rereadCheckAndAll(section, cur_rec, _fsmgr);
		// the 'create_temporary_path' value controlles
		// the write-out of a binary header.
		return outputCurrentBinaryRec(
			path, create_temporary_path, writemode,
			beforePosition, afterPosition );
	}
	else{
		// UDF text.
		if(useMode) *useMode = 1;
		return outputCurrentTextRec(
			path,
			create_temporary_path,
			writemode,
			divide_record_mode,
			beforePosition,
			afterPosition,
			outputLine,
			divide_record_link_name );
	}
}

////////////////////////////////////////////////////////////////////
// Binary output.
////////////////////////////////////////////////////////////////////
bool CacheExportManager::isBinaryRecord( unsigned int recordNumber )
{
	return _fsmgr->isBinaryRecord(recordNumber);
}
bool CacheExportManager::needBinaryHeader( unsigned int recordNumber )
{
	bool outputHeader = _fsmgr->isBinaryRecord(recordNumber);
	if(recordNumber <= 0){
		return outputHeader;
	}
	else{
		if(!outputHeader) return false;
		if (_fsmgr->isBinaryRecord(recordNumber - 1)) {
			return false;
		}
		/* Check again to need "begin{binary}" keyword. */
		if (outputHeader) {
			REC_POSITION_TYPE begin_binary_pos = _fsmgr->beginBinaryPosition();
			if (begin_binary_pos > 0) {
				/* "begin{binary}" keyword already exists. */
				outputHeader = false;
			}
		}
	}
	return outputHeader;
}

bool CacheExportManager::outputCurrentBinaryRec(
	const char* path,
	short create_temporary_path,
	short writemode,  // 0:overwrite,1:append
	REC_POSITION_TYPE* beforePosition,
	REC_POSITION_TYPE* afterPosition
)
{
	bool rb = false;
	bool  outHeaderAsRequired = false;
	// check/create file path.
	// Open the stream.
	unsigned int rec = _controller->getCurrentRec();
	unsigned int section = StructureCacheManager::recordToSection(rec);
	BinaryOutput binaryOut;
	bool appendMode = false;
	if(writemode) appendMode = true;

	if(create_temporary_path){
		createOutputFileName(path);
		if(!binaryOut.open(_output_file_name.c_str(),appendMode)) return false;
	}
	else if(_fsmgr->isDividedRecord()){
		string dir_path;
		string record_prefix;
		string outputfile;
		if(!makeDivideBinaryPaths(dir_path, record_prefix, _fsmgr, _operation)){
			return false;
		}
		buildDivideRecordFileName(rec, dir_path, record_prefix, outputfile);
		appendMode = false;
		if(!binaryOut.open(outputfile.c_str(),appendMode)) return false;
	}
	else{
		if(!path) return false;
		if(!binaryOut.open(path,appendMode)) return false;
		outHeaderAsRequired = true;
	}
	
	StructureCacheOperation* operation = _operation;
	const char* label = 0;
	if(rec > 0){
		// Get a record label.
		if(!_fsmgr->getRecInfoExtend(rec,&label)){ //,false)){
			return false;
		}
	}
	// Binary header.(optional)
	if(outHeaderAsRequired){
		// Write a binary header.
		bool outputHeader = true;
		if(rec > 0){
			if (_fsmgr->isBinaryRecord(rec - 1)) {
				outputHeader = false;
			}
		}
		if (outputHeader) { /* Check again to need "begin{binary}" keyword. */
			REC_POSITION_TYPE begin_binary_pos  = _fsmgr->beginBinaryPosition();
			if (begin_binary_pos > 0) {
				/* "begin{binary}" keyword already exists. */
				outputHeader = false;
			}
		}
		if(outputHeader){
			// begin{binary} ////////////////////////////////////
			if(!binaryOut.outBEGIN()) return false;
			UserBinaryUdfData user_data;
			_fsmgr->getUserBinaryUdfData(user_data);
			if(!binaryOut.outHEADER(operation, &user_data)){
				return false;
			}
		}
	}
	// Write a record.
	if(beforePosition) *beforePosition = binaryOut.tell();
	if(binaryOut.outRECORD(rec,label,operation)){
		if(afterPosition) *afterPosition = binaryOut.tell();
		return true;
	}
	return false;
}

bool writeBinaryEmptyRecord(
	unsigned int ii_rec, 
	CacheRecordOperation* recop, 
	FileStreamManager* fsmgr, 
	StructureCacheOperation* operation,
	FILE* binaryWriteOpened
)
{
	// Empty record output.
	BinaryOutput temporary;
	temporary.setOutput(binaryWriteOpened);
	if (ii_rec <= 0) { // Binary common record. ///////////////
		fsmgr->setBinaryInitialRecPos(temporary.tell());
	} ///////////////////////////////////////////////////
	unsigned int section = StructureCacheManager::recordToSection(ii_rec);
	char* label = recop->getRecordLabel(ii_rec);
	if (!temporary.outEMPTY(section, label, operation)) {
		return false;
	}
	temporary.resetOutput();
	return true;
}

bool writeBinaryEditRecord(
	unsigned int ii_rec,
	CacheRecordOperation* recop,
	FileStreamManager* fsmgr,
	StructureCacheOperation* operation,
	FILE* binaryWriteOpened,
	bool& useVirtualRecord
)
{
	char* record_filepath = recop->getEditedRecordFilePath(ii_rec);
	if (record_filepath) {
		BinaryOutput temporary;
		if (temporary.openRead(record_filepath)) {
			temporary.setOutput(binaryWriteOpened);
			if (ii_rec <= 0) { // Binary common record. ///////////////
				fsmgr->setBinaryInitialRecPos(temporary.tell());
			} ///////////////////////////////////////////////////
			if (temporary.extractCopy(-1)) {
				useVirtualRecord = true;
			}
			temporary.resetOutput();
		}
	}
	else { // Empty record output.
		if (!writeBinaryEmptyRecord(ii_rec, recop, fsmgr, operation, binaryWriteOpened)) {
			return false;
		}
		useVirtualRecord = true;
	}
	return true;
}

bool CacheExportManager::transBinaryFile(	// true:binary, false:text
	FILE*         binaryWriteOpened,
	unsigned int& edit_flag,
	int write_flag // 1:write until binary name map,
)
{
	if(!_fsmgr->isBinaryMode()) return false;
	char* sourceFile = _fsmgr->getFile();
	// IN: original file, OUT: target file.
	BinaryOutput standardBinIO;
	FILE* sourceFP = standardBinIO.openRead(sourceFile);
	if(!sourceFP) return false;
	REC_POSITION_TYPE position;
	// begin{binary} ////////////////////////////////////
	standardBinIO.setOutput(binaryWriteOpened);
	if(!standardBinIO.outBEGIN()){
		standardBinIO.resetOutput();
		return false;
	}
	standardBinIO.resetOutput();
	// Binary Header ///////////////////////////////////////////
	position = _fsmgr->beginBinaryPosition();
	if(position < 0){ // New binary header is created.
		standardBinIO.setOutput(binaryWriteOpened);
		_fsmgr->beginBinary(standardBinIO.tell());
		UserBinaryUdfData user_data;
		_fsmgr->getUserBinaryUdfData(user_data);
		if(!standardBinIO.outHEADER(_operation, &user_data)){
			standardBinIO.resetOutput();
			return false;
		}
		standardBinIO.resetOutput();
		//return false;
	}
	else{ // The binary header has already existed. 
		if(standardBinIO.seekRead(position)){
			// binary header.
			if(!_operation->binaryCopy(binaryWriteOpened, sourceFP)) return false;
			// COMPARING POSITIONS AT TOP OF "NPIF":'N'.
			if(edit_flag == 0){
				REC_POSITION_TYPE src_pos = ftell(sourceFP);
				REC_POSITION_TYPE out_pos = ftell(binaryWriteOpened);
				if(src_pos != out_pos){
					edit_flag = 1; // Need to reload.
					//printf("[[[SET EDIT_FLAG TO RELOAD UDF]]]\n");
				}
			}
			// binary name map.
			if(!_operation->binaryCopy(binaryWriteOpened, sourceFP)) return false;
		}
	}
	if(write_flag == 1){
		return true;
	}

	//--- CHANGED RECORD -----------------------------------------
	CacheRecordOperation* recop = _fsmgr->getRecordOperation();
	unsigned int start_change_record = recop->getStartChangeRecord();
	unsigned int start_append_record = recop->getStartNewRecord();
	//------------------------------------------------------------
	unsigned int startRecord = _fsmgr->getBinaryStartRecord();
	// Record ///////////////////////////////////////////
	bool useVirtualRecord = false;
	unsigned int ii;
	unsigned int loopRecordSize = recop->getRecordNumber();
	//-unsigned int loopRecordSize = _fsmgr->getFileRecSize();
	for(ii = startRecord; ii <= loopRecordSize ; ii++ ){
		if(ii >= start_append_record) break;
		if(!_fsmgr->isBinaryRecord(ii)) continue;
		useVirtualRecord = false;
		if(recop->isEditedRecord( ii )){
			if (!writeBinaryEditRecord(ii, recop, _fsmgr, _operation, binaryWriteOpened, useVirtualRecord)) {
				return false;
			}
		}
		if(!useVirtualRecord){
			if(recop->isOriginal( ii )){
				unsigned int original_record = recop->getOriginalRecord( ii );
				if(original_record <= 0){
					position = _fsmgr->getBinaryInitialRecPos();
				}
				else if(!_fsmgr->getRecInfo(original_record,0,0,&position)){
					return false;
				}
				if(position >= 0){
					// Seek the reading position.
					standardBinIO.seekRead(position);
					// Write position.
					position = BinaryOutput::tell(binaryWriteOpened);
					if(ii <= 0){ // Binary common record. ///////////////////////////////////
						if(!_operation->binaryCopy(binaryWriteOpened, sourceFP)) return false;
						_fsmgr->setBinaryInitialRecPos(position);
					} ///////////////////////////////////////////////////////////////////////
					else{
						if(!_operation->binaryCopy(binaryWriteOpened, sourceFP)) return false;
						// * For avoiding the future crash in insert records. Set "original_record".
						_fsmgr->changeRecInfo(original_record,false,false,true,0,0,position);
						if(ii != original_record){ // Need to reload.
							if(edit_flag <= 0) edit_flag = 1; // Reload file.
						}
					}
				}
			}
			else{
				if (!writeBinaryEditRecord(ii, recop, _fsmgr, _operation, binaryWriteOpened, useVirtualRecord)) {
					return false;
				}
			}
		}
	}
	// Append record. ///////////////////////////////////////
	if(loopRecordSize < start_append_record){
		return true;
	}
	UINT apnd_recno = 0;
	const char* apnd_label = 0x0;
	for(ii = start_append_record;ii<=loopRecordSize;ii++){
		if(recop->isEditedRecord( ii )){
			char* filepath = recop->getEditedRecordFilePath( ii );
			if(filepath){
				BinaryOutput temporary;
				if(temporary.openRead( filepath )){
					temporary.setOutput(binaryWriteOpened);
					position = temporary.tell();
					if(temporary.extractCopy(-1)){
						useVirtualRecord = true;
						apnd_recno = ii;
						apnd_label = recop->getRecordLabel( ii );
						// From temporary record file to real record.
						if(!_fsmgr->setRecInfo(apnd_recno,apnd_label,0,position)){ // Binary!
							return false;
						}
					}
					temporary.resetOutput();
				}
			}
			else{ // Empty record output.
				if (!writeBinaryEmptyRecord(ii, recop, _fsmgr, _operation, binaryWriteOpened)) {
					return false;
				}
			}
		}
	}
	return true;
}

bool 
CacheExportManager::copyOneBinaryFile(
	FILE* ofp,
	const char* sourceFilepath,
	REC_POSITION_TYPE* readbyte
)
{
	if(!ofp) return false;
	if(!sourceFilepath) return false;
	FILE *ifp;
#if defined(WIN32) || defined(WIN64) || defined(CYGWIN_WIN32PATH)
	ifp = fopen(sourceFilepath,"rb");
#else
	ifp = fopen(sourceFilepath,"r");
#endif
	if(!ifp) return false;
	bool rb = _operation->binaryCopySpec(ofp,ifp,-1,readbyte);
	fclose(ifp);
	return rb;
}

// Binary data are written to another file.
bool 
CacheExportManager::extractBinaryRecord(
	FILE* ofp,
	UINT  record
)
{
	if(!ofp) return false;
	bool rb;
	UINT bgnline;
	REC_POSITION_TYPE bgnpos;
	const char* filepath = (const char*)_fsmgr->getFile();
	rb = _fsmgr->getRecInfo(record,&bgnline,&bgnpos);
	if(!rb) return false;
	FILE *ifp = 0x0;
	int origin = SEEK_SET;
#if defined(WIN32) || defined(WIN64) || defined(CYGWIN_WIN32PATH)
	ifp = fopen(filepath,"rb");
#else
	ifp = fopen(filepath,"r");
#endif
	if(UtilHugeFile::seek( ifp,bgnpos,origin ) == 0){ // OK
		rb = _operation->binaryCopy(ofp,ifp);
		fclose(ifp);
		return rb;
	}
	return false;
}

// WRITE DEFINE ONLY.
// CAN NOT OVERWRITE.
bool
CacheExportManager::writeDefineFile(
	const char* outfile,
	bool withCommon
)
{
	bool rb = false;
	bool binMode = false;
	bool overwrite = false;
	E_CACHE_OUTPUT_FILE efile;
	const char* filename = 0x0;
	string outname;
	unsigned int edit_flag = 0;
	//-int apnd_no = 0;
	int attr_flg = 0;
	const char* current_outfile_path = 0x0;
	if(_fsmgr->isBinaryMode()) binMode = true;
	if(outfile){
		// Save to another file name.
		filename = outfile;

		efile = eCACHE_OUTPUT_FILE_OVERWRITE;
		efile = eCACHE_OUTPUT_FILE_OTHERFILE;
		rb = outputFilename(efile,filename,outname);
		if(rb){
			unsigned char write_mask = 1;
			if(withCommon) write_mask |= 2;
			current_outfile_path = outname.c_str();
			rb = fcmgr_output_trans_files(
					current_outfile_path,
					current_outfile_path,
					this,
					overwrite,
					write_mask,
					binMode,
					edit_flag,
					attr_flg);
			return rb;
		}
	}
	return rb;
}

// The operations of the virtual record files.
bool
CacheExportManager::transFile(
	const char* outfile,
	int& reload,
	bool overwrite,
	bool clear_virtual
)
{
	reload = 1;
	bool rb = false;
	bool binMode = false;
	E_CACHE_OUTPUT_FILE efile;
	const char* filename = 0x0;
	string outname;
	string backname;
	unsigned int edit_flag = 0;
	//-int apnd_no = 0;
	int attr_flg = 0;
	bool to_reload = false;
	const char* current_outfile_path = 0x0;
	const char* final_outfile_path = outfile;
	if(_fsmgr->isBinaryMode()){
		binMode = true;
		if(_fsmgr->getSaveDivideMode()){
			if (_fsmgr->isDividedRecord()) {
				if (!writeDivideBinaryRecordFile(final_outfile_path, to_reload)) {
					return false;
				}
			}
			else {
				// Single binary file.
				if (!splitDivideBinaryRecordFile(final_outfile_path)) {
					return false;
				}
				to_reload = true;
			}
		}
		else {
			if (_fsmgr->isDividedRecord()) {
				to_reload = true;
			}
		}
	}
	if(outfile){
		// Save to another file name.
		filename = outfile;
		efile = eCACHE_OUTPUT_FILE_OTHERFILE;
		rb = outputFilename(efile,filename,outname);
		if(rb){
			current_outfile_path = outname.c_str();
			rb = fcmgr_output_trans_files(
					current_outfile_path,
					final_outfile_path,
					this,
					overwrite,
					0, // write_mask
					binMode,
					edit_flag,
					attr_flg);
			if(rb && clear_virtual){
				// All temporary files are erased.(erased in clear().)
				CacheRecordOperation* recop = _fsmgr->getRecordOperation();
				recop->clearRecordOperation();
				//-vrec->clearVirtualRecord();
				if(edit_flag > 0 || attr_flg > 0 || to_reload){
					reload = 1;
				}
				else{
					reload = 0;
				}
			}
		}
	}
	else{
		///
		//*** OVERWRITE ONLY ( APPENDED 2 OR MORE RECORDS ) *** 
		//
		// If only Append, virtual records will be moved to real records.
		int file_mode = 0; // 0:new,!0:append
		filename = _fsmgr->getFile();
		final_outfile_path = filename;
		// Write permission is checked.
		if(fcache_file_access( filename ) < 6) return false;
		CacheRecordOperation* recop = _fsmgr->getRecordOperation();
		edit_flag = recop->sizeEditRecord();
		
		// It is checked whether Header has been edited.
		vector<DataAttrSet*> dataset;
		DataAttrCache* attr = this->getDataAttr();
		int attr_edit = attr->getEdited(dataset);
		if(attr_edit > 0) edit_flag = 1;
		/////////////////////////////////////////
		if(recop->isAppendOnly()){ // append record.
			reload = 0;    // Without reload.
			file_mode = 1; // append file.
			outname = filename;
			rb = true;
		}
		else{
			efile = eCACHE_OUTPUT_FILE_OVERWRITE;
			rb = outputFilename(efile,filename,outname);
			if(edit_flag <= 0){ //- && erase_no > 0){
				reload = 0; // Without reload.
			}
		}
		if(rb){
			if (to_reload) {
				reload = 1;
			}
			if(file_mode){ // Appended records only.
				current_outfile_path = outname.c_str();
				rb = fcmgr_output_append_files(current_outfile_path,this); //-,edit_no,apnd_no);
				if(!rb) return false;
			}
			else{ // Editted & appended records.
				current_outfile_path = outname.c_str();
				rb = fcmgr_output_trans_files(
						current_outfile_path,
						final_outfile_path,
						this,
						overwrite,
						0, // write_mask
						binMode,
						edit_flag,
						attr_flg);
				if(!rb) return false;
				if(edit_flag > 0 || attr_flg > 0){
					reload = 1;
				}
			}
			if(rb && clear_virtual){
				if(!file_mode){
					// Backup file.
					fbackup_name(filename,backname);
					if(false) rename(filename,backname.c_str()); // No use.
					else{ // No backup-file!
						if(fcache_file_exist(filename) > 0){
							fcache_file_delete( filename );
						}
					}
					rename(outname.c_str(),filename);
				}
				// All temporary files are erased.(erased in clear().)
				recop->clearRecordOperation();
				//-vrec->clearVirtualRecord();
			}
		}
	}
	return rb;
}

// TEXT FILE ONLY!
//bool 
//CacheExportManager::outputCurrentTextRec(
//	const char* path,
//	short create_temporary_path, // 0:use path,1:create path.
//	short writemode,             // 0:overwrite,1:append
//	REC_POSITION_TYPE* beforePosition,
//	REC_POSITION_TYPE* afterPosition,
//	REC_LINENUMBER_TYPE* outputLine
//)
//{
//	// Output a text file.
//	bool rb = false;
//	// check/create file path.
//	const char* filep = 0x0;
//	if(create_temporary_path){
//		createOutputFileName(path);
//		filep = _output_file_name.c_str();
//	}
//	else{
//		if(!path) return false;
//		filep = path;
//	}
//	// Open the stream.
//	ofstream ofs;
//	if(writemode){
//		ofs.open(filep,ios::app);
//		ofs.seekp(0,ios::end);
//	}
//	else{
//		ofs.open(filep);
//	}
//	// Write out.
//	if(ofs.good()){
//		UDF_WRITE_CLASS wrt = UDF_WRITE_CLASS(_fsmgr,_define,_operation);
//		wrt.setStream( &ofs );
//		UINT rec = _controller->getCurrentRec();
//		if(beforePosition) *beforePosition = ofs.tellp();
//		rb = wrt.outData( rec, 0, outputLine );
//		if(afterPosition) *afterPosition = ofs.tellp();
//		ofs.close();
//	}
//	return true;
//}

bool 
CacheExportManager::outputCurrentTextRec(
	const char* path,
	short create_temporary_path, // 0:use path,1:create path.
	short writemode,             // 0:overwrite,1:append
	short divide_record_mode,    // 0:normal(no-divide),1:depend on state,2:1 + with record-line.
	REC_POSITION_TYPE* beforePosition,
	REC_POSITION_TYPE* afterPosition,
	REC_LINENUMBER_TYPE* outputLine,
	string* divide_record_link_name
)
{
	// Output a text file.
	bool rb = false;
	// check/create file path.
	const char* filep = 0x0;
	if(create_temporary_path){
		createOutputFileName(path);
		filep = _output_file_name.c_str();
	}
	else{
		if(!path) return false;
		filep = path;
	}
	// Open the stream.
	TextUdfOutput udfOutput(filep,(writemode > 0));
	if(!udfOutput.goodOutStream()) return false;
	// Write out.
	UINT rec = _controller->getCurrentRec();
	if(beforePosition) *beforePosition = udfOutput.tell();
	if(udfOutput.outputTextRecord(
			rec,
			NULL, // Record label.
			_fsmgr,
			_define,
			_operation,
			NULL, // family_name.
			outputLine,
			divide_record_mode,
			divide_record_link_name )
	){
		if(afterPosition) *afterPosition = udfOutput.tell();
		return true;
	}
	return false;
}

bool
CacheExportManager::outputCurrentRecord(
	const char* filepath,
	short writemode,	// 0:overwrite,1:append
	short filemode      // 0:current,-1:binary,1:text
)
{
	if(!filepath) return false;
	const char* curpath = (const char*)_fsmgr->getFile();
	if(strcmp(curpath,filepath) == 0){
		return false;
	}
	short divide_record_mode = 0;
	return outputCurrentRecMultiways(filepath,0,writemode,filemode,divide_record_mode);
}

//
// The data of a record is written.
//
//bool
//CacheExportManager::outputOneRecord(
//	const char* filepath,
//	UINT  record,	
//	int writemode	// 0:overwrite,1:append
//)
//{
//	if(!filepath) return false;
//	bool rb = false;
//	const char* curpath = (const char*)_fsmgr->getFile();
//	if(strcmp(curpath,filepath) == 0){
//		// A file name is the same.
//		return false;
//	}
//	if(record > 0){
//		FILE* ofp = 0x0;
//#if defined(WIN32) || defined(WIN64) || defined(CYGWIN_WIN32PATH)
//		if(writemode){
//			ofp = fopen(filepath,"ab");
//		}
//		else{
//			ofp = fopen(filepath,"wb");
//		}
//#else
//		if(writemode){
//			ofp = fopen(filepath,"a");
//		}
//		else{
//			ofp = fopen(filepath,"w");
//		}
//#endif
//		CacheRecordOperation* recop = _fsmgr->getRecordOperation();
//		rb = recop->isEditedRecord(record);
//		//-TemporaryCacheRecord* tmprec;
//		//-CacheVirtualRecord* vrec = _fsmgr->getVirtualRecord();
//		//-rb = vrec->getVRec(eVIRTUAL_ACCESS_BOTH,record,&tmprec); // BOTH(EDIT/APPEND)
//		if(rb){
//			const char* recfile = recop->getEditedRecordFilePath( record );
//			//-const char* recfile = tmprec->getFile();
//			if(_fsmgr->exfile(recfile)){
//				if(isBinaryRecord(record)){
//					rb = copyOneBinaryFile(ofp,recfile);
//				}
//				else{
//					rb = copyOneTextFile(ofp,recfile);
//				}
//				fclose(ofp);
//				return rb;
//			}
//		}
//		// Both binary/text.
//		rb = extractOneRecord(ofp,record);
//		// CLOSE.
//		fclose(ofp);
//	}
//	else{
//		// Write common.(0:use path)
//		rb = outputCurrentRecMultiways(filepath,0,writemode);
//	}
//	return rb;
//}

//
// The data of a record is written.
//
// Return bytes or -1 in error.
static int outputReturn(
	FILE *ofp
)
{
#ifdef WIN32
		int n = fputc( 0x0d, ofp );
		if(n == EOF) return -1;
		n = fputc( 0x0a, ofp );
		if(n == EOF) return -1;
		return 2;
#else
		return fprintf(ofp,"\n");
#endif
}

bool
CacheExportManager::extractOneRecord(
	FILE* ofp,
	UINT  record
)
{
	if(isBinaryRecord(record)){
		return extractBinaryRecord(ofp,record);
	}
	return extractTextRecord(ofp,record);
}

bool
CacheExportManager::extractTextRecord(
	FILE* ofp,
	UINT  record
)
{
	bool rb;
	short end_is_eof = 0;
	UINT bgnline,endline;
	REC_POSITION_TYPE bgnpos,endpos;
	const char* filepath = (const char*)_fsmgr->getFile();
	rb = _fsmgr->getRecInfo(record,&bgnline,&bgnpos);
	if(!rb) return false;
	if(bgnpos <= 0){
		rb = fcmgr_search_begin_record(filepath,&bgnpos);
		if(!rb) return false;
		if(bgnpos > 0) bgnpos--;
	}
	rb = _fsmgr->getRecInfo(record+1,&endline,&endpos);
	if(!rb){
		end_is_eof = 1;
	}
	int origin;
	int ch;
	long i,modify = 0;
	FILE *tfp = 0x0;

#if defined(WIN32) || defined(WIN64) || defined(CYGWIN_WIN32PATH)
	tfp = fopen(filepath,"rb");
	origin = SEEK_SET;
#else
	tfp = fopen(filepath,"r");
	origin = SEEK_SET;
#endif

	if(fseek( tfp,bgnpos,origin ) == 0){ // OK
		ch = fgetc( tfp );
		modify++;
		while( (ch != 0x0a) && ( feof(tfp) == 0 ) ){
			ch = fgetc( tfp );
			modify++;
		}
	}
	else{ // FALSE!
		fclose(tfp);
		return false;
	}

	if(end_is_eof <= 0){
		REC_POSITION_TYPE byte = endpos - bgnpos - modify;
		for(i=0;(i < byte ) && ( feof( tfp ) == 0 ) ;i++){
			ch = fgetc( tfp );
			fputc( ch, ofp );
		}
		while( ((ch != 0x0a) && (ch != 0x0d)) && ( feof(tfp) == 0 ) ){
			ch = fgetc( tfp );
			fputc( ch, ofp );
		}
		fprintf(ofp,"\n");
	}
	else{
		size_t numcount = CACHE_OUTPUT_BUFFERLEN;
		size_t numread = 0;
		char buffer[CACHE_OUTPUT_BUFFERMAX];
		while(!feof(tfp)){
			numread = fread( buffer,sizeof(char),numcount,tfp);
			fwrite( buffer, sizeof(char), numread, ofp );
		}

		// Newline.
		outputReturn( ofp );
	}
	fclose(tfp);
	return true;
}

//
// A file is written to another file
// with inspecting the size and the line number of the file.
bool
CacheExportManager::copyOneTextFile(
	FILE* ofp,   // Opened "w+"/"wb+", if check_last_newline == true.
	const char* sourceFilepath,
	REC_POSITION_TYPE* readbyte,
	REC_LINENUMBER_TYPE* readline,
	bool check_last_newline
)
{
	FILE *tfp;
	char* sp;
	char buffer[CACHE_OUTPUT_BUFFERMAX];
	if(sourceFilepath == 0x0) return false;
	const char* path = sourceFilepath;
	//------------------------------------
	int last_written = 0;
	if(check_last_newline){
		if(ftell( ofp ) > 2){
			long offset = -3;
			if(fseek( ofp, offset, SEEK_END ) == 0){
				last_written = getc( ofp );
				if(last_written == 0x0a || last_written == 0x0d){
					last_written = getc( ofp );
					if(last_written == 0x0a || last_written == 0x0d){
						last_written = getc( ofp );
					}
				}
			}
			fseek( ofp, 0L, SEEK_END );
		}
	}
	//------------------------------------
#if defined(WIN32) || defined(WIN64) || defined(CYGWIN_WIN32PATH)
	tfp = fopen(path,"rb");
#else
	tfp = fopen(path,"r");
#endif
	if(!tfp) return false;
	size_t numcount = CACHE_OUTPUT_BUFFERLEN;
	size_t numread = 0;
	size_t allread = 0;
	size_t allline = 0;
	//------------------------------------
	if(check_last_newline){
		if(last_written == 0x0a || last_written == 0x0d){
			if(!feof(tfp)){
				last_written = getc( tfp );
				if(last_written == 0x0a || last_written == 0x0d){
					last_written = getc( tfp );
					if(last_written == 0x0a || last_written == 0x0d){
						// Empty.
					}
					else{
						ungetc(last_written,tfp);
					}
				}
				else{
					ungetc(last_written,tfp);
				}
			}
		}
	}
	//------------------------------------
	unsigned int i;
	while(!feof(tfp)){
		numread = fread( buffer,1,numcount,tfp);
		fwrite( buffer, 1, numread, ofp );
		// bytes.
		allread += numread;
		// line no.
		if(readline){
			sp = buffer;
			for(i=0;i<numread;i++){
				if(*sp++ == 0x0a) allline++;
			}
		}
	}
	fclose(tfp);
	if(readbyte) *readbyte = allread;
	if(readline) *readline = allline;
	return true;
}

void CacheExportManager::outputEmptyTextRecord(
	FILE *ofp,
	const char* label
)
{
	const char* labelp = "";
	if(label){
		labelp = label;
	}
	outputReturn( ofp );
	fprintf( ofp,"\\begin{record}{\"%s\"}", labelp ); outputReturn( ofp );
	fprintf( ofp,"\\begin{data}" ); outputReturn( ofp );
	fprintf( ofp,"\\end{data}" ); outputReturn( ofp );
	fprintf( ofp,"\\end{record}" ); outputReturn( ofp );
}

bool CacheExportManager::outputDividedRecord(
	FILE *ofp,
	const char* label,
	const char* link_name,
	REC_POSITION_TYPE* writebyte,
	REC_LINENUMBER_TYPE* writeline
)
{
	if(!link_name) return false;
	const char* labelp = "";
	if(label){
		labelp = label;
	}
	// Writing bytes.
	REC_POSITION_TYPE writing_bytes = 0;
	int rt = 0;
	// NEW LINE.
	rt = outputReturn( ofp );
	if(rt >= 0) writing_bytes += rt;
	else return false;
	// \\begin{record}{"label"}
	rt = fprintf( ofp,"\\begin{record}{\"%s\"}", labelp );
	if(rt >= 0) writing_bytes += rt;
	else return false;
	rt = outputReturn( ofp );
	if(rt >= 0) writing_bytes += rt;
	else return false;
	// \\link{"link_name"}
	rt = fprintf( ofp,"\\link{\"%s\"}", link_name );
	if(rt >= 0) writing_bytes += rt;
	else return false;
	rt = outputReturn( ofp );
	if(rt >= 0) writing_bytes += rt;
	else return false;
	// \\end{record}
	rt = fprintf( ofp,"\\end{record}" );
	if(rt >= 0) writing_bytes += rt;
	else return false;
	rt = outputReturn( ofp );
	if(rt >= 0) writing_bytes += rt;
	else return false;

	if(writebyte){
		*writebyte = writing_bytes;
	}
	if(writeline){
		*writeline = 4;
	}
	return true;
}


static bool file_exist( const char* filename )
{
	return (fcache_file_exist( filename ) > 0);
}

bool
CacheExportManager::fnext_name( const char* filename, string& nextname )
{
	nextname = filename;
	nextname = nextname + "~";
	while(fcache_file_exist(nextname.c_str())){
		nextname = nextname + "~";
	}
	return true;
}
bool
CacheExportManager::fbackup_name( const char* filename, string& nextname )
{
	string pathpart, filepart, tempname;
	_fsmgr->splitpath(filename,pathpart,filepart);
	tempname = filepart;
	tempname = "~" + tempname;
	_fsmgr->makepath(pathpart.c_str(),tempname.c_str(),nextname);
	while(file_exist(nextname.c_str())){
		tempname = "~" + tempname;
		_fsmgr->makepath(pathpart.c_str(),tempname.c_str(),nextname);
	}
	return true;
}

// A temporary alias file name is returned.
// Input value of outstatus.
//   (1) eCACHE_OUTPUT_FILE_OTHERFILE
//   (2) eCACHE_OUTPUT_FILE_OVERWRITE
// Output value of outstatus.
//   (1) eCACHE_OUTPUT_FILE_ABNORMAL
//   (2) eCACHE_OUTPUT_FILE_NORMAL
//   (3) eCACHE_OUTPUT_FILE_EXIST
bool
CacheExportManager::outputFilename(
	E_CACHE_OUTPUT_FILE& outstatus,   // IN/OUT
	const char* filename,             // IN
	string& outname                   // OUT
)
{
	short fexist = 0;
	if(file_exist(filename)){
		fexist = 1;
	}
	switch(outstatus){
	// When OVERWRITE(ing), a temporary output file is found and returned.
	case eCACHE_OUTPUT_FILE_NORMAL:
	case eCACHE_OUTPUT_FILE_OVERWRITE:
		if(!fexist){
			outstatus = eCACHE_OUTPUT_FILE_ABNORMAL;
			return false;
		}
		if(fnext_name(filename,outname)){
			outstatus = eCACHE_OUTPUT_FILE_NORMAL;
			return true;
		}
		break;
	// When saved to the file of a different name,
	// it investigates whether there is already the same file.
	case eCACHE_OUTPUT_FILE_OTHERFILE:
		if(fexist){
			outstatus = eCACHE_OUTPUT_FILE_EXIST;
			outname = filename;
			return true; // false; // true:OVERWRITING!
		}
		else{
			outstatus = eCACHE_OUTPUT_FILE_NORMAL;
			outname = filename;
			return true;
		}
	}
	return false;
}

bool CacheExportManager::saveQuickCurrentRecord( const char* outputFile )
{
	UINT CurRec = _controller->getCurrentRec();
	if(_controller->isRecordEdited(0)) return false;
	if(_fsmgr->quickOutputAppendRecord()){
		short useMode = 0;
		const char* label = 0;
		// Check filename. //////////////////////////////////
		const char* curfile = _fsmgr->getFile();
		if(outputFile){
			if(_fsmgr->pathCompare(curfile,outputFile) != 0){
				return false;
			}
		}
		/////////////////////////////////////////////////////
		string labelString;
		string outputDivideRecordLinkName;
		REC_LINENUMBER_TYPE line_number = 0;
		REC_LINENUMBER_TYPE lastLine = 0;
		REC_POSITION_TYPE beforePosition = 0;
		REC_POSITION_TYPE afterPosition = 0;
		// Output current record data.
		bool rb = outputCurrentRecMultiways(
			curfile,
			0,
			1,
			0,
			2, // divide_record_mode = 0:normal(no-divide),1:depend on state,2:1 + with record-line.
			&useMode,
			&beforePosition,
			&afterPosition,
			&line_number,
			&outputDivideRecordLinkName);
		if(rb){
			// Delete virtual record data.
			if(_fsmgr->getRecInfoExtend(CurRec,&label)){ // ,false)){
				if(label) labelString = label;
			}
			CacheRecordOperation* recop = _fsmgr->getRecordOperation();
			recop->clearRecordOperation();
			//-CacheVirtualRecord* vrec = _fsmgr->getVirtualRecord();
			//-vrec->clearVirtualRecord();
			// Append proper record data.
			switch(useMode){
			case -1: // binary.
				lastLine = 0;
				_fsmgr->setLastPosition(afterPosition);
				break;
			case 1: // udf text.
				lastLine = _fsmgr->getLastLineNo();
				line_number += lastLine;
				_fsmgr->setLastLineNo(line_number);
				_fsmgr->setLastPosition(afterPosition);
				break;
			case 2: // xml.
				lastLine = 0;
				break;
			}
			// Binary,Text or other.
			if(outputDivideRecordLinkName.length() <= 0){
				_fsmgr->setRecInfo( CurRec, labelString.c_str(), lastLine, beforePosition );
			}
			else{
				_fsmgr->setRecInfo( CurRec, labelString.c_str(), lastLine, beforePosition, outputDivideRecordLinkName.c_str() );
			}
		}
		return rb;
	}
	return false;
}

///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
CacheFileCommandData::CacheFileCommandData( string& file, UINT record )
{
	_file = file;
	_record = record;
}
CacheFileCommandData::CacheFileCommandData( CacheFileCommandData& cd )
{
	_file = cd._file;
	_record = cd._record;
}
static bool
comp_rec( const CacheFileCommandData* t1, const CacheFileCommandData* t2 )
{
	return ( t1->_record < t2->_record );
}
/********************************
bool
CacheExportManager::searchFile()
{
	FILE* fp;
	int rec;
	char *fs, buff[256],dummy[32],file[256];
	string curname = _fsmgr->getFile();
	vector<string> extname;
	//fudf_create_extfile_name(curname,extname);
	for(int i=0;i<extname.size();i++){
		fp = fopen(extname[i].c_str(),"r");
		if(fp){
			fs = fgets(buff,255,fp);
			if(fs){
				if(!strncmp(buff,CCACHE_OUTPUT_COMMAND_REPLACE,strlen(CCACHE_OUTPUT_COMMAND_REPLACE))){
					sscanf(buff,"%s%d,%s}",dummy,&rec,file);
					CacheFileCommandData* cfcd = new CacheFileCommandData(extname[i],(UINT)rec);
					_searched.push_back(cfcd);
				}
			}
			fclose(fp);
		}
	}
	if(_searched.size() >= 2){
		sort(_searched.begin(),_searched.end(),comp_rec);
	}
	return true;
}
CacheFileCommandData*
CacheExportManager::nextCond()
{
	CacheFileCommandData* ccc;
	do {
		ccc = _searched[_cond_counter];
		_cond_counter++;
		if(_cond_counter >= _searched.size()){
			return 0x0;
		}
	} while(_cond_prev_rec != ccc->_record );
	_cond_prev_rec = ccc->_record;
	return ccc;
}
********************************/

