
#ifdef WIN32
#pragma warning (disable:4786)
#endif

#ifdef WIN32
# include <stdio.h>
# include <io.h>
#else
#include <unistd.h>
#endif

#include "FileStreamManager.h"
#include "CacheEventController.h"
#include "CacheExportManager.h"
#include "StructureCacheOperation.h"
#include "DataAttrCache.h"
#include "cache_util_function.h"
#include "StructureCacheManager.h"
#include "binaryOutput.h"
#include "BinaryFileLib.h"
#include "textudfOutput.h"
#include "util_file_seek.h"
#include "cache_file_utility.h"
#include <string>
#include <vector>
#include <sstream>
#include <algorithm>
using namespace std;


bool makeDivideBinaryPaths(
	string& divide_directory_path,
	string& divide_record_prifix,
	FileStreamManager* fmgr,
	StructureCacheOperation* operation,
	bool if_not_mkdir
)
{
	bool directory_ok = false;
	if(operation->getDivideDirectory()){
		// Read from DIVD header data.
		divide_directory_path = operation->getDivideDirectory();
		int dir_status = fcache_directory(divide_directory_path.c_str());
		if(dir_status > 0){
			directory_ok = true;
		}
	}
	if(!directory_ok){
		// Use default data.
		string read_path = fmgr->getFile();
		if(getDivideBinaryDirName(read_path, divide_directory_path)){
			int dir_status = fcache_directory(divide_directory_path.c_str());
			if(dir_status > 0){
				directory_ok = true;
			}
			else if(if_not_mkdir) {
				dir_status = fcache_dir_create(divide_directory_path.c_str());
				if (dir_status >= 0) {
					directory_ok = true;
				}
			}
		}
	}
	if(directory_ok){
		if(operation->getDivideRecordPrefix()){
			// Read from DIVD header data.
			divide_record_prifix = operation->getDivideRecordPrefix();
		} else {
			// Use default data.
			string read_path = fmgr->getFile();
			getDivideBinaryRecordPrefix(read_path, divide_record_prifix);
		}
	}
	return divide_record_prifix.length() > 0;
}

void setEraseFiles(CacheRecordOperation* recop, FileStreamManager* fsmgr, vector<string>& erase_files)
{
	vector<unsigned int> erase_records;
	if (recop->getErasedOriginalRecords(erase_records) > 0) {
		bool is_binary_record;
		string record_file;
		for (unsigned int i = 0; i < erase_records.size(); i++) {
			unsigned int original_erase_rec = (erase_records[i] + 1); // COMMON=0, STEP0=1, STEP1=2, ...
			if (fsmgr->getRecInfo(original_erase_rec, 0, 0, 0, &record_file, &is_binary_record)) {
				if (fcache_file_exist(record_file.c_str()) <= 0) {
					continue;
				}
				// Create temporary name.
				string rename_path;
				fudf_create_extfile_name(record_file, rename_path);
				rename(record_file.c_str(), rename_path.c_str());
				erase_files.push_back(rename_path);
			}
		}
	}
}

struct TempFilePair {
	string cur_file_path;
	string new_file_path;
};

bool getDivideBinaryRecordPath(
	const char* udffile_path,
	const char* record_file_name,
	unsigned int record,
	string& record_path,
	int get_opt,
	bool make_directory){
	string outputfile;
	if (udffile_path) {
		outputfile = udffile_path;
	}
	string dir_path;
	string record_prefix;
	if (!getDivideBinaryDirName(outputfile, dir_path)) {
		return false;
	}
	if (get_opt > 0) {
		record_path = dir_path;
		return true;
	}
	if (record_file_name) {
		if (record_file_name[0]) {
			record_prefix = record_file_name;
		}
	}
	if (record_prefix.length() == 0) {
		if (!getDivideBinaryRecordPrefix(outputfile, record_prefix)) {
			return false;
		}
	}
	outputfile.clear();
	if (make_directory) {
		if (fcache_directory(dir_path.c_str()) == 0) {
			if (fcache_dir_create(dir_path.c_str()) != 0) {
				return false;
			}
		}
	}
	buildDivideRecordFileName(record, dir_path, record_prefix, record_path);
	return true;
}

void makeLabel(unsigned int record, string& rec_label)
{
	stringstream   strbuf;
	strbuf << "Rec " << record;
	rec_label = strbuf.str();
}

bool writeDivideBinaryRecord(
	StructureCacheOperation* operation,
	const char* file_path, // UDF file path.
	const char* label,
	const char* record_file_name)
{
	BinaryOutput binaryOut;
	if (!binaryOut.open(file_path, false)) {
		return false;
	}
	int section = 1;
	return operation->writeBinaryData(section, label, binaryOut.getOutput(), false);
}

bool CacheExportManager::writeDivideBinaryRecordFile(const char* outfile_path, bool& reload)
{
	reload = false;
	if (!_fsmgr->isBinaryMode()) return false;
	bool common_record_written = false;

	// SAVE FILE TYPE.
	bool format_changed = false;
	string outputfile;
	bool save_as_other = false;
	if (outfile_path) {
		save_as_other = true;
		outputfile = outfile_path;
	}

	string dir_path;
	string record_prefix;
	if (save_as_other) {
		if (!getDivideBinaryDirName(outputfile, dir_path)) {
			return false;
		}
		if (!getDivideBinaryRecordPrefix(outputfile, record_prefix)) {
			return false;
		}
		outputfile.clear();
		if (fcache_directory(dir_path.c_str()) == 0) {
			if (fcache_dir_create(dir_path.c_str()) != 0) {
				return false;
			}
		}
		else {
			// Delete record files.
		}
	}
	else {
		if (!makeDivideBinaryPaths(dir_path, record_prefix, _fsmgr, _operation)) {
			return false;
		}
	}
	bool is_binary_record;
	string record_file;
	vector<string> erase_files;
	vector<TempFilePair> rename_files;

	//--- CHANGED RECORD -----------------------------------------
	CacheRecordOperation* recop = _fsmgr->getRecordOperation();
	unsigned int start_change_record = recop->getStartChangeRecord();
	unsigned int start_append_record = recop->getStartNewRecord();
	//------------------------------------------------------------
	if (save_as_other == false) { // Set erased record files.
		setEraseFiles(recop, _fsmgr, erase_files);
	}
	//------------------------------------------------------------
	unsigned int startRecord = _fsmgr->getBinaryStartRecord();
	// Record ///////////////////////////////////////////
	bool useVirtualRecord = false;
	unsigned int ii;
	unsigned int loopRecordSize = recop->getRecordNumber();
	for (ii = startRecord; ii <= loopRecordSize; ii++) {
		if (ii >= start_append_record) break;
		if (!_fsmgr->isBinaryRecord(ii)) continue;
		if (recop->isEditedRecord(ii)) {
			// Add 2021/10/29.
			unsigned int original_record;
			if (recop->isOriginal(ii)) {
				original_record = recop->getOriginalRecord(ii);
			}
			if (save_as_other == true || ii == original_record) {
				continue;
			}
			if (!_fsmgr->getRecInfo(original_record, 0, 0, 0, &record_file, &is_binary_record)) {
				continue;
			}
			erase_files.push_back(record_file);
		}
		else {
			unsigned int original_record;
			if (recop->isOriginal(ii)) {
				original_record = recop->getOriginalRecord(ii);
			}
			if (save_as_other == false && ii == original_record) {
				continue;
			}
			if (!_fsmgr->getRecInfo(original_record, 0, 0, 0, &record_file, &is_binary_record)) {
				continue;
			}
			reload = true;
			/* Copy original record file. */
			buildDivideRecordFileName(ii, dir_path, record_prefix, outputfile);
			if (save_as_other) {
				if (fcache_file_copy(record_file.c_str(), outputfile.c_str()) != 0) {
					continue;
				}
			}
			else {
				if (fcache_file_exist(record_file.c_str()) <= 0) {
					continue;
				}
				// Create temporary name.
				string rename_path;
				fudf_create_extfile_name(record_file, rename_path);
				TempFilePair pair;
				pair.cur_file_path = rename_path;
				pair.new_file_path = outputfile;
				rename_files.push_back(pair);
				rename(record_file.c_str(), rename_path.c_str());
				//if (fcache_file_copy(record_file.c_str(), outputfile.c_str()) != 0) {
				//	continue;
				//}
				//_fsmgr->setRecFile(ii, outputfile.c_str(), true);
			}
		}
	}
	for(ii = startRecord; ii <= loopRecordSize ; ii++ ){
		if(ii >= start_append_record) break;
		if(!_fsmgr->isBinaryRecord(ii)) continue;
		if(recop->isEditedRecord( ii )){
			reload = true;
			char* record_filepath = recop->getEditedRecordFilePath( ii );
			if (record_filepath) {
				BinaryOutput temporary;
				if (temporary.openRead(record_filepath)) {
					if (save_as_other) {
						buildDivideRecordFileName(ii, dir_path, record_prefix, outputfile);
						if (temporary.open(outputfile.c_str(), false) != NULL) {
							if (!temporary.extractCopy(-1)) {
								return false;
							}
							temporary.closeWrite();
						}
						//_fsmgr->setRecFile(ii, outputfile.c_str(), true);
					}
					else { // Save self.
						/* Copy edited record file. */
						if (!_fsmgr->getRecInfo(ii, 0, 0, 0, &record_file, &is_binary_record)) {
							buildDivideRecordFileName(ii, dir_path, record_prefix, record_file);
							if (record_file.length() == 0) {
								continue;
							}
						}
						if (fcache_file_exist(record_file.c_str()) > 0) {
							if (fcache_file_delete(record_file.c_str()) < 0) {
								return false;
							}
						}
						if (temporary.open(record_file.c_str(), false) != NULL) {
							if (!temporary.extractCopy(-1)) {
								return false;
							}
							temporary.closeWrite();
						}
						//_fsmgr->setRecFile(ii, record_file.c_str(), true);
					}
					/*- Common binary divide record has written. -*/
					if (ii == 0) {
						common_record_written = true; /* common record file is going to delete. */
					}
				}
				else {
					return false;
				}
			}
		} else{
			// pass
		}
	}

	// Erase the erased record files.
	for (ii = 0; ii < erase_files.size(); ii++) {
		if (fcache_file_exist(erase_files[ii].c_str())  > 0) {
			fcache_file_delete(erase_files[ii].c_str());
		}
	}
	// Rename the record files.
	for (ii = 0; ii < rename_files.size(); ii++) {
		if (fcache_file_exist(rename_files[ii].new_file_path.c_str())  > 0){
			fcache_file_delete(rename_files[ii].new_file_path.c_str());
		}
		rename(rename_files[ii].cur_file_path.c_str(), rename_files[ii].new_file_path.c_str());
	}
	
	// Append record. ///////////////////////////////////////
	if(loopRecordSize < start_append_record){
		recop->clearRecordOperation( common_record_written );
		return true;
	}
	UINT apnd_recno = 0;
	const char* apnd_label = 0x0;
	for (ii = start_append_record; ii <= loopRecordSize; ii++) {
		if (recop->isEditedRecord(ii)) {
			reload = true;
			char* record_filepath = recop->getEditedRecordFilePath(ii);
			if (record_filepath) {
				BinaryOutput temporary;
				if (temporary.openRead(record_filepath)) {
					buildDivideRecordFileName(ii, dir_path, record_prefix, outputfile);
					if (temporary.open(outputfile.c_str(), false) != NULL) {
						if (!temporary.extractCopy(-1)) {
							return false;
						}
						temporary.closeWrite();
					}
					// Do not need to set.
					//const char* label = recop->getRecordLabel(ii);
					//_fsmgr->setRecInfo(ii, label, 0, 0, record_file.c_str(), true);
					/*- Common binary divide record has written. -*/
					if (ii == 0) {
						common_record_written = true; /* common record file is going to delete. */
					}
				}
			}
		}
	}
	recop->clearRecordOperation( common_record_written );
	return true;
}

bool CacheExportManager::splitDivideBinaryRecordFile(const char* outfile_path)
{
	if (_fsmgr->isBinaryMode() == false) return false;
	if (_fsmgr->isDividedRecord() == true) return false;
	if (_fsmgr->getSaveDivideMode() == false) return false;

	/* Can not convert the same file name.
	if (outfile_path == 0 || outfile_path[0] == 0) {
		return false;
	} */

	// SAVE FILE TYPE.
	string outputfile;
	bool save_as_other = false;
	if (outfile_path) {
		save_as_other = true;
		outputfile = outfile_path;
	}

	string dir_path;
	string record_prefix;
	if (save_as_other) {
		if (!getDivideBinaryDirName(outputfile, dir_path)) {
			return false;
		}
		if (!getDivideBinaryRecordPrefix(outputfile, record_prefix)) {
			return false;
		}
		outputfile.clear();
		if (fcache_directory(dir_path.c_str()) == 0) {
			if (fcache_dir_create(dir_path.c_str()) != 0) {
				return false;
			}
		}
		else {
			// Delete record files.
		}
	}
	else {
		// Record directory is created if it is not.
		if (!makeDivideBinaryPaths(dir_path, record_prefix, _fsmgr, _operation, true)) {
			return false;
		}
	}
	bool is_binary_record;
	string record_file;

	//--- CHANGED RECORD -----------------------------------------
	CacheRecordOperation* recop = _fsmgr->getRecordOperation();
	unsigned int start_change_record = recop->getStartChangeRecord();
	unsigned int start_append_record = recop->getStartNewRecord();
	//------------------------------------------------------------
	unsigned int startRecord = _fsmgr->getBinaryStartRecord();
	// Record ///////////////////////////////////////////
	bool useVirtualRecord = false;
	unsigned int ii;
	unsigned int loopRecordSize = recop->getRecordNumber();
	for (ii = startRecord; ii <= loopRecordSize; ii++) {
		if (ii >= start_append_record) break;
		if (!_fsmgr->isBinaryRecord(ii)) continue;
		if (recop->isEditedRecord(ii)) {
			char* record_filepath = recop->getEditedRecordFilePath(ii);
			if (record_filepath) {
				BinaryOutput temporary;
				if (temporary.openRead(record_filepath)) {
					if (save_as_other) {
						buildDivideRecordFileName(ii, dir_path, record_prefix, outputfile);
						if (temporary.open(outputfile.c_str(), false) != NULL) {
							if (!temporary.extractCopy(-1)) {
								return false;
							}
							temporary.closeWrite();
						}
						//_fsmgr->setRecFile(ii, outputfile.c_str(), true);
					}
					else { // Save self.
						/* Copy edited record file. */
						if (!_fsmgr->getRecInfo(ii, 0, 0, 0, &record_file, &is_binary_record)) {
							buildDivideRecordFileName(ii, dir_path, record_prefix, record_file);
							if (record_file.length() == 0) {
								continue;
							}
						}
						if (fcache_file_delete(record_file.c_str()) < 0) {
							return false;
						}
						if (temporary.open(record_file.c_str(), false) != NULL) {
							if (!temporary.extractCopy(-1)) {
								return false;
							}
							temporary.closeWrite();
						}
						//_fsmgr->setRecFile(ii, record_file.c_str(), true);
					}
				}
				else {
					return false;
				}
			}
		}
		else {
			unsigned int record_number = ii;
			if (recop->isOriginal(ii)) {
				record_number = recop->getOriginalRecord(ii);
			}
			if (save_as_other) {
				/* Copy original record file. */
				buildDivideRecordFileName(ii, dir_path, record_prefix, outputfile);
				if (ii == 0) {
					// Write common data.
					BinaryOutput binIO;
					FILE* writeFP = binIO.open(outputfile.c_str(), false);
					binIO.outRECORD(ii, 0, _operation);
					_fsmgr->setRecFile(ii, outputfile.c_str(), true);
					binIO.close();
					continue;
				}
				// Write record data.
				bool rb;
				UINT line;
				REC_POSITION_TYPE pos;
				const char* sourceFile = (const char*)_fsmgr->getFile();
				rb = _fsmgr->getRecInfo(record_number, &line, &pos);
				if (!rb) {
					continue;
				}
				BinaryOutput binIO;
				FILE* sourceFP = binIO.openRead(sourceFile);
				if (!sourceFP) {
					continue;
				}
				if (!binIO.seekRead(pos)) {
					return false;
				}
				FILE* writeFP = binIO.open(outputfile.c_str(), false);
				if (!writeFP) {
					continue;
				}
				// Copy record data.
				if (!_operation->binaryCopy(writeFP, sourceFP)) {
					continue;
				}
				binIO.close();
				_fsmgr->setRecFile(ii, outputfile.c_str(), true);
			}
		}
	}

	// Append record. ///////////////////////////////////////
	if (loopRecordSize < start_append_record) {
		recop->clearRecordOperation( false );
		return true;
	}
	UINT apnd_recno = 0;
	const char* apnd_label = 0x0;
	for (ii = start_append_record; ii <= loopRecordSize; ii++) {
		if (recop->isEditedRecord(ii)) {
			char* record_filepath = recop->getEditedRecordFilePath(ii);
			if (record_filepath) {
				BinaryOutput temporary;
				if (temporary.openRead(record_filepath)) {
					buildDivideRecordFileName(ii, dir_path, record_prefix, outputfile);
					if (temporary.open(outputfile.c_str(), false) != NULL) {
						if (!temporary.extractCopy(-1)) {
							return false;
						}
						temporary.closeWrite();
					}
					// Do not need to set.
					//const char* label = recop->getRecordLabel(ii);
					//_fsmgr->setRecInfo(ii, label, 0, 0, record_file.c_str(), true);
				}
			}
		}
	}
	recop->clearRecordOperation( false ); // Do not delete a common file.
	return true;
}

bool CacheExportManager::mergeDivideBinaryRecordFile(FILE* binaryWriteOpened)
{
	if (!_fsmgr->isBinaryMode()) return false;
	if (!_fsmgr->isDividedRecord()) return false;

	BinaryOutput standardBinIO;
	FILE* sourceFP;

	// begin{binary} ////////////////////////////////////
	standardBinIO.setOutput(binaryWriteOpened);
	if (!standardBinIO.outBEGIN()) {
		standardBinIO.resetOutput();
		return false;
	}
	if (!standardBinIO.outHEADER(_operation, 0)) {
		standardBinIO.resetOutput();
		return false;
	}

	const char* rec_label = 0x0;
	bool is_binary_record;
	string record_file;
	
	//--- CHANGED RECORD -----------------------------------------
	CacheRecordOperation* recop = _fsmgr->getRecordOperation();
	unsigned int start_change_record = recop->getStartChangeRecord();
	unsigned int start_append_record = recop->getStartNewRecord();
	//------------------------------------------------------------
	unsigned int startRecord = _fsmgr->getBinaryStartRecord();
	// Record ///////////////////////////////////////////
	bool useVirtualRecord = false;
	unsigned int ii;
	unsigned int loopRecordSize = recop->getRecordNumber();
	//-unsigned int loopRecordSize = _fsmgr->getFileRecSize();
	for (ii = startRecord; ii <= loopRecordSize; ii++) {
		if (ii >= start_append_record) break;
		if (!_fsmgr->isBinaryRecord(ii)) continue;
		useVirtualRecord = false;
		if (ii == 0) {
			// Write common data.
			standardBinIO.outRECORD(ii, 0, _operation);
			standardBinIO.closeRead();
			continue;
		}
		if (recop->isEditedRecord(ii)) {
			char* record_filepath = recop->getEditedRecordFilePath(ii);
			if (record_filepath) {
				sourceFP = standardBinIO.openRead(record_filepath);
				if (sourceFP) {
					if (!_operation->binaryCopy(binaryWriteOpened, sourceFP)) {
						return false;
					}
					standardBinIO.closeRead();
					useVirtualRecord = true;
				}
				else {
					// Write an empty record.
					standardBinIO.outEMPTY(ii, 0, _operation);
					useVirtualRecord = true;
				}
			}
			else {
				// Write an empty record.
				standardBinIO.outEMPTY(ii, 0, _operation);
				useVirtualRecord = true;
			}
		}
		else {
			unsigned int record_number = ii;
			if (recop->isOriginal(ii)) {
				record_number = recop->getOriginalRecord(ii);
			}
			if (!_fsmgr->getRecInfo(record_number, &rec_label, 0, 0, &record_file, &is_binary_record)) {
				// Write an empty record.
				standardBinIO.outEMPTY(record_number, rec_label, _operation);
			}
			sourceFP = standardBinIO.openRead(record_file.c_str());
			if (sourceFP) {
				if (!_operation->binaryCopy(binaryWriteOpened, sourceFP)) {
					return false;
				}
				standardBinIO.closeRead();
			}
			else {
				// Write an empty record.
				standardBinIO.outEMPTY(record_number, 0, _operation);
			}
		}
	}

	// Append record. ///////////////////////////////////////
	if (loopRecordSize < start_append_record) {
		recop->clearRecordOperation();
		return true;
	}
	UINT apnd_recno = 0;
	const char* apnd_label = 0x0;
	for (ii = start_append_record; ii <= loopRecordSize; ii++) {
		if (recop->isEditedRecord(ii)) {
			char* filepath = recop->getEditedRecordFilePath(ii);
			rec_label = recop->getRecordLabel(ii);
			if (filepath) {
				sourceFP = standardBinIO.openRead(filepath);
				if (sourceFP) {
					if (!_operation->binaryCopy(binaryWriteOpened, sourceFP)) {
						return false;
					}
					standardBinIO.closeRead();
				}
				else {
					// Write an empty record.
					standardBinIO.outEMPTY(ii, rec_label, _operation);
				}
			}
			else { // Empty record output.
				// Write an empty record.
				standardBinIO.outEMPTY(ii, rec_label, _operation);
			}
		}
	}
	recop->clearRecordOperation();
	return true;
}

