
/*
 *  Material Modeling Platform Package
 *  Copyright(c) 2007, The Japan Research Institute, Ltd.
 *  All rights reserved.
 *  
 *  $Id: CacheRecordOperation.cpp,v 1.10 2007/04/29 08:50:12 nishimoto Exp $
 */

#include "CacheRecordOperation.h"
#include "file_stream_record.h"
#include "cache_util_character.h"
using namespace std;

char* fcache_record_operation_strdup( const char* ss )
{
	if(!ss) return NULL;
	int len = strlen( ss );
	char* cc = new char[len + 1];
	strcpy(cc,ss);
	return cc;
}

bool fcache_record_operation_file_exist( const char* filename )
{
	return (fcache_file_exist( filename ) > 0);
}
static char* get_abc( int number, char* buffer )
{
	buffer[0] = '\0';
	if(number <= 0){
		return buffer;
	}
	number--;
	char *ss, *p, tmp[128];
	ss = buffer;
	p = tmp;
	if(number < 0){
		number = -number;
	}
	do{
		*p++ = (char)(number%10 + (int)'A');
		number /= 10;
	} while(number);
	while(p != tmp){
		*ss++ = *--p;
	}
	*ss++ = 0x0;
	return buffer;
}
void fcache_record_divide_filename( char* buffer, unsigned int record, int number )
{
	if(number <= 1){
		sprintf(buffer,"rec%08d.du",record);
	}
	else if(number <= 2){
		sprintf(buffer,"rec%08dA.du",record);
	}
	else if(number <= 3){
		sprintf(buffer,"rec%08dB.du",record);
	}
	else{
		char sbuff[128];
		number--;
		get_abc( number, sbuff );
		sprintf(buffer,"rec%08d%s.du",record,sbuff);
	}
}

// filepath
bool fcache_make_output_directory( const char* outfile, const char* post_fix, string& outpath )
{
	const char* curr_file = outfile;
	if(!curr_file) return false;
	if(!*curr_file) return false;
	outpath = curr_file;
	if(post_fix){
		if(post_fix[0] != '.'){
			outpath = outpath + ".";
		}
		outpath = outpath + post_fix;
	}
	return true;
}

bool fcache_make_record_divide_filename( string& name, const string& directory, unsigned int record, bool check_exist )
{
	char buffer[128];
	string path;
	for(int n = 0; n < 1000 ; n++ ){
		buffer[0] = '\0';
		fcache_record_divide_filename( buffer, record, n );
		name = buffer;
		// Make path.
		path = directory + "/";
		path = path + buffer;
		if(!check_exist){
			return true;
		}
		// Check exist.
		if(fcache_record_operation_file_exist( path.c_str() )){
			continue;
		}
		return true;
	}
	return false;
}

//
// class CacheEditedRecord.
//
CacheEditedRecord::CacheEditedRecord( unsigned int record )
:file_path(0),record_label(0)
{
	original_record_number = record;
}
CacheEditedRecord::~CacheEditedRecord()
{
	clear();
}
void CacheEditedRecord::clear()
{
	// DELETE FILE.
	deleteFile();
	
	// MEMORY.
	if(record_label) delete[] record_label;
	if(file_path) delete[] file_path;
	original_record_number = 0;
	record_label = 0x0;
	file_path = 0x0;
}
void CacheEditedRecord::deleteFile()
{
	// DELETE FILE.
	if(file_path){
		if(file_path[0]){
			if(fcache_record_operation_file_exist(file_path)){
				fcache_file_delete( file_path );
			}
		}
	}
}
void CacheEditedRecord::setRecordNo( const unsigned int record )
{
	original_record_number = record;
}
void CacheEditedRecord::setFile( const char* path )
{
	if(file_path) delete[] file_path;
	file_path = 0x0;
	if(path) file_path = fcache_record_operation_strdup(path);
}
void CacheEditedRecord::setLabel( const char* label )
{
	if(record_label) delete[] record_label;
	record_label = 0x0;
	if(label) record_label = fcache_record_operation_strdup(label);
}

// Edited container.
bool cache_compare_record_number(
	const CacheEditedRecord* a,
	const CacheEditedRecord* b
)
{
	return ( a->getRecordNo() < b->getRecordNo() );
}

//
// class CacheChangeRecord.
//
CacheChangeRecord::CacheChangeRecord(unsigned char change_type)
:change_type(change_type)
{
	this->change_type = change_type;
	memset(&record, 0, sizeof(InsertedOrOriginalRecord));
}
CacheChangeRecord::~CacheChangeRecord()
{
	switch(change_type){
	case CACHE_RECORD_OPERATION_OLD: // ORIGINAL.
		//if(record.original.file_path){
		//	delete[] record.original.file_path;
		//}
		break;
	case CACHE_RECORD_OPERATION_NEW: // INSERTED.
		if(record.insert.record_label){
			delete[] record.insert.record_label;
		}
		if(record.insert.file_path){
			// DELETE FILE.
			deleteFile();
			// MEMORY.
			delete[] record.insert.file_path;
			record.insert.file_path = NULL;
		}
		break;
	}
}
void CacheChangeRecord::deleteFile()
{
	if(change_type == CACHE_RECORD_OPERATION_NEW){
		if(record.insert.file_path){
			if(record.insert.file_path[0]){
				// DELETE FILE.
				if(fcache_record_operation_file_exist(record.insert.file_path)){
					fcache_file_delete( record.insert.file_path );
					//unlink(record.insert.file_path);
				}
			}
		}
	}
}
// Input.
bool CacheChangeRecord::setOld(unsigned int original_record) // , char* file)
{
	switch(change_type){
	case CACHE_RECORD_OPERATION_OLD: // ORIGINAL.
		record.original.original_record_number = original_record;
		//if(record.original.file_path){
		//	delete[] record.original.file_path;
		//	record.original.file_path = NULL;
		//}
		//if(file){
		//	record.original.file_path = fcache_record_operation_strdup( file );
		//}
		return true;
	}
	return false;
}
bool CacheChangeRecord::setNew(const char* label, char* file)
{
	switch(change_type){
	case CACHE_RECORD_OPERATION_NEW: // INSERTED.
		if(record.insert.record_label){
			delete[] record.insert.record_label;
			record.insert.record_label = NULL;
		}
		if(record.insert.file_path){
			delete[] record.insert.file_path;
			record.insert.file_path = NULL;
		}
		if(label){
			record.insert.record_label = fcache_record_operation_strdup( label );
		}
		if(file){
			record.insert.file_path = fcache_record_operation_strdup( file );
		}
		return true;
	}
	return false;
}
bool CacheChangeRecord::setLabel( const char* label )
{
	switch(change_type){
	case CACHE_RECORD_OPERATION_NEW: // INSERTED.
		if(record.insert.record_label){
			delete[] record.insert.record_label;
			record.insert.record_label = NULL;
		}
		if(label){
			record.insert.record_label = fcache_record_operation_strdup( label );
		}
		return true;
	}
	return false;
}
bool CacheChangeRecord::setFile( const char* file )
{
	switch(change_type){
	case CACHE_RECORD_OPERATION_OLD: // ORIGINAL.
		//if(record.original.file_path){
		//	delete[] record.original.file_path;
		//	record.original.file_path = NULL;
		//}
		//if(file){
		//	record.original.file_path = fcache_record_operation_strdup( file );
		//}
		//return true;
		return false;
	case CACHE_RECORD_OPERATION_NEW: // INSERTED.
		if(record.insert.file_path){
			delete[] record.insert.file_path;
			record.insert.file_path = NULL;
		}
		if(file){
			record.insert.file_path = fcache_record_operation_strdup( file );
		}
		return true;
	}
	return false;
}
// Output.
unsigned char CacheChangeRecord::getType()
{
	return change_type;
}
bool CacheChangeRecord::isOld()
{
	return (CACHE_RECORD_OPERATION_OLD == change_type);
}
bool CacheChangeRecord::isNew()
{
	return (CACHE_RECORD_OPERATION_NEW == change_type);
}
char* CacheChangeRecord::getLabel()
{
	switch(change_type){
	case CACHE_RECORD_OPERATION_NEW: // INSERTED.
		return record.insert.record_label;
	}
	return NULL;
}
bool CacheChangeRecord::sameLabel( const char* label )
{
	if(!label) return false;
	switch(change_type){
	case CACHE_RECORD_OPERATION_NEW: // INSERTED.
		if(record.insert.record_label){
			return ( strcmp(record.insert.record_label,label) == 0 );
		}
	}
	return false;
}
char* CacheChangeRecord::getFile()
{
	switch(change_type){
	case CACHE_RECORD_OPERATION_NEW: // INSERTED.
		return record.insert.file_path;
	}
	return NULL;
}
unsigned int CacheChangeRecord::getOriginalRecordNumber()
{
	switch(change_type){
	case CACHE_RECORD_OPERATION_OLD: // ORIGINAL.
		return record.original.original_record_number;
	}
	return CACHE_RECORD_NO_RECORD;
}

//
// class CacheRecordWork.
//
// private.
void CacheRecordWork::clearShiftRecordNumber() {
	minimum_edit_record = CACHE_RECORD_SHIFT_NOVALUE;
	//maximum_edit_record = 0;
	//maximum_erase_record = 0;
	serial_shift_record_number = CACHE_RECORD_SHIFT_NOVALUE;
}
bool CacheRecordWork::isShiftRecordNumber() {
	return !(serial_shift_record_number == CACHE_RECORD_SHIFT_NOVALUE);
}
void CacheRecordWork::setShiftRecordNumber(unsigned int shift) {
	serial_shift_record_number = shift;
}
unsigned int CacheRecordWork::getShiftRecordNumber() {
	if (!isShiftRecordNumber()) printf("Illegal using getShiftRecordNumber()\n");
	return serial_shift_record_number;
}
void CacheRecordWork::clearChangedRecord()
{
	if(change_records.size() > 0){
		unsigned int ii,nn = change_records.size();
		for(ii=0;ii<nn;ii++){
			CacheChangeRecord* crec = change_records[ii];
			if(crec){
				deleteRecord( crec );
			}
		}
		change_records.clear();
	}
}

// Free memory & delete temporary file.
void CacheRecordWork::deleteRecord( CacheChangeRecord* rec )
{
	if(!rec) return;
	if(rec->getType() == CACHE_RECORD_OPERATION_OLD){
		unsigned int original = rec->getOriginalRecordNumber();
		eraseEditedRecordByEraseRecordOperation( original );
	}
	delete rec;
}

void CacheRecordWork::eraseEditedRecordByEraseRecordOperation( unsigned int original )
{
	//printf("ERASED EDITED ORIGINAL RECORD:[%d]\n", original);
	if(!edited_records) return;
	if(edited_records->size() <= 0) return;
	CacheEditedRecord temp(original);
	CacheEditedRecord* pfind = edited_records->searchCut( &temp );
	if(pfind){
		delete pfind;
	}
}

void CacheRecordWork::check_minimum_edit_record(unsigned int record) {
	if (record < minimum_edit_record) {
		minimum_edit_record = record;
	}
}

// public.
CacheRecordWork::CacheRecordWork(unsigned int number_of_records)
{
	//file_creator_number = 0;
	clearShiftRecordNumber();

	original_number_of_records = number_of_records;
	//current_udffile = fcache_record_operation_strdup( udffile );
	// Sorted by record index.
	edited_records = new util_sorted::util_sorted_vector<CacheEditedRecord,CacheCompareRecordNumber,CacheEqualLabel>();
}
CacheRecordWork::~CacheRecordWork()
{
	//if(current_udffile){
	//	delete[] current_udffile;
	//}
	if(edited_records){
		delete edited_records;
		edited_records = NULL;
	}
	clearChangedRecord();
}
void CacheRecordWork::clearRecordOperation()
{
	clearChangedRecord(); // WITH ERASING EDITED FILES.
	if(edited_records){
		edited_records->clearDeleteElement(); // WITH ERASING EDITED FILES.
	}
	clearShiftRecordNumber();
	//clearDirectory();
}
// Delete temporary files.
void CacheRecordWork::deleteFiles()
{
	unsigned int i,n;
	if(edited_records){
		n = edited_records->size();
		for(i=0;i<n;i++){
			CacheEditedRecord* erec = edited_records->get( i );
			if(erec){
				erec->deleteFile();
			}
		}
	}
	n = change_records.size();
	for(i=0;i<n;i++){
		CacheChangeRecord* crec = change_records[i];
		if(crec){
			crec->deleteFile();
		}
	}
}
// Input.
bool CacheRecordWork::eraseRecord( unsigned int start, unsigned int number )
{
	if(number <= 0) return true; // trivial.
	unsigned int i;
	unsigned int nchanged = change_records.size();
	CacheChangeRecord* rec;
	if(original_number_of_records <= 0){ // No records.
		if(nchanged > 0){
			if(start >= nchanged){
				// error.
			}
			else{
				//   vvv <- ERASE
				// ++++++++
				if((start+number) > nchanged){
					number = nchanged - start;
				}
				for(i=0;i<number;i++){
					rec = change_records[i+start];
					deleteRecord( rec );
				}
				unsigned int new_num = nchanged - number;
				for(i=start+number;i<nchanged;i++){
					rec = change_records[i];
					change_records[i - number] = rec;
				}
				change_records.resize(new_num);
				return true;
			}
		}
		else{
			if(start <= 0){ // do nothing.
				return true;
			}
			// error.
		}
	}
	else{ // Some records.
		if(isShiftRecordNumber()){
			if(nchanged > 0){
				if(start < getShiftRecordNumber()){
					unsigned int end_index = start + number - 1;
					if(end_index < getShiftRecordNumber()){
						//+++ ERASE EDITED ORIGINAL RECORD +++
						for(i=start;i<(start+number);i++){
							eraseEditedRecordByEraseRecordOperation( i ); // erase original edited record.
						}
						//++++++++++++++++++++++++++++++++++++
						//   vvv <- ERASE
						// oooooo++++++++
						unsigned int nadd = getShiftRecordNumber() - end_index - 1;
						if(nadd > 0){
							change_records.resize( nchanged + nadd);
							if(nchanged > 0){
								unsigned int jj,nn = nchanged;
								for(jj=1;jj<=nn;jj++){
									i = nchanged - jj;
									rec = change_records[i];
									change_records[i+nadd] = rec;
								}
							}
							for(i=0;i<nadd;i++){
								rec = new CacheChangeRecord(CACHE_RECORD_OPERATION_OLD);
								unsigned int original_record = i + start + number;
								rec->setOld( original_record);
								change_records[i] = rec;
							}
							setShiftRecordNumber(start);
							return true;
						}
						else{
							//    vvv <- ERASE
							// oooooo++++++++
							setShiftRecordNumber(start);
							return true;
						}
					}
					else{
						//     vvvv <- ERASE
						// oooooo++++++++
						unsigned int nshiftrec = getShiftRecordNumber();
						//+++ ERASE EDITED ORIGINAL RECORD +++
						for(i=start;i<nshiftrec;i++){
							eraseEditedRecordByEraseRecordOperation( i ); // erase original edited record.
						}
						//++++++++++++++++++++++++++++++++++++
						unsigned int ndel = start + number - nshiftrec;
						for(i=0;i<ndel;i++){
							rec = change_records[i];
							deleteRecord( rec );
						}
						for(i=ndel;i<nchanged;i++){
							rec = change_records[i];
							change_records[i - ndel] = rec;
						}
						change_records.resize( nchanged - ndel );
						setShiftRecordNumber(start);
						return true;
					}
				}
				else{
					//        vvv <- ERASE
					// oooooo++++++++
					unsigned int start_array = start - getShiftRecordNumber();
					if(start_array >= nchanged){
						return false;
					}
					if((start_array + number) > nchanged){
						number = nchanged - start_array;
					}
					for(i=0;i<number;i++){
						rec = change_records[i+start_array];
						deleteRecord( rec );
					}
					unsigned int nres = nchanged - start_array - number;
					for(i=0;i<nres;i++){
						rec = change_records[i+start_array+number];
						change_records[i+start_array] = rec;
					}
					change_records.resize( start_array + nres);
					return true;
				}
			}
			else{ // No change_records!
				//   vvv <- ERASE
				// oooooo-------
				if(start < getShiftRecordNumber()){
					unsigned int end_index = start + number - 1;
					if(end_index < getShiftRecordNumber()){
						//+++ ERASE EDITED ORIGINAL RECORD +++
						for(i=start;i<(start+number);i++){
							eraseEditedRecordByEraseRecordOperation( i ); // erase original edited record.
						}
						//++++++++++++++++++++++++++++++++++++
						unsigned int nadd = getShiftRecordNumber() - end_index - 1;
						if(nadd > 0){
							change_records.resize(nadd);
							for(i=0;i<nadd;i++){
								rec = new CacheChangeRecord(CACHE_RECORD_OPERATION_OLD);
								unsigned int original_record = i + start + number;
								rec->setOld( original_record);
								change_records[i] = rec;
							}
							setShiftRecordNumber(start);
							return true;
						}
						else{
							setShiftRecordNumber(start);
							return true;
						}
					}
					else{
						//     vvv <- ERASE
						// oooooo-------
						//+++ ERASE EDITED ORIGINAL RECORD +++
						unsigned int nshiftrec = getShiftRecordNumber();
						for(i=start;i<nshiftrec;i++){
							eraseEditedRecordByEraseRecordOperation( i ); // erase original edited record.
						}
						//++++++++++++++++++++++++++++++++++++
						setShiftRecordNumber(start);
						return true;
					}
				}
				// error.
			}
		}
		else{ // Some records & no changes.
			if(nchanged > 0){
				return false; // Implementaion error.
			}
			else{ // No change. <OK>
				if(start < original_number_of_records){
					//+++ ERASE EDITED ORIGINAL RECORD +++
					for(i=start;i<(start+number);i++){
						eraseEditedRecordByEraseRecordOperation( i ); // erase original edited record.
					}
					//++++++++++++++++++++++++++++++++++++
					if((start + number) >= original_number_of_records){
						//     vvv <- ERASE
						// ooooooo
						setShiftRecordNumber(start);
						return true;
					}
					else{
						//     vvv <- ERASE
						// ooooooooo
						unsigned int nadd = original_number_of_records - start - number;
						change_records.resize( nadd );
						for(i=0;i<nadd;i++){
							rec = new CacheChangeRecord(CACHE_RECORD_OPERATION_OLD);
							unsigned int original_record = i + start + number;
							rec->setOld( original_record);
							change_records[i] = rec;
						}
						setShiftRecordNumber(start);
						return true;
					}
				}
			}
		}
	}
	return false;
}
bool CacheRecordWork::insertRecord( unsigned int start, unsigned int number, bool front )
{
	if(number <= 0) return true; // trivial.
	unsigned int i;
	unsigned int nchanged = change_records.size();
	CacheChangeRecord* rec;
	if(original_number_of_records <= 0){ // No records.
		if(nchanged <= 0){
			if(start <= 0){ // <OK>
				change_records.resize(number);
				for(i=0;i<number;i++){
					rec = new CacheChangeRecord(CACHE_RECORD_OPERATION_NEW);
					change_records[i] = rec;
				}
				setShiftRecordNumber( 0 );
				return true;
			}
			else{
				// error.
			}
		}
		else{ // some inserted. <OK>
			if(!front){
				start = start + 1;
			}
			if(start <= nchanged){
				//   v3 <- INSERT
				// ++++++  -> ++xxx++++
				change_records.resize(nchanged + number);
				if(start < nchanged){
					unsigned int jj,nn = nchanged - start;
					for(jj=1;jj<=nn;jj++){
						i = nchanged - jj;
						rec = change_records[i];
						change_records[i+number] = rec;
					}
				}
				for(i=0;i<number;i++){
					rec = new CacheChangeRecord(CACHE_RECORD_OPERATION_NEW);
					change_records[i+start] = rec;
				}
				setShiftRecordNumber( 0 );
				return true;
			}
			else{
				// error.
			}
		}
	}
	else{ // original records > 0.
		if(!front){
			start = start + 1;
		}
		if(isShiftRecordNumber()){
			if(nchanged <= 0){ // Erased last only.
				if(start > getShiftRecordNumber()){
					start = getShiftRecordNumber(); // No error.
				}
				if(start <= getShiftRecordNumber()){
					//   v3 <- INSERT
					// ooooo---  -> ooxxxooo---
					unsigned int nadd = getShiftRecordNumber() - start + number;
					change_records.resize( nadd );
					for(i=0;i<number;i++){
						rec = new CacheChangeRecord(CACHE_RECORD_OPERATION_NEW);
						change_records[i] = rec;
					}
					for(i=number;i<nadd;i++){
						rec = new CacheChangeRecord(CACHE_RECORD_OPERATION_OLD);
						unsigned int original_record = start + (i - number);
						rec->setOld(original_record);
						change_records[i] = rec;
					}
					setShiftRecordNumber( start );
					return true;
				}
				else{
					// error.
				}
			}
			else{ // some inserted.
				if(start < getShiftRecordNumber()){ // <OK>
					//   v3 <- INSERT
					// ooooo+++  -> ooxxxooo+++
					unsigned int nadd = getShiftRecordNumber() - start + number;
					change_records.resize( nchanged + nadd );
					unsigned int jj,nn = nchanged;
					for(jj=1;jj<=nn;jj++){
						i = nchanged - jj;
						rec = change_records[i];
						change_records[nadd + i] = rec;
					}
					for(i=0;i<number;i++){
						rec = new CacheChangeRecord(CACHE_RECORD_OPERATION_NEW);
						change_records[i] = rec;
					}
					for(i=number;i<nadd;i++){
						rec = new CacheChangeRecord(CACHE_RECORD_OPERATION_OLD);
						unsigned int original_record = start + (i - number);
						rec->setOld( original_record);
						change_records[i] = rec;
					}
					setShiftRecordNumber( start );
					return true;
				}
				else{ // <OK>
					if(start > ( nchanged + getShiftRecordNumber())){
						start = nchanged + getShiftRecordNumber();
					}
					//       v3 <- INSERT
					// ooooo+++  -> ooooo+xxx++
					change_records.resize( nchanged + number );
					unsigned int start_array = start - getShiftRecordNumber();
					if(start_array < nchanged){
						unsigned int jj,nn = nchanged - start_array;
						for(jj=1;jj<=nn;jj++){
							i = nchanged - jj;
							rec = change_records[i];
							change_records[number + i] = rec;
						}
					}
					for(i=0;i<number;i++){
						rec = new CacheChangeRecord(CACHE_RECORD_OPERATION_NEW);
						change_records[i + start_array] = rec;
					}
					return true;
				}
			}
		}
		else{
			if(nchanged <= 0){ // No changed. <OK>
				if(start > original_number_of_records){
					start = original_number_of_records;
				}
				//    v3 <- INSERT
				// ooooooo  -> oooxxxoooo
				unsigned int nadd = original_number_of_records - start + number;
				change_records.resize( nadd );
				for(i=0;i<number;i++){
					rec = new CacheChangeRecord(CACHE_RECORD_OPERATION_NEW);
					change_records[i] = rec;
				}
				for(i=number;i<nadd;i++){
					rec = new CacheChangeRecord(CACHE_RECORD_OPERATION_OLD);
					unsigned int original_record = start + (i - number);
					rec->setOld( original_record);
					change_records[i] = rec;
				}
				setShiftRecordNumber( start );
				return true;
			}
			else{
				// Implementation error.
			}
		}
	}
	return false;
}
bool CacheRecordWork::appendRecord( unsigned int number )
{
	unsigned int nrec = getRecordNumber();
	if(nrec == CACHE_RECORD_NO_RECORD){
		return false;
	}
	return insertRecord( nrec, number, true );
}
bool CacheRecordWork::setRecordLabel( unsigned int record, const char* label )
{
	//CacheChangeRecord* rec = getChangeRecord( record );
	//if(rec){
	//	return rec->setLabel( label );
	//}
	//return false;
	if(!label) return false;
	if(record >= getRecordNumber()){
		return false;
	}
	// Status
	unsigned char status = getStatus( record );
	if(status == 'O'){ // Original.
		unsigned int original = getOriginalRecord( record );
		CacheEditedRecord temp(original);
		CacheEditedRecord* pfind = edited_records->search( &temp );
		if(pfind){
			pfind->setLabel( label );
			return true;
		}
		else{
			// New. Label edited.
			CacheEditedRecord* newrec = new CacheEditedRecord(original);
			newrec->setLabel( label );
			edited_records->put( newrec );
			return true;
		}
	}
	else if(status == 'I'){
		CacheChangeRecord* rec = getChangeRecord( record );
		if(rec){
			return rec->setLabel( label );
		}
	}
	return false;
}
// Query.
bool CacheRecordWork::isChanged()
{
	return (isEdited() || isRecordChanged());
}
bool CacheRecordWork::isEdited()
{
	return (edited_records->size() > 0);
}
bool CacheRecordWork::isRecordChanged()
{
	return (CACHE_RECORD_SHIFT_NOVALUE != serial_shift_record_number);
}
// Output.
unsigned int CacheRecordWork::getRecordNumber()
{
	if(isShiftRecordNumber()){
		unsigned int nchanged = change_records.size();
		if(nchanged > 0){
			return nchanged + getShiftRecordNumber();
		}
		else{
			return getShiftRecordNumber();
		}
	}
	else{
		return original_number_of_records;
	}
}
unsigned int CacheRecordWork::getStartChangeRecord()
{
	if(minimum_edit_record <= serial_shift_record_number){
		return minimum_edit_record;
	}
	return serial_shift_record_number;
}
// Last original (not edited) current record.
unsigned int CacheRecordWork::getStartNewRecord()
{
	//unsigned int start = getStartChangeRecord();
	//unsigned int nrecord = getRecordNumber();
	//for(unsigned int i=start; i < nrecord ; i++ ){
	//	if(getStatus( i ) == 'O'){
	//		unsigned int original = getOriginalRecord( i );
	//		CacheEditedRecord temp(original);
	//		CacheEditedRecord* pfind = edited_records->search( &temp );
	//		if(!pfind){ // Not edited !
	//			start = i + 1;
	//		}
	//	}
	//}
	//return start;
	unsigned int start = getStartChangeRecord();
	unsigned int max_original_changed = getMaxOriginalChangedRecord();
	if(max_original_changed != CACHE_RECORD_NO_RECORD){
		start = max_original_changed + 1;
	}
	return start;
}
// Return changed record index.
unsigned int CacheRecordWork::getMaxOriginalChangedRecord()
{
	unsigned int start = getStartChangeRecord();
	if(start == CACHE_RECORD_NO_RECORD){
		start = 0;
	}
	unsigned int nrecord = getRecordNumber();
	unsigned int max_original_changed = CACHE_RECORD_NO_RECORD;
	for(unsigned int ii=start;ii < nrecord;ii++){
		if(getStatus( ii ) == 'O'){
			max_original_changed = ii;
		}
	}
	return max_original_changed;
}

unsigned char CacheRecordWork::getStatus( unsigned int record )
{
	if(isShiftRecordNumber()){
		unsigned int nchanged = change_records.size();
		if(record >= getShiftRecordNumber()){
			if(nchanged > 0){
				if(record >= (nchanged + getShiftRecordNumber())){
					return 0x0; // No record.
				}
				else{
					unsigned int array_index = record - getShiftRecordNumber();
					CacheChangeRecord* rec = change_records[ array_index ];
					switch(rec->getType()){
					case CACHE_RECORD_OPERATION_OLD:
						return 'O';
					case CACHE_RECORD_OPERATION_NEW:
						return 'I';
					}
				}
			}
		}
		else{
			return 'O'; // Original.
		}
	}
	else{
		if(record < original_number_of_records){
			return 'O';
		}
	}
	return 0x0;
}
unsigned int CacheRecordWork::getOriginalRecord( unsigned int record )
{
	if(isShiftRecordNumber()){
		unsigned int nchanged = change_records.size();
		if(record >= getShiftRecordNumber()){
			if(nchanged > 0){
				if(record >= (nchanged + getShiftRecordNumber())){
					return CACHE_RECORD_NO_RECORD; // No record.
				}
				else{
					unsigned int array_index = record - getShiftRecordNumber();
					CacheChangeRecord* rec = change_records[ array_index ];
					switch(rec->getType()){
					case CACHE_RECORD_OPERATION_OLD:
						return rec->getOriginalRecordNumber();
					case CACHE_RECORD_OPERATION_NEW:
						return CACHE_RECORD_NO_RECORD;
					}
				}
			}
		}
		else{
			return record; // Original.
		}
	}
	else{
		if(record < original_number_of_records){
			return record;
		}
	}
	return CACHE_RECORD_NO_RECORD;
}

unsigned int CacheRecordWork::getErasedOriginalRecords(vector<unsigned int>& records)
{
	records.clear();
	if (isShiftRecordNumber()) {
		vector<bool> remains;
		remains.resize(original_number_of_records, false);
		for (unsigned int i = 0; i < serial_shift_record_number; i++) {
			remains[i] = true;
		}

		for (unsigned int i = 0; i < change_records.size(); i++) {
			unsigned int rec = change_records[i]->getOriginalRecordNumber();
			if (rec < original_number_of_records) {
				remains[rec] = true;
			}
		}
		for (unsigned int i = 0; i < original_number_of_records; i++) {
			if (remains[i] == false) { // Erase.
				// SET RECORD INDEXES.
				records.push_back(i);
			}
		}
	}
	return records.size();
}

char* CacheRecordWork::getRecordLabel( unsigned int record )
{
	// Status
	unsigned char status = getStatus( record );
	if(status == 'O'){ // Original.
		unsigned int original = getOriginalRecord( record );
		CacheEditedRecord temp(original);
		CacheEditedRecord* pfind = edited_records->search( &temp );
		if(pfind){
			return pfind->getLabel();
		}
	}
	else if(status == 'I'){
		CacheChangeRecord* rec = getChangeRecord( record );
		if(rec){
			return rec->getLabel();
		}
	}
	return NULL;
}

bool CacheRecordWork::findRecordFromLabel( const char* label, unsigned int& record )
{
	unsigned int i,num = getRecordNumber();
	CacheEditedRecord* erec;
	CacheChangeRecord* crec;
	for(i=0;i<num;i++){
		unsigned char status = getStatus( i );
		switch(status){
		case 'O':
			erec = getEditRecord( i );
			if(erec){
				if(erec->sameLabel(label)){
					record = i;
					return true;
				}
			}
			break;
		case 'I':
			crec = getChangeRecord( i );
			if(crec){
				if(crec->sameLabel(label)){
					record = i;
					return true;
				}
			}
			break;
		default:
			break;
		}
	}
	return false;
}

CacheChangeRecord* CacheRecordWork::getChangeRecord( unsigned int record )
{
	if(isShiftRecordNumber()){
		unsigned int nchanged = change_records.size();
		if(record >= getShiftRecordNumber()){
			if(nchanged > 0){
				if(record >= (nchanged + getShiftRecordNumber())){
					return NULL; // No record.
				}
				else{
					unsigned int array_index = record - getShiftRecordNumber();
					CacheChangeRecord* rec = change_records[ array_index ];
					return rec;
				}
			}
		}
	}
	return NULL;
}
void CacheRecordWork::debugPrint()
{
	unsigned int num = getRecordNumber();
	unsigned int i,rec;
	for(i=0;i<num;i++){
		unsigned char status = getStatus( i );
		switch(status){
		case 'O':
			rec = getOriginalRecord( i );
			printf( "[%d]", rec );
			break;
		case 'I':
			printf( "[I]" );
			break;
		default:
			printf( "[X]" );
			break;
		}
	}
	printf( "\n" );
	for(i=0;i<num;i++){
		bool status = isEditedRecord( i );
		if(status){
			printf( " e " );
		}
		else{
			printf( " - " );
		}
	}
	printf( "\n" );
}

//
// Edit flag.
//
// Edited old/Appended.
bool CacheRecordWork::setEditRecord( unsigned int record, const char* label, const char* filepath )
{
	// Check record number.
	if(record >= getRecordNumber()){
		return false;
	}
	// CHECK MIN-MAX RECORD.
	check_minimum_edit_record( record );
	// Status
	unsigned char status = getStatus( record );
	if(status == 'O'){ // Original.
		unsigned int original = getOriginalRecord( record );
		// Check alreday.
		if(isEditedRecord( record )){
			// Change label or filepath.
			CacheEditedRecord temp(original);
			CacheEditedRecord* pfind = edited_records->search( &temp );
			if(pfind){
				if(label){
					pfind->setLabel( label );
				}
				if(filepath){
					pfind->setFile( filepath );
				}
			}
			return true;
		}
		// New.
		CacheEditedRecord* newrec = new CacheEditedRecord(original);
		if(label){
			newrec->setLabel( label );
		}
		if(filepath){
			newrec->setFile( filepath );
		}
		edited_records->put( newrec );
		return true;
	}
	else if(status == 'I'){
		CacheChangeRecord* rec = getChangeRecord( record );
		if(!rec) return false;
		if(label){
			rec->setLabel( label );
		}
		if(filepath){
			rec->setFile( filepath );
		}
		return true;
	}
	return false;
}

// Record moved.
bool CacheRecordWork::isMovedRecord( unsigned int record )
{
	unsigned char status = getStatus( record );
	if(!status) return false;

	if(status == 'O'){
		unsigned int original = getOriginalRecord( record );
		return (original != record);
	}
	else if(status == 'I'){
		return true;
	}
	return false;
}
// Is the edited temporary record file ? (Edited old/Appended).
bool CacheRecordWork::isEditedRecord( unsigned int record )
{
	unsigned char status = getStatus( record );
	if(!status) return false;

	if(status == 'O'){
		unsigned int original = getOriginalRecord( record );
		CacheEditedRecord temp(original);
		CacheEditedRecord* pfind = edited_records->search( &temp );
		if(pfind) return true;
		return false;
	}
	else if(status == 'I'){
		return true;
	}
	return false;
}
// Edited old/Appended.
char* CacheRecordWork::getEditedRecordFilePath( unsigned int record )
{
	// Status
	unsigned char status = getStatus( record );
	if(status == 'O'){ // Original.
		unsigned int original = getOriginalRecord( record );
		CacheEditedRecord temp(original);
		CacheEditedRecord* pfind = edited_records->search( &temp );
		if(pfind){
			return pfind->getFile();
		}
		return NULL;
	}
	else if(status == 'I'){
		CacheChangeRecord* rec = getChangeRecord( record );
		if(rec){
			return rec->getFile();
		}
	}
	return NULL;
}
void CacheRecordWork::setEditedRecordFilePath( unsigned int record, const char* filepath )
{
	// Status
	unsigned char status = getStatus( record );
	if(status == 'O'){ // Original.
		unsigned int original = getOriginalRecord( record );
		CacheEditedRecord temp(original);
		CacheEditedRecord* pfind = edited_records->search( &temp );
		if(pfind){
			pfind->setFile(filepath);
		}
	}
	else if(status == 'I'){
		CacheChangeRecord* rec = getChangeRecord( record );
		if(rec){
			rec->setFile(filepath);
		}
	}
}
// Only edited old record.
CacheEditedRecord* CacheRecordWork::getEditRecord( unsigned int record )
{
	unsigned char status = getStatus( record );
	if(!status) return NULL;
	if(status == 'O'){
		unsigned int original = getOriginalRecord( record );
		CacheEditedRecord temp(original);
		CacheEditedRecord* pfind = edited_records->search( &temp );
		return pfind;
	}
	else if(status == 'I'){
		return NULL;
	}
	return NULL;
}

//
// Number
//
bool CacheRecordWork::isAppendOnly()
{
	if(serial_shift_record_number == original_number_of_records){
		return (sizeEditRecord() <= 0 && change_records.size() > 0);
	}
	return false;
}
unsigned int CacheRecordWork::sizeEditRecord()
{
	return edited_records->size();
}
unsigned int CacheRecordWork::sizeChangeRecord()
{
	return change_records.size();
}

//
// class CacheRecordOperation. 
//
bool CacheRecordOperation::is_common() {
	return (_common_record != NULL);
}
void CacheRecordOperation::check_common() {
	if (!_common_record) {
		_common_record = new CacheEditedRecord(0);
	}
}
void CacheRecordOperation::delete_common() {
	if (_common_record) {
		delete _common_record; // WITH ERASING A EDITED FILE.
		_common_record = NULL;
	}
}

void CacheRecordOperation::check_object() {
	if (!_cache_record_work) {
		_cache_record_work = new CacheRecordWork(work_number_of_records);
	}
}
void CacheRecordOperation::delete_object() {
	if (_cache_record_work) {
		delete _cache_record_work;
		_cache_record_work = NULL;
	}
}
// public 
CacheRecordOperation::CacheRecordOperation(unsigned int number_of_records) {
	_cache_record_work = NULL;
	work_number_of_records = number_of_records;
	// This class.
	_common_record = NULL;
	clearDirectory();
}
CacheRecordOperation::~CacheRecordOperation() {
	delete_object();
	// Commoon.
	delete_common();
	// DIrectory.
	clearDirectory();
}
void CacheRecordOperation::clearRecordOperation(bool clear_commmon) {
	if (_cache_record_work) {
		work_number_of_records = _cache_record_work->getRecordNumber();
		_cache_record_work->clearRecordOperation();
		delete _cache_record_work;
		_cache_record_work = NULL;
	}
	if (clear_commmon) {
		// Commoon.
		delete_common();
		// Directory.
		clearDirectory();
	}
}

//
// Temporary directory.
//
bool CacheRecordOperation::directory() {
	return _set_record_temporary_directory;
}
void CacheRecordOperation::registDirectory() {
	_set_record_temporary_directory = true;
}
void CacheRecordOperation::clearDirectory() {
	// Temporary directory.
	// DELETE DIRECTORY.
	if (_record_temporary_directory.length() > 0) {
		const char* path = _record_temporary_directory.c_str();
		fcache_dir_delete(path);
	}
	// DELETE NAME.
	_set_record_temporary_directory = false;
	_record_temporary_directory = "";
}
std::string& CacheRecordOperation::getDirectory() {
	return _record_temporary_directory;
}
void CacheRecordOperation::setDirectory(std::string& dirpath) {
	_record_temporary_directory = dirpath;
}
//
// Input.
//
bool CacheRecordOperation::eraseRecord(unsigned int start, unsigned int number) {
	if (start <= 0) {
		start = 1;
	}
	unsigned int rec_id = start - 1;
	//-------------------------------------------------------
	check_object();
	return _cache_record_work->eraseRecord(rec_id, number);
}
bool CacheRecordOperation::insertRecord(unsigned int start, unsigned int number, bool front) {
	if (start <= 0) {
		start = 1;
	}
	unsigned int rec_id = start - 1;
	//-------------------------------------------------------
	check_object();
	return _cache_record_work->insertRecord(rec_id, number, front);
}
bool CacheRecordOperation::appendRecord(unsigned int number) {
	check_object();
	return _cache_record_work->appendRecord(number);
}
bool CacheRecordOperation::setRecordLabel(unsigned int record, const char* label) {
	if (record <= 0) {
		return false;
	}
	unsigned int rec_id = record - 1;
	//-------------------------------------------------------
	check_object();
	return _cache_record_work->setRecordLabel(rec_id, label);
}
//
// Output.
//
// Query.
bool CacheRecordOperation::isChanged() {
	if (is_common()) return true; // COMMON.
	if (!_cache_record_work) return false;
	return _cache_record_work->isChanged();
}
bool CacheRecordOperation::isEdited() {
	if (is_common()) return true; // COMMON.
	if (!_cache_record_work) return false;
	return _cache_record_work->isEdited();
}
bool CacheRecordOperation::isRecordEdited() { // NO CHECK COMMON.
	if (!_cache_record_work) return false;
	return _cache_record_work->isEdited();
}
bool CacheRecordOperation::isRecordChanged() { // NO CHECK COMMON.
	if (!_cache_record_work) return false;
	return _cache_record_work->isRecordChanged();
}
// Record operation (Insert/Append/Erase)
unsigned int CacheRecordOperation::getRecordNumber() {
	if (!_cache_record_work) return work_number_of_records;
	return _cache_record_work->getRecordNumber();
}
unsigned int CacheRecordOperation::getStartChangeRecord() { // FOR OUTPUT FILE.
	if (is_common()) return 0;
	if (!_cache_record_work) return (work_number_of_records + 1);
	return _cache_record_work->getStartChangeRecord() + 1;
}
unsigned int CacheRecordOperation::getStartNewRecord() { // FOR OUTPUT FILE. AFTER THIS RECORD, THESE ARE ONLY APPENDED RECORDS.
	if (!_cache_record_work) return (work_number_of_records + 1);
	return _cache_record_work->getStartNewRecord() + 1;
}
unsigned int CacheRecordOperation::getMaxOriginalChangedRecord() { // Return changed record index.
	if (!_cache_record_work) return 0;
	return _cache_record_work->getMaxOriginalChangedRecord() + 1;
}

// 0x0: No record (SKIP), 'O': OLD record, 'I': INSERTED record.
//unsigned char getStatus( unsigned int record );
bool CacheRecordOperation::isOriginal(unsigned int record) {
	if (record <= 0) {
		return true;
	}
	unsigned int rec_id = record - 1;
	//-------------------------------------------------------
	if (!_cache_record_work) {
		if (record <= work_number_of_records) return true; // REMARK!!!
		return false;
	}
	return _cache_record_work->isOriginal(rec_id);
}
// OLD record
unsigned int CacheRecordOperation::getOriginalRecord(unsigned int record) {
	if (record <= 0) {
		return 0;
	}
	unsigned int rec_id = record - 1;
	//-------------------------------------------------------
	if (!_cache_record_work) return record;
	return _cache_record_work->getOriginalRecord(rec_id) + 1;
}
unsigned int CacheRecordOperation::getErasedOriginalRecords(vector<unsigned int>& records)
{
	if (!_cache_record_work) return 0;
	return _cache_record_work->getErasedOriginalRecords(records);
}
// Get the original record number after editing records by a simple function.
unsigned int CacheRecordOperation::getOriginalRecordSimpleWay(unsigned int record)
{
	if (isChanged()) {
		// Is the edited temporary record file ?
		bool is_edited = isEditedRecord(record);
		if (is_edited) {
			// Exist the edited temporary record file.
			return record;
		}
		if (isOriginal(record)) {
			// Return the original record number.
			unsigned int original = getOriginalRecord(record);
			if (original != CACHE_RECORD_NO_RECORD) {
				return original;
			}
		}
	}
	return record;
}

// Edited old/Appended.
char* CacheRecordOperation::getRecordLabel(unsigned int record) { // FOR NEW RECORD.
	if (record <= 0) {
		return NULL;
	}
	unsigned int rec_id = record - 1;
	//-------------------------------------------------------
	if (!_cache_record_work) return NULL;
	return _cache_record_work->getRecordLabel(rec_id);
}
// Edited old/Appended.
bool CacheRecordOperation::findRecordFromLabel(const char* label, unsigned int& record) {
	if (!label) return false;
	if (!_cache_record_work) return false;
	unsigned int rec_id = 0;
	if (_cache_record_work->findRecordFromLabel(label, rec_id)) {
		record = rec_id + 1;
		return true;
	}
	return false;
}
//CacheChangeRecord* getChangeRecord( unsigned int record );
void CacheRecordOperation::debugPrint() {
	if (!_cache_record_work) return;
	_cache_record_work->debugPrint();
}
//
// Edit flag.
//
// Edited old/Appended.
bool CacheRecordOperation::setEditRecord(unsigned int record, const char* label, const char* filepath) {
	if (record <= 0) {
		check_common();
		//---> Ignore label.
		if (filepath) {
			_common_record->setFile(filepath);
		}
		return true;
	}
	unsigned int rec_id = record - 1;
	//-------------------------------------------------------
	check_object();
	return _cache_record_work->setEditRecord(rec_id, label, filepath);
}
// Record moved. Return false for record = 0, true for INSERTED or MOVED ORIGINAL. 
bool CacheRecordOperation::isMovedRecord(unsigned int record) {
	if (record <= 0) {
		return false;
	}
	unsigned int rec_id = record - 1;
	//-------------------------------------------------------
	if (!_cache_record_work) return false;
	return _cache_record_work->isMovedRecord(rec_id);
}
// Is the edited temporary record file ? (Edited old/Appended)
bool CacheRecordOperation::isEditedRecord(unsigned int record) {
	if (record <= 0) {
		return is_common();
	}
	unsigned int rec_id = record - 1;
	//-------------------------------------------------------
	if (!_cache_record_work) return false;
	return _cache_record_work->isEditedRecord(rec_id);
}
// Edited old/Appended.
char* CacheRecordOperation::getEditedRecordFilePath(unsigned int record) {
	if (record <= 0) {
		if (is_common()) {
			return _common_record->getFile();
		}
	}
	unsigned int rec_id = record - 1;
	//-------------------------------------------------------
	if (!_cache_record_work) return NULL;
	return _cache_record_work->getEditedRecordFilePath(rec_id);
}
void CacheRecordOperation::setEditedRecordFilePath(unsigned int record, const char* filepath) {
	if (record <= 0) {
		check_common();
		//---> Ignore label.
		if (filepath) {
			_common_record->setFile(filepath);
		}
		return;
	}
	unsigned int rec_id = record - 1;
	//-------------------------------------------------------
	check_object();
	_cache_record_work->setEditedRecordFilePath(rec_id, filepath);
}
void CacheRecordOperation::resetCommonEdited() {
	delete_common();
}
bool CacheRecordOperation::isCommonEdited() {
	return is_common();
}
bool CacheRecordOperation::isAppendOnly() {
	if (is_common()) return false; // COMMON WAS EDITED.
	if (!_cache_record_work) return false;
	return _cache_record_work->isAppendOnly();
}
unsigned int CacheRecordOperation::sizeEditRecord() {
	if (!_cache_record_work) return 0;
	return _cache_record_work->sizeEditRecord();
}
unsigned int CacheRecordOperation::sizeChangeRecord() {
	if (!_cache_record_work) return 0;
	return _cache_record_work->sizeChangeRecord();
}

//
// For divided records operations.
//
CacheDivideRecordData::CacheDivideRecordData( unsigned int record )
:move_or_copy_file(false),executed_move_or_copy_or_delete(false),record_number(record),
 source_file_path(NULL),record_file_name(NULL),record_label(NULL)
{
}
CacheDivideRecordData::~CacheDivideRecordData()
{
	if(source_file_path){
		delete[] source_file_path;
	}
	if(record_file_name){
		delete[] record_file_name;
	}
	if(record_label){
		delete[] record_label;
	}
}
// Move or Copy file.
void CacheDivideRecordData::setMoveFile( bool is_move )
{
	move_or_copy_file = is_move;
}
bool CacheDivideRecordData::isMoveFile()
{
	return move_or_copy_file;
}
bool CacheDivideRecordData::isExecuted()
{
	return executed_move_or_copy_or_delete;
}
void CacheDivideRecordData::setExecuted( bool executed )
{
	executed_move_or_copy_or_delete = executed;
}

// Record number, >= 1.
unsigned int CacheDivideRecordData::getRecordNo() const 
{
	return record_number;
}

void CacheDivideRecordData::setSourceFilePath( const char* file )
{
	if(source_file_path){
		delete[] source_file_path;
	}
	source_file_path = fcache_record_operation_strdup( file );
}
const char* CacheDivideRecordData::getSourceFilePath()
{
	return source_file_path;
}

void CacheDivideRecordData::setFileName( const char* file )
{
	if(record_file_name){
		delete[] record_file_name;
	}
	record_file_name = fcache_record_operation_strdup( file );
}
const char* CacheDivideRecordData::getFileName()
{
	return record_file_name;
}
void CacheDivideRecordData::setLabel( const char* label )
{
	if(record_label){
		delete[] record_label;
	}
	record_label = fcache_record_operation_strdup( label );
}
const char* CacheDivideRecordData::getLabel()
{
	return record_label;
}

// class CacheDividedRecordOperation.
CacheDividedRecordOperation::CacheDividedRecordOperation(
	CacheRecordOperation* recop,
	REC_LABELPOS_DATA*    recdt
)
:overwrite(false),divide_records(NULL),record_opp(recop),record_dtp(recdt)
{
	//num_record = 0;
	link_file_with_directory_record = CACHE_RECORD_NO_RECORD;
}
CacheDividedRecordOperation::~CacheDividedRecordOperation()
{
	if(divide_records){
		divide_records->clearDeleteElement();
		delete divide_records;
	}
	if(delete_files.size() > 0){
		unsigned int i,n = delete_files.size();
		for(i=0;i<n;i++){
			char *p = delete_files[i];
			if(p){
				delete[] p;
			}
		}
		delete_files.clear();
	}
}
// file.
void CacheDividedRecordOperation::setSourceOutputFile(
	bool overwrite,
	const char* source_filepath,
	const char* output_filepath
)
{
	this->overwrite = overwrite;
	if(!source_filepath) return;
	if(!output_filepath) return;
	this->source_filepath = source_filepath;
	this->output_filepath = output_filepath;
	// make dirctory name.
	string namestring;
	if(!fcache_make_output_directory( output_filepath, "directory", namestring )){
		return;
	}
	// split.
	string filename;
	futl_filepart(namestring.c_str(), filename);
	
	this->output_directory_path = namestring;
	this->output_directory_name = filename;
	
}
// directory.
void CacheDividedRecordOperation::setOutputDirectoryPath( const char* directory_path )
{
	if(!directory_path) return;
	this->output_directory_path = directory_path;
}

// Output path / directory-name + file.
// directory-PATH + "/" + filename.
bool CacheDividedRecordOperation::getFilePath(
	string& outpath,
	const char* filename,
	const char* delimiter // = NULL
)
{
	if(!filename) return false;
	if(!delimiter){
		delimiter = fcache_get_delimiter();
	}
	outpath = output_directory_path + delimiter;
	outpath = outpath + filename;
	return true;
}
// directory-NAME + "/" + filename
bool CacheDividedRecordOperation::getDirectoryFile(
	string& directory_file,
	const char* filename,
	const char* delimiter // = NULL
)
{
	if(!filename) return false;
	if(!delimiter){
		delimiter = fcache_get_delimiter();
	}
	directory_file = output_directory_name + delimiter;
	directory_file = directory_file + filename;
	return true;
}

//--
bool CacheDividedRecordOperation::isLinkFileWithDirectory()
{
	return (link_file_with_directory_record != CACHE_RECORD_NO_RECORD);
}
unsigned int CacheDividedRecordOperation::getStartLinkFileWithDirectory()
{
	return link_file_with_directory_record;
}
void CacheDividedRecordOperation::setRecordLinkFileWithDirectory( unsigned int record )
{
	if(record < link_file_with_directory_record){
		link_file_with_directory_record = record;
	}
}

// start devided record.
unsigned int CacheDividedRecordOperation::startDevidedRecord()
{
	if(divide_records){
		if(divide_records->size() > 0){
			CacheDivideRecordData* recp = divide_records->get( 0 );
			if(recp){
				return recp->getRecordNo();
			}
		}
	}
	return CACHE_RECORD_NO_RECORD;
}

// new divided record files.
unsigned int CacheDividedRecordOperation::sizeDividedRecord()
{
	if(divide_records){
		return divide_records->size();
	}
	return 0;
}
CacheDivideRecordData* CacheDividedRecordOperation::getDivideRecord( unsigned int record )
{
	if(divide_records){
		CacheDivideRecordData rec( record );
		CacheDivideRecordData* recp = divide_records->search( &rec );
		return recp;
	}
	return NULL;
}
// Delete files.
unsigned int CacheDividedRecordOperation::sizeDeleteFile()
{
	return delete_files.size();
}
bool CacheDividedRecordOperation::deleteFiles()
{
	if(delete_files.size() <= 0) return true;
	bool rb = true;
	unsigned int i,n = delete_files.size();
	for(i=0;i<n;i++){
		char* filepath = delete_files[i];
		if(filepath){
			if(fcache_file_delete( filepath ) < 0){
				rb = false;
			}
		}
	}
	if(rb){
		// Delete directory.
		if(sizeDividedRecord() <= 0){
			if(output_directory_path.length() > 8){
				fcache_dir_delete( output_directory_path.c_str(), false );
			}
		}
	}
	return rb;
}

// Execute.
bool CacheDividedRecordOperation::executeFileOperation( unsigned int record )
{
	if(!divide_records) return true;
	if(output_directory_path.length() <= 2){
		return false;
	}
	CacheDivideRecordData* drecp = getDivideRecord( record );
	if(drecp){
		return executeOperation( drecp );
	}
	return true;
}

bool CacheDividedRecordOperation::executeRemainingFileOperation()
{
	if(!divide_records) return true;
	if(output_directory_path.length() <= 2){
		return false;
	}
	unsigned int i,n = divide_records->size();
	for(i=0;i<n;i++){
		CacheDivideRecordData* drecp = divide_records->get( i );
		if(!drecp) continue;
		if(drecp->isExecuted()) continue;
		// Executing.
		executeOperation( drecp );
	}
	return true;
}

bool CacheDividedRecordOperation::executeOperation( CacheDivideRecordData* drecp )
{
	if(!drecp) return false;

	const char* source_file_path = drecp->getSourceFilePath();
	const char* file_name = drecp->getFileName();
	const char* directoty_path = output_directory_path.c_str();
	string fullpath;
	futl_makepath(directoty_path, file_name, fullpath);

	// Make directory.
	if(fcache_directory( directoty_path ) != 2){
		if(fcache_dir_create( directoty_path ) != 0){
			return false;
		}
	}

	if(drecp->isMoveFile()){
		int rt = fcache_file_copy( source_file_path, fullpath.c_str() );
		if(rt == 0){ // NORMAL.
			fcache_file_delete( source_file_path );
		}
		drecp->setExecuted( true );
		return (rt == 0);
	}
	else{ // Copy.
		int rt = fcache_file_copy( source_file_path, fullpath.c_str() );
		drecp->setExecuted( true );
		return (rt == 0);
	}
	return true;
}

// Make source file path.
bool CacheDividedRecordOperation::makeSourceFilePath(
	const char* record_file,
	string& file_path
)
{
	string path = record_file;
	const char* curr_file = source_filepath.c_str();
	if(path.find("/") == string::npos){
		fcache_make_output_directory( curr_file, "directory", path );
	}
	else{
		path = curr_file;
	}
	return futl_makepath(path.c_str(), record_file, file_path);
}

// Construct divided & delete record data.
void CacheDividedRecordOperation::makeDivideRecordData()
{
	if(!record_opp) return;
	if(!record_dtp) return;
	if(!record_dtp->isDividedRecord()) return;
	//if(!record_opp->isChanged()){
	//	return;
	//}
	bool record_insert_delete = record_opp->isRecordChanged();
	bool create_no_change_record_data = false;
	bool delete_old_record_file = false;
	if(overwrite){
		delete_old_record_file = true;
	}
	else{ // Save as an another file.
		create_no_change_record_data = true;
	}

#if defined(USING_NONDIVIDED_RECORD)
	if(record_dtp->existNondividedRecord()){
		// Mix divided & non-divided record.
	}
	else{
		// All records are divided.
	}
#endif

	// Start record at 1.
	string filename;
	unsigned int i;
	unsigned int total_record = record_opp->getRecordNumber();
	unsigned int start_changed_record = record_opp->getStartChangeRecord();
	unsigned int orig_record = record_dtp->sizeRecord();
	//num_record = total_record;

	const char* label = NULL;
	char* source_file = NULL;
	char* target_file = NULL;

	// Original only. -----------------------------------------------
	if(create_no_change_record_data){ // Copy no-change record files.
		string work_original_path;
		for(i=1;i<start_changed_record;i++){
			const char* original_record_file = record_dtp->getRecFile( i ); // File name only, not path!
			if(original_record_file){
				label = record_dtp->getLabel( i );
				//+++ Same old file name.
				// if(!fcache_make_record_divide_filename( filename, this->output_directory_path, i, true)){
				filename = original_record_file;
				if(filename.find("/") != string::npos){
					setRecordLinkFileWithDirectory( i );
				}
				if(!futl_filepart( original_record_file, filename )){
					printf("Can not make a record file name : record = %d.\n", i);
					continue;
				}
				//+++
				bool is_move = false; // copy files.
				makeSourceFilePath(original_record_file, work_original_path);
				original_record_file = work_original_path.c_str();
				addDivideFile( i, original_record_file, filename.c_str(), label, is_move );
			}
		}
	}

	// Edited/Insert/Append record.
	for(i=start_changed_record;i<=total_record;i++){
		label = NULL;
		source_file = NULL;
		target_file = NULL;
		if(record_opp->isEditedRecord( i )){ // Original & edited, or inserted or appended.
			label = record_opp->getRecordLabel( i );
			source_file = record_opp->getEditedRecordFilePath( i ); // Temporary file.
			if(source_file){
				if(record_opp->isOriginal( i )){
					unsigned int original = record_opp->getOriginalRecord( i );
					const char* original_record_file = record_dtp->getRecFile( original );
					if(original_record_file){
						// Delete file.
						if(delete_old_record_file){
							string work_original_path;
							makeSourceFilePath(original_record_file, work_original_path);
							original_record_file = work_original_path.c_str();
							addDeleteFilePath( original_record_file );
						}
					}
					if(!label){
						label = record_dtp->getLabel( original );
					}
				}
				//+++ New file name.
				if(!fcache_make_record_divide_filename( filename, this->output_directory_path, i, true)){
					printf("Can not make a record file name : record = %d.\n", i);
					continue;
				}
				//+++
				bool is_move = true;
				addDivideFile( i, source_file, filename.c_str(), label, is_move );
			}
		}
		if(!source_file){
			// Original & no edit.
			if(record_opp->isOriginal( i )){
				unsigned int original = record_opp->getOriginalRecord( i );
				const char* original_record_file = record_dtp->getRecFile( original );
				if(original_record_file){ // these are divided record files.
					label = record_dtp->getLabel( original );
					if(i != original){
						//+++ New file name.
						if(!fcache_make_record_divide_filename( filename, this->output_directory_path, i, true)){
							printf("Can not make a record file name : record = %d.\n", i);
							continue;
						}
						//+++
						bool is_move = overwrite;
						string work_original_path;
						makeSourceFilePath(original_record_file, work_original_path);
						original_record_file = work_original_path.c_str();
						addDivideFile( i, original_record_file, filename.c_str(), label, is_move );
					}
					else{ // Save as an another file.
						if(create_no_change_record_data){
							//+++ New file name.
							if(!fcache_make_record_divide_filename( filename, this->output_directory_path, i, true)){
								printf("Can not make a record file name : record = %d.\n", i);
								continue;
							}
							//+++
							bool is_move = false; // copy files.
							string work_original_path;
							makeSourceFilePath(original_record_file, work_original_path);
							original_record_file = work_original_path.c_str();
							addDivideFile( i, original_record_file, filename.c_str(), label, is_move );
						}
					}
				}
			}
		}
	}

	// delete record file.
	if(delete_old_record_file){
		vector<bool> original_record_exist;
		original_record_exist.resize( orig_record + 1, false );
		for(i=0;i<start_changed_record;i++){
			original_record_exist[i] = true;
		}
		for(i = start_changed_record ; i<=total_record;i++){
			if(record_opp->isOriginal( i )){
				unsigned int original = record_opp->getOriginalRecord( i );
				original_record_exist[ original ] = true;
			}
		}
		for(i = start_changed_record ; i <= orig_record ; i++ ){
			string work_original_path;
			if(original_record_exist[ i ] == false){
				const char* original_record_file = record_dtp->getRecFile( i );
				if(original_record_file){
					makeSourceFilePath(original_record_file, work_original_path);
					original_record_file = work_original_path.c_str();
					addDeleteFilePath( original_record_file );
				}
			}
		}
	}
}

void CacheDividedRecordOperation::addDivideFile(
	unsigned int rec,
	const char* source_file_path,
	const char* destin_file_name,
	const char* label,
	bool is_move
)
{
	if(rec <= 0) return;
	if(!source_file_path) return;
	if(!destin_file_name) return;
	if(!label){
		label = "";
	}
	CacheDivideRecordData* drecp = new CacheDivideRecordData( rec );
	drecp->setSourceFilePath( source_file_path );
	drecp->setFileName( destin_file_name ); // file name.
	drecp->setLabel( label );
	drecp->setMoveFile( is_move );
	if(!divide_records){
		divide_records = new msorted_vector();
	}
	divide_records->put( drecp );
}

void CacheDividedRecordOperation::addDeleteFilePath( const char* filepath )
{
	if(!filepath) return;
	if(!*filepath) return;
	char *p = fcache_record_operation_strdup( filepath );
	delete_files.push_back( p );
}
