
/*
 *  Material Modeling Platform Package
 *  Copyright(c) 2002, The Japan Research Institute, Ltd.
 *  All rights reserved.
 *  
 *  $Id: cache_util_character.cpp,v 1.7 2008/12/17 06:57:38 t352149 Exp $
 */

#ifdef WIN32
#pragma warning (disable:4786)
#endif
#include <stdio.h>
#include <stdlib.h>
#ifndef MACOSX
#include <malloc.h>
#endif
#include <ctype.h>
#include <string.h>
#include <limits.h>
#include "cache_util_character.h"
using namespace std;

#define UDB_MAX_BUFF_SIZE	8192
static const char	UDB_UNIX_NL[] = { 0x0A, '\0'};  // LF
static const char	UDB_WDOS_NL[] = { 0x0D, 0x0A, '\0'}; // CR{LF
static const char	UDB_MACI_NL[] = { 0x0D, '\0'};  // CR
static const char   UDB_LINEFEED = 0x0A;
static const char   UDB_CARRIAGE = 0x0D;
static const char	UDB_ESCAPE = '\\';
static const char	UDB_TERMINAL = '\0';
static const char	UDB_NEW_LINE_TXT[] = "\\n";
static const char	UDB_DBL_QUOT_TXT[] = "\\\"";
static const char	UDB_SGL_QUOT_TXT[] = "\\\'";
static const char	UDB_ARYBR[] = {'[',']','\0'};
static const char	UDB_DELIMITER[] = {'\\','/','\0'};
void udb_test_print(){
	printf("[%s]\n",UDB_UNIX_NL);
	printf("[%s]\n",UDB_WDOS_NL);
	printf("[%s]\n",UDB_MACI_NL);
}

int futl_pos_character( const char *text, const char cc )
{
	int ipos = 0;
	const char* pp = text;
	while(*pp != UDB_TERMINAL){
		if(*pp == cc){
			return ipos;
		}
		pp++;
		ipos++;
	}
	return -1;
}

// Escape characters are removed.
// '\n' -> LF, '\"' -> ", '\'' -> '
int futl_convert_escape(
	const char* input,
	char** alloced_txt
)
{
	short escape = 0;
	const char* pp = input;
	char *new_txt = new char[strlen(input)+1];
	int nn = 0;
	while(*pp != UDB_TERMINAL){
		switch(*pp){
		case UDB_ESCAPE:
			if(escape <= 0) escape = 1;
			else{
				new_txt[nn++] = UDB_ESCAPE;
				escape = 0;
			}
			break;
		case 'n':
			if(escape > 0){
				new_txt[nn++] = UDB_UNIX_NL[0];
			}
			else{
				new_txt[nn++] = *pp;
			}
			escape = 0;
			break;
		case '\"':
		case '\'':
			new_txt[nn++] = *pp;
			escape = 0;
			break;
		default:
			new_txt[nn++] = *pp;
			escape = 0;
		}
		pp++;
	}
	new_txt[nn++] = UDB_TERMINAL;
	*alloced_txt = new_txt;
	return nn;
}
// Escape characters are attached.
// '\n' <- LF, '\"' <- ", '\'' <- ', '\\' <- '\'
int futl_unconvert_escape(
	const char* input,
	char** alloced_txt
)
{
	const char* pp = input;
	int nn = 0;
	//// Text length.
	while(*pp != UDB_TERMINAL){
		switch(*pp){
		case UDB_CARRIAGE:
		case UDB_LINEFEED:
		case UDB_ESCAPE:
		case '\"':
		case '\'':
			nn = nn + 2;
			break;
		default:
			nn++;
		}
		pp++;
	}
	//////////////////////////////////////////
	pp = input;
	short cr_lf = 0;
	char *new_txt = new char[nn+1];
	nn = 0;
	while(*pp != UDB_TERMINAL){
		switch(*pp){
		case UDB_CARRIAGE:
			if((*(pp+1)) == UDB_LINEFEED) cr_lf = 1;
			else{
				cr_lf = 0;
				new_txt[nn++] = UDB_ESCAPE;
				new_txt[nn++] = 'n';
			}
			break;
		case UDB_LINEFEED:
			new_txt[nn++] = UDB_ESCAPE;
			new_txt[nn++] = 'n';
			cr_lf = 0;
			break;
		case UDB_ESCAPE:
		case '\"':
		case '\'':
			new_txt[nn++] = UDB_ESCAPE;
			new_txt[nn++] = *pp;
			cr_lf = 0;
			break;
		default:
			new_txt[nn++] = *pp;
			cr_lf = 0;
		}
		pp++;
	}
	new_txt[nn++] = UDB_TERMINAL;
	*alloced_txt = new_txt;
	return nn;
}
// string versions.
int futl_convert_escape(
	const string& input,
	string& output
)
{
	char* alloced_txt = 0x0;
	int rc = futl_convert_escape(input.c_str(),&alloced_txt);
	if(rc > 0) output = alloced_txt;
	if(alloced_txt) delete alloced_txt;
	return rc;
}
int futl_unconvert_escape(
	const string& input,
	string& output
)
{
	char* alloced_txt = 0x0;
	int rc = futl_unconvert_escape(input.c_str(),&alloced_txt);
	if(rc > 0) output = alloced_txt;
	if(alloced_txt) delete alloced_txt;
	return rc;
}

// Directory delimiter.
char futl_get_path_delimiter(char os)
{
	switch(os){
	case 'u': return UDB_DELIMITER[1];
	case 'w': return UDB_DELIMITER[0];
	default:
#if defined(WIN32)
		return UDB_DELIMITER[0];
#else
		return UDB_DELIMITER[1];
#endif
	}
}
// Split string
int futl_split_string(
	const char split,
	const char* text,
	string& head,
	string& tail,
	bool ignore_space,
	bool backward,
	bool tail_default, // If not found, text puts on tail.
	bool exclude_split
)
{
	if(!text) return 0;
	int length = 0;
	int find = -1;
	const char* ss;
	char *str1 = 0x0,*str2 = 0x0;
	if(backward){
		length = strlen(text);
		for(int i=length-1;i>=0;i--){
			if(text[i] == split){
				find = i;
				break;
			}
		}
	}
	else{
		for(ss = text;*ss != '\0';ss++){
			if(find < 0 && *ss == split){
				find = length;
			}
			length++;
		}
	}
	if(find < 0){
		int nn = 0;
		str1 = (char*)malloc(length+1);
		for(ss = text;*ss != '\0';ss++){
			if(ignore_space && futl_char_is_space(*ss)){
				continue;
			}
			str1[nn++] = *ss;
		}
		str1[nn] = 0x0;
		if(tail_default){
			head = "";
			tail = str1;
		}
		else{
			head = str1;
			tail = "";
		}
		if(str1) free(str1);
		return nn;
	}
	else{
		int nn1=0,nn2=0;
		str1 = (char*)malloc(find+1);
		str2 = (char*)malloc(length-find+1);
		length = 0;
		for(ss = text;*ss != '\0';ss++,length++){
			if(ignore_space && futl_char_is_space(*ss)){
				continue;
			}
			if(length < find) str1[nn1++] = *ss;
			else{
				if(exclude_split){
					// Split character is excluded.
					if(length == find) continue;
				}
				str2[nn2++] = *ss;
			}
		}
		str1[nn1] = 0x0;
		str2[nn2] = 0x0;
		head = str1;
		tail = str2;
		if(str1) free(str1);
		if(str2) free(str2);
		return nn1+nn2;
	}
}
// Remove spaces. //
bool futl_char_is_space( char cc )
{
	switch(cc){
	case ' ':
	case '\t':
		return true;
	}
	return false;
}
int futl_ignore_space_strcpy(char* s1, const char* s2)
{
	int nn = 0;
	char* sp1 = s1;
	const char* sp2 = s2;
	while(*sp2 != '\0'){
		if(futl_char_is_space(*sp2)){
			sp2++;
		}
		else{
			*sp1 = *sp2;
			sp1++;
			sp2++;
			nn++;
		}
	}
	*sp1 = '\0';
	return nn;
}
int futl_ignore_space_strcat(char* s1, const char* s2)
{
	int nn = 0;
	char* sp1 = s1;
	const char* sp2 = s2;
	while(*sp1 != '\0'){
		sp1++;
		nn++;
	}
	while(*sp2 != '\0'){
		if(futl_char_is_space(*sp2)){
			sp2++;
		}
		else{
			*sp1 = *sp2;
			sp1++;
			sp2++;
			nn++;
		}
	}
	*sp1 = '\0';
	return nn;
}
char* futl_ignore_space_strdup(const char* s2)
{
	if(!s2) return 0x0;
	char* ss = (char*)malloc(strlen(s2)+1);
	char* sp1 = ss;
	const char* sp2 = s2;
	while(*sp2 != '\0'){
		if(futl_char_is_space(*sp2)){
			sp2++;
		}
		else{
			*sp1 = *sp2;
			sp1++;
			sp2++;
		}
	}
	*sp1 = '\0';
	return ss;
}
char* futl_trim_space_strdup(const char* s2)
{
	if(!s2) return 0x0;
	int nn = strlen(s2);
	char* ss = (char*)malloc(nn+1);
	if(nn <= 0){
		*ss = '\0';
		return ss;
	}
	// nn > 0:
	char* sp1 = ss;
	const char* sp2 = s2;
	const char* sp3 = &(s2[nn - 1]);
	// triming front spaces.
	while(*sp2 != '\0'){
		if(futl_char_is_space(*sp2)){
			sp2++;
		}
		else{
			break;
		}
	}
	// triming tail spaces.
	while(sp3 > sp2){
		if(futl_char_is_space(*sp3)){
			--sp3;
		}
		else{
			break;
		}
	}
	// result.
	if(sp3 == sp2){
		if(!futl_char_is_space(*sp2)){
			*sp1 = *sp2;
			sp1++;
		}
		*sp1 = '\0';
		return ss;
	}
	for(const char* p = sp2; p <= sp3 ; p++ ){
		*sp1 = *p;
		sp1++;
	}
	*sp1 = '\0';
	return ss;
}
int futl_remove_space(const string& input,string& output)
{
	char* ss = futl_ignore_space_strdup(input.c_str());
	if(!ss) return 0;
	output = ss;
	free(ss);
	return output.size();
}
int futl_remove_space(string& inout)
{
	char* ss = futl_ignore_space_strdup(inout.c_str());
	if(!ss) return 0;
	inout = ss;
	free(ss);
	return inout.size();
}
int futl_trim_space( string& inout )
{
	if(inout.length() <= 0) return 0; // no change.
	char* p = futl_trim_space_strdup(inout.c_str());
	if(p){
		inout = p;
		free(p);
	}
	return (int)inout.length();
}
unsigned int futl_find_unspace(
	const string& words
)
{
	unsigned int n = words.length();
	if(n <= 0) return 0;
	unsigned int i = 0;
	for(i=0;i<n;i++){
		char c = words[i];
		if(futl_char_is_space( c )) continue;
		break;
	}
	//if(i >= n) i--;
	return i;
}
// Compare strings with options.
const char* futl_string_character(
	const char* ss,
	const char* cc
)
{
	if(!ss || !cc) return 0x0;
	const char* sp;
	const char* cp;
	for(sp = ss ; *sp ; sp++){
		for(cp = cc ; *cp != '\0' ; cp++){
			if(*sp == *cp) return sp;
		}
	}
	return 0x0;
}
const char* futl_string_character(
	const char* ss,
	const char cc
)
{
	if(!ss) return 0x0;
	const char* sp;
	for(sp = ss ; *sp ; sp++){
		if(*sp == cc) return sp;
	}
	return 0x0;
}
int futl_string_compare(
	const char* s1,
	const char* s2,
	const char* ignore_char,
	bool ignore_space
)
{
	if(!s1 && !s2) return 0;
	else if(s1 && !s2) return 1;
	else if(!s1 && s2) return -1;
	bool next_loop;
	int icomp = 0;
	const char* sp1 = s1;
	const char* sp2 = s2;
	while(*sp1 != '\0' && *sp2 != '\0'){
		next_loop = false;
		if(ignore_char){
			if(futl_string_character(ignore_char,*sp1)){
				if(futl_string_character(ignore_char,*sp2)){
					sp1++; sp2++;
					next_loop = true;
				}
			}
			else if(futl_string_character(ignore_char,*sp2)){
				if(futl_string_character(ignore_char,*sp1)){
					sp1++; sp2++;
					next_loop = true;
				}
			}
		}
		if(ignore_space){
			if(futl_char_is_space(*sp1)){
				sp1++;
				next_loop = true;
			}
			if(futl_char_is_space(*sp2)){
				sp2++;
				next_loop = true;
			}
		}
		if(next_loop) continue;
		// Comparing character.
		if((icomp = (int)((*sp1) - (*sp2))) != 0) return icomp;
		sp1++;
		sp2++;
	}
	return (int)((*sp1) - (*sp2));
}
/*********************/
/* Duplicating string with triming. */
char	*futl_trimstrdup( char *str )
{
	int	i,startp,endp,lenp;
	char	*out = NULL;
	if(str == NULL) return NULL;
	lenp = strlen(str);
	startp = 0;
	endp = lenp;
	for(i=0;str[i] != '\0' ; i++){
		switch(str[i]){
		case ' ':
		case '\t':
			continue;
		}
		startp = i;
		goto OUT_LOOP1;
	}
	if(lenp > 0){
		endp = 0;
		goto OUT_LOOP2;
	}
OUT_LOOP1:
	for(i=lenp-1;i >= 0 ; i-- ){
		switch(str[i]){
		case ' ':
		case '\t':
			continue;
		}
		endp = i + 1;
		goto OUT_LOOP2;
	}
OUT_LOOP2:
	lenp = endp - startp;
	if(lenp <= 0) lenp = 0;
	out = (char *)malloc(lenp+1);
	if(out == NULL) return NULL;
	strncpy(out,&(str[startp]),lenp);
	out[lenp] = '\0';
	return out;
}
// Untrim: head+input+tail -> output(ignoring spaces).
int futl_untrimstring(
	const char* head,
	const char* tail,
	const string& input,
	string& output,
	bool if_not_found
)
{
	bool head_op = false;
	bool tail_op = false;
	unsigned short head_size = 0;
	unsigned short tail_size = 0;
	int insize = input.size();
	const char* inputp = input.c_str();
	/////////////////////////////////
	if(if_not_found){
		if(head){
			head_size = strlen(head);
			if(strncmp(inputp,head,head_size)) head_op = true;
		}
		if(tail){
			tail_size = strlen(tail);
			if(insize < tail_size) tail_op = true;
			else{
				const char* tailp = &(inputp[insize-tail_size]);
				if(strcmp(tailp,tail)) tail_op = true;
			}
		}
	}
	else{
		if(head) head_op = true;
		if(tail) tail_op = true;
	}
	//// If the input string is empty. ////
	if(insize <= 0){
		output = "";
		if(head) output += head;
		if(tail) output += tail;
		return head_size+tail_size;
	}
	///////////////////////////////////////
	int nn = 0;
	char* ss = (char*)malloc(insize+head_size+tail_size+1);
	char* sp1 = ss;
	const char* sp2;
	//// Header character. ////
	if(head_op){
		for(sp2 = head;*sp2 != '\0';sp2++){
			*sp1++ = *sp2;
		}
		nn += head_size;
	}
	//// Main character. ////
	for(sp2 = inputp;*sp2 != '\0';sp2++){
		if(futl_char_is_space(*sp2)) continue;
		*sp1++ = *sp2;
		nn++;
	}
	//// Tail character. ////
	if(tail_op){
		for(sp2 = tail;*sp2 != '\0';sp2++){
			*sp1++ = *sp2;
		}
		nn += tail_size;
	}
	//// End ////
	*sp1 = '\0';
	output = ss;
	if(ss) free(ss);
	return nn;
}

// remove bracket"[]".
static int remove_bracket( const string& ss, string& out, bool& is_bracket )
{
	is_bracket = false;
	if(ss.size() <= 0) return 0;
	const char* pp = ss.c_str();
	if(pp[0] == '['){
		is_bracket = true;
		out = &(pp[1]);
		if(out.size() > 0){
			if(out[out.size()-1] == ']')
				out.erase(out.size()-1,1);
		}
	}
	else out = pp;
	return out.size();
}
// remove bracket"[]".
string futl_remove_bracket(
	const string& ss,       // IN
	bool nospace_operation, // IN(ignore space option)
	bool *is_bracket        // OUT(option)
)
{
	bool bracket = false;
	if(ss.size() <= 0) return ss;
	string output;
	if(nospace_operation){
		string nospace;
		if(futl_remove_space(ss,nospace) <= 0) return nospace;
		remove_bracket(nospace,output,bracket);
		if(is_bracket) *is_bracket = bracket;
		return output;
	}
	remove_bracket(ss,output,bracket);
	if(is_bracket) *is_bracket = bracket;
	return output;
}
// string literal is extracted.
int futl_extract_literal( const char* text, string& literal )
{
	short quote = 0; // 1:single,2:double
	char* ss = (char*)malloc(strlen(text)+1);
	char* pp = ss;
	const char* sp = text;
	while(*sp != '\0'){
		switch(*sp){
		case '\'':
			if(quote == 0) quote = 1;
			else if(quote == 1) goto End_Loop;
			else *pp++ = *sp;
			break;
		case '\"':
			if(quote == 0) quote = 2;
			else if(quote == 2) goto End_Loop;
			else *pp++ = *sp;
			break;
		default:
			if(quote > 0) *pp++ = *sp;
		}
		sp++;
	}
End_Loop:
	*pp = '\0';
	literal = ss;
	if(ss) free(ss);
	return literal.size();
}
// ignore character.
int futl_ignore_character(
	const char* ignore,
	const char* text,
	string& literal
)
{
	bool found;
	char* ss = (char*)malloc(strlen(text)+1);
	char* pp = ss;
	const char* sp = text;
	const char* gg;
	while(*sp != '\0'){
		found = false;
		gg = ignore;
		while(*gg != '\0'){
			if(*sp == *gg){
				found = true;
				break;
			}
			gg++;
		}
		if(!found) *pp++ = *sp;
		sp++;
	}
	*pp = '\0';
	literal = ss;
	if(ss) free(ss);
	return literal.size();
}
// The same letter as the single character in the 'characters' 
// is found from the 'text'.
const char* futl_find_character(
	const char* text,
	const char* characters
)
{
	const char* sp = text;
	const char* cp = characters;
	while(*sp != '\0'){
		cp = characters;
		while(*cp != '\0'){
			if(*sp == *cp) return sp;
			cp++;
		}
		sp++;
	}
	return 0x0;
}

/* Is text a (int/float) number? */
int	futl_isnumber(const char* text)
{
	int	other = 0;
	int	ret = 1;
	const char *pp;
	pp = text;
	for( ; *pp != '\0'; pp++){
		switch(*pp){
		case ' ':
		case '\t':
		case '+':
		case '-':
			continue;
		default:
			goto Next;
		}
	}
	return 0;
Next:
	for( ; *pp != '\0'; pp++){
		if(isdigit(*pp)) continue;
		switch(*pp){
		case '.':
			other++;
			break;
		default:
			return 0;
		}
	}
	if(other >= 2) ret = 0;
	return ret;
}

/* Extended alphabet */
int	futl_isexalpha(char cc)
{
	if(isalpha(cc)) return 1;
	if(isdigit(cc)) return 1;
	switch(cc){
	case '$':
	case '@':
	case '_':
		return 1;
	}
	return 0;
}
int futl_isexalpha(char* ss)
{
	if(!ss) return 0;
	while(*ss != '\0'){
		if(futl_isexalpha(*ss)) return 1;
		ss++;
	}
	return 0;
}

static bool isnot_number(const char cc)
{
	if(isalpha(cc)) return true;
	switch(cc){
	case '$':
	case '@':
	case '_':
		return true;
	}
	return false;
}
int futl_isnot_number(const char* ss)
{
	if(!ss) return 0;
	short prev = 0;
	while(*ss != '\0'){
		if(prev){
			switch(*ss){
			case 'e':
			case 'E':
			case 'f':
				break;
			default:
				if(isnot_number(*ss)) return 1;
			}
		}
		else if(isnot_number(*ss)) return 1;
		// Special operation. (1.2e-5...)
		if(isdigit(*ss)) prev = 1;
		else if(*ss == '.') prev = 1;
		else prev = 0;
		ss++;
	}
	return 0;
}

// STRING OPERATIONS
void futl_string_replace( string& text, const char* str1, const char* str2 )
{
	if(strcmp(str1,str2) == 0) return;
	int str1len = strlen(str1);
	string::size_type spos = 0;
	while((spos = text.find(str1)) != text.npos){
		text.replace(spos,str1len,str2);
	}
}

int futl_string_splitfields(
	vector<string>& fields,
	const string& text,
	const char* delimit
)
{
	string::size_type spos = 0;
	string::size_type epos = 0;
	int slen = 0;
	int dlen = strlen(delimit);
	if(dlen <= 0) return 0;
	while((epos = text.find(delimit,spos)) != string::npos){
		slen = epos - spos;
		if(slen <= 0) fields.push_back("");
		else          fields.push_back(text.substr(spos,slen));
		spos = epos + dlen;
	}
	// The last token.
	epos = text.length();
	slen = epos - spos;
	if(slen > 0) fields.push_back(text.substr(spos,slen));
	return fields.size();
}

/* isbracket */
#define mc_isbracket(cc)	(((cc)==('['))?(1):(((cc)==(']'))?(2):(0)))
int	futl_isbracket( char cc )
{
	switch(cc){
	case '[':
		return 1;
	case ']':
		return 2;
	}
	return 0;
}
int futl_top_bracket(const char* ss)
{
	if(!ss) return 0;
	while(*ss != '\0'){
		switch(*ss){
		case ' ':
		case '\t':
			break;
		default:
			if(futl_isbracket(*ss)) return 1;
			return 0;
		}
		ss++;
	}
	return 0;
}
int futl_check_bracket_couple(const char* ss)
{
	if(!ss) return 0;
	int bra;
	int count = 0;
	int max_cnt = 0;
	while(*ss != '\0'){
		switch(*ss){
		case ' ':
		case '\t':
			break;
		default:
			bra = futl_isbracket(*ss);
			if(bra == 1){
				count++;
			}
			else if(bra == 2){
				count--;
			}
			if(count < 0) return -1;
			else if(count > max_cnt) max_cnt = count;
		}
		ss++;
	}
	if(count == 0) return max_cnt;
	else           return -1;
}

// name = "a.b[].c.d[].e" & level = 1,
// then return "a.b[]"
bool futl_array_name_level( int level, string& name )
{
	int nameLen = name.length();
	int nameCnt = 0;
	int br_count = 0;
	for(const char* pp = name.c_str(); *pp ; pp++ ){
		switch(*pp){
		case '.':
			if(br_count == level){
				goto LoopEnd;
			}
			break;
		case ']':
			br_count++;
			break;
		default:
			break;
		}
		nameCnt++;
	}
LoopEnd:
	if(nameCnt >= nameLen) return true;
	name.erase(nameCnt,nameLen-nameCnt);
	return true;
}

/*********************/
/*
 * aa[10].bb[100][100] -> aa[].bb[][]
 */
int     futl_ignore_array_index(
	char    *in_name,
	char    *out_name
)
{
	char	*ss,*tt;
	int		brsw = 0;
	ss = in_name;
	tt = out_name;
	while((*ss) != UDB_ARYBR[2]){
		switch(*ss){
		case '[':
			if(!brsw) *(tt++)=(*ss);
			brsw ++;
			break;
		case ']':
			brsw --;
			if(brsw < 0) brsw = 0;
			if(!brsw) *(tt++)=(*ss);
			break;
		default:
			if(brsw <= 0){
				*(tt++)=(*ss);
			}
		}
		ss++;
	}
	*tt = UDB_ARYBR[2];
	if(brsw)	return 1;
	else		return 0;
}
int     futl_ignore_array_index(
	const char* data_path,
	string& out_name
)
{
	if(!data_path) return 0;
	char* ignorename = (char*)malloc(strlen(data_path)+1);
	ignorename[0] = 0x0;
	int rc = futl_ignore_array_index((char*)data_path,ignorename);
	out_name = ignorename;
	free(ignorename);
	return rc;
}
int     futl_ignore_array_index(
	const string& in_name,
	string& out_name
)
{
	if(in_name.size() <= 0) return 0;
	const char*	data_path = in_name.c_str();
	char* ignorename = (char*)malloc(in_name.size()+1);
	ignorename[0] = 0x0;
	int rc = futl_ignore_array_index((char*)data_path,ignorename);
	out_name = ignorename;
	free(ignorename);
	return rc;
}

/*
 * Is the definition type name an array type? 
 */
int	futl_is_last_array_type(const char *name)
{
	int	strnn;
	strnn = strlen(name);
	if(strnn <= 2){
		return 0;
	}
	strnn--;
	if(name[strnn] == UDB_ARYBR[1] || name[strnn-1] == UDB_ARYBR[1]){
		return 1;
	}
	return 0;
}


/* Counting array bracket set([]) at the last variable.(b[].a[][] -> 2) */
int	futl_count_last_array_bracket(const char *name)
{
	if(!name) return 0;
	int	counter = 0;
	int	bracket = 0;
	for(const char* sp = name;*sp;sp++){
		switch(*sp){
		case '[':
			bracket++;
			break;
		case ']':
			bracket--;
			if(!bracket) counter++;
			break;
		case ' ':
			break;
		default:
			if(!bracket) counter = 0;
		}
	}
	if(bracket) return -1;
	return counter;
}

/* Counting array bracket set([]) in variable.(b[].a[][] -> 3) */
int	futl_count_all_array_bracket(const char *name)
{
	if(!name) return 0;
	int	counter = 0;
	int	bracket = 0;
	for(const char* sp = name;*sp;sp++){
		switch(*sp){
		case '[':
			bracket++;
			break;
		case ']':
			bracket--;
			if(!bracket) counter++;
			break;
		case ' ':
			break;
		}
	}
	if(bracket) return -1;
	return counter;
}

int	futl_count_all_and_last_array_bracket(const char *name, int& lastArray)
{
	if(!name) return 0;
	lastArray = 0;
	int	counter = 0;
	int	bracket = 0;
	for(const char* sp = name;*sp;sp++){
		switch(*sp){
		case '[':
			bracket++;
			break;
		case ']':
			bracket--;
			if(!bracket){
				lastArray++;
				counter++;
			}
			break;
		case ' ':
			break;
		default:
			if(!bracket) lastArray = 0;
		}
	}
	if(bracket) return -1;
	return counter;
}

// "...abc.cde" -> return 3 & "abc.cde".
int futl_dot_and_name(const char *name, string& out)
{
	if(!name) return -1;
	int count = 0;
	const char* p = name;
	for( ; *p ; p++ ){
		switch(*p){
		case '.':
			count++;
			continue;
		case ' ':
			continue;
		}
		break;
	}
	if(*p) out = p;
	else   out = "";
	return count;
}

/* Counting struct level. */
/* aa -> 0            */
/* a.b[].c.d[][] -> 3 */
int	futl_get_struct_level(const char *name)
{
	int	i = 0;
	const char* p = name;
	while(*p != '\0'){
		if(*p == '.') i++;
		p++;
	}
	return i;
}

/* Counting array bracket set([]) in each struct part. */
/* a.b[].c.d[][] -> part[0]=0,part[1]=1,part[2]=0,part[3]=2 */
int	futl_count_struct_array_bracket(
	char	*name,
	int	*part
)
{
	int	i;
	int	strnn;
	int	brsw = 0;
	//int	brnum=0;
	int	stnum=0;
	part[stnum] = 0;
	strnn = strlen(name);
	for(i=1;i<strnn;i++){
		switch(name[i]){
		case '[':
			brsw = 1;
			break;
		case ']':
			if(brsw > 0){
				brsw = 0;
				part[stnum]++;
			}
			break;
		case '.':
			stnum++;
			part[stnum] = 0;
			break;
		}
	}
	stnum++;
	return stnum;
}

int	futl_max_struct_array_bracket( const char* name )
{
	int	i;
	int	strnn;
	int	brsw = 0;
	int	partn = 0;
	int	partmax = 0;
	if(name == NULL) return 0;
	strnn = strlen(name);
	for(i=1;i<strnn;i++){
		switch(name[i]){
		case '[':
			brsw = 1;
			break;
		case ']':
			if(brsw > 0){
				brsw = 0;
				partn++;
			}
			break;
		case '.':
			if(partn > partmax) partmax = partn;
			partn = 0;
			break;
		}
	}
	if(partn > partmax) partmax = partn;
	return partmax;
}

/* Counting array bracket up to the given strruct. */
/* a.b[].c[][].d.e[].f[][][].g[].h[] & upto=6 -> 4 */
int	futl_count_upto_array_bracket(
	char	*name,
	int	upto
)
{
	int	i, num, *starray = 0x0;
	int	brcount = 0;
	int	anum = futl_count_all_array_bracket(name);
	if(anum > 0){
		starray = (int*)calloc(anum,sizeof(int));
		num = futl_count_struct_array_bracket(name,starray);
		if(upto <= 0) upto = num;
		for(i=0;i<upto;i++){
			if(starray[i] > 0){
				brcount++;
			}
		}
		if(starray) free(starray);
	}
	return brcount;
}

/* Counting same struct part. */
/*
 * Getting the structure part name.
 */
/* aaaa.bbbb[].cccc[][].ddddd.eeeee.xxxxxxx.yyyyyyy  */
/*  [0]  [1]    [2]      [3]   [4]   [5]      [6]    */
/* if index = 2, getting 'cccc[][]'.           */
int	futl_get_struct_part_name(
	char	*name,		/* IN  */
	int	indx,		/* IN  */
	char	*part_name	/* OUT */
)
{
	int	nn;
	int	stnum = 0;
	char	*pp = name;
	
	nn = 0;
	while(*pp != '\0'){
		switch(*pp){
		case '.':
			stnum++;
			break;
		default:
			if(stnum == indx){
				part_name[nn] = *pp;
				nn++;
			}
		}
		pp++;
		if(stnum > indx) break;
	}
	part_name[nn] = '\0';
	return nn;
}


/* aaaa.bbbb[].cccc[][].ddddd.eeeee.xxxxxxx.yyyyyyy  */
/*  [1]  [2]    [3]      [4]   [5]   [6]      [7]    */
/* if index = 3, getting 'aaaa.bbbb[].cccc[][]'.           */
int	futl_get_struct_upto_part_name(
	char	*name,		/* IN  */
	int	indx,		/* IN  */
	char	*part_name	/* OUT */
)
{
	int	i,nn;
	int	strnn;
	int	stnum=1;
	
	nn = 0;
	strnn = strlen(name);
	for(i=0;i<strnn;i++){
		switch(name[i]){
		case '.':
			stnum++;
			break;
		}
		if(stnum > indx) break;
		nn++;
		part_name[i] = name[i];
		part_name[i+1] = '\0';
	}
	return nn;
}


/*
 * If name1="a.b[10].c.d[10][10]" and name2="a.b[8].x.d[10]",
 * this function returns 2. Index is ignored.
 */
int	futl_count_same_struct_part(
	char	*name1,
	char	*name2
)
{
	int	i, strn[2];
	char	*buf1 = 0x0,*buf2 = 0x0;
	char	*str1 = 0x0,*str2 = 0x0;
	int	stnum=0;
	strn[0] = strlen(name1); strn[1] = strlen(name2);
	buf1 = (char*)malloc(strn[0]+2);
	str1 = (char*)malloc(strn[0]+2);
	buf2 = (char*)malloc(strn[1]+2);
	str2 = (char*)malloc(strn[1]+2);
	strn[0] = 1;
	strn[1] = 1;
	futl_ignore_array_index(name1,buf1);
	futl_ignore_array_index(name2,buf2);
	for(i=1; strn[0] > 0 && strn[1] > 0 ;i++){
		strn[0] = futl_get_struct_part_name(buf1,i,str1);
		strn[1] = futl_get_struct_part_name(buf2,i,str2);
		if(strn[0] <= 0 || strn[1] <= 0){
			goto DIFFERENT;
		}
		else{
			if(strcmp(str1,str2) == 0){
				stnum = i;
			}
			else{
				goto DIFFERENT;
			}
		}
	}
DIFFERENT:
	if(buf1) free(buf1);
	if(buf2) free(buf2);
	if(str1) free(str1);
	if(str2) free(str2);
	return stnum;
}

/*
 * If name1="a.b[10].c.d[10][10]" and name2="a.b[8].x.d[10]",
 * this function returns 0 or not 0.
 * aaa[].bbb.ccc.ddd[][]
 *  (1)  (2) (3)  (4)  = indx
 */
int	futl_compare_struct_name(
	int	indx,
	char	*name1,
	char	*name2
)
{
	int	i;
	int	strn[2];
	char	*buf1 = 0x0,*buf2 = 0x0;
	char	*str1 = 0x0,*str2 = 0x0;
	int	ret=0;
	strn[0] = strlen(name1); strn[1] = strlen(name2);
	buf1 = (char*)malloc(strn[0]+2);
	str1 = (char*)malloc(strn[0]+2);
	buf2 = (char*)malloc(strn[1]+2);
	str2 = (char*)malloc(strn[1]+2);
	strn[0] = 1;
	strn[1] = 1;
	futl_ignore_array_index(name1,buf1);
	futl_ignore_array_index(name2,buf2);
	for(i=1; (strn[0] > 0 && strn[1] > 0) && i <= indx ;i++){
		strn[0] = futl_get_struct_part_name(buf1,i,str1);
		strn[1] = futl_get_struct_part_name(buf2,i,str2);
		if(strn[0] <= 0 || strn[1] <= 0){
			ret = strn[0] - strn[1];
			goto DIFFERENT;
		}
		else{
			if((ret = strcmp(str1,str2)) != 0){
				goto DIFFERENT;
			}
		}
	}
DIFFERENT:
	if(buf1) free(buf1);
	if(buf2) free(buf2);
	if(str1) free(str1);
	if(str2) free(str2);
	return ret;
}

/*
 *  name1="a[0].b[10][2].c.d[10][10]" and name2="a[5].b[8][6].x.d[10]",
 *  this function returns 3.
 *
 */
int	futl_count_array_in_same_struct(
	char	*name1,
	char	*name2
)
{
	int	i;
	int	stcnt = 0;
	int	ardim = 0;
	int	num, *starray = 0x0;
	int	anum = 0;

	stcnt = futl_count_same_struct_part( name1, name2 );
	if(stcnt <= 0) return 0;
	anum = futl_count_all_array_bracket(name1);
	if(anum < 0) return 0;
	starray = (int*)calloc(anum,sizeof(int));
	num = futl_count_struct_array_bracket(name1,starray);
	for(i=0;i<stcnt;i++){
		ardim += starray[i];
	}
	if(starray) free(starray);
	return ardim;
}

/*
 * If name is aa[].bb[0][1][2].cc.dd[2][1],
 * array_index[0]=-1
 * array_index[1]=0
 * array_index[2]=1
 * array_index[3]=2
 * array_index[4]=2
 * array_index[5]=1.
 */
int	futl_get_array_index(
	const char	*name,
	int	*array_indx,
	char* ignorename
)
{
	short brsw = 0;
	short error = 0;
	short nullflag = 0;
	unsigned short brnum = 0;
	unsigned short incnt = 0;
	const char *sp;
	char *cc,buff[64]; /* String of Number. */
	char *pp = 0x0;
	int  val = 0;
	cc = buff;
	if(ignorename) pp = ignorename;
	for(sp = name ; *sp != '\0' ; sp++ ){
		switch(*sp){
		case '[':
			//////////////////////
			if(brsw == 0){
				if(ignorename){
					if(!futl_char_is_space(*sp)){
						(*pp++) = *sp;
					}
				}
			}
			//////////////////////
			brsw++;
			if(brsw == 1){
				cc = buff;
				*cc = '\0';
				incnt = 0;
			}
			break;
		case ']':
			if(brsw == 1){
				*cc = '\0';
				if(incnt > 0){
					if(nullflag) return (-99);
					if(futl_isnumber(buff)){
						val = atoi(buff);
						if(val < 0){
							val = 0;
							error = 1;
						}
					}
					else{
						error = 1;
						val = 0;
					}
					array_indx[brnum] = val;
				}
				else{
					nullflag = 1;
					array_indx[brnum] = -1;
				}
				incnt = 0;
				brnum++;
			}
			else{
				*cc = *sp;
				cc++;
				incnt++;
				if(incnt > 50) return (-99);
			}
			brsw --;
			if(brsw < 0) return (-99);
			//////////////////////
			if(brsw == 0){
				if(ignorename){
					if(!futl_char_is_space(*sp)){
						(*pp++) = *sp;
					}
				}
			}
			//////////////////////
			break;
		default:
			if(brsw > 0){
				if(!futl_char_is_space(*sp)){
					(*cc++) = *sp;
				}
				incnt++;
				if(incnt > 50) return (-99);
			}
			//////////////////////
			else if(brsw == 0){
				if(ignorename){
					if(!futl_char_is_space(*sp)){
						(*pp++) = *sp;
					}
				}
			}
			//////////////////////
		}
	}
	if(ignorename) *pp = 0x0;
	if(error > 0 || brsw != 0) return (-99);
	return brnum;
}
// A return value is the number of [] of name.
// aindex.size() is the number of available indexes.
//   name   ->    aindex
// "a[][][]"      []
// "a[][][2]      [-1,-1,2](allowPartNullIndex=true) or Error
// "a[][1][]"     [-1,1](allowPartNullIndex=true)
int	futl_get_array_index(
	const char	*name,
	vector<unsigned int>& aindex,
	char* ignorename,		// OUT:OPTION
	const char* ignorechar, // IN:OPTION
	const bool allowPartNullIndex, // = false
	const unsigned int nullValue // = UINT_MAX
)
{
	short brsw = 0;
	short error = 0;
	short each_loop = 0;
	unsigned short nullflag = 0;
	unsigned short brnum = 0;
	unsigned short incnt = 0;
	const char *sp;
	char *cc,buff[32]; /* String of Number. Max size is 30. */
	char *pp = 0x0;
	int  val = 0;
	cc = buff;
	if(ignorename) pp = ignorename;
	aindex.resize(0);
	for(sp = name ; *sp != '\0' ; sp++ ){
		each_loop = 0;
		switch(*sp){
		case '[':
			//////////////////////
			if(brsw == 0){
				if(ignorename){
					if(!futl_char_is_space(*sp)){
						(*pp++) = *sp;
					}
				}
			}
			//////////////////////
			brsw++;
			if(brsw == 1){
				cc = buff;
				*cc = '\0';
				incnt = 0;
			}
			break;
		case ']':
			if(brsw == 1){
				*cc = '\0';
				if(incnt > 0){
					if(nullflag > 0){
						if(allowPartNullIndex){
							for(val=0;val<nullflag;val++) aindex.push_back(nullValue);
							nullflag = 0;
						}
						else return (-99);
					}
					if(futl_isnumber(buff)){
						val = atoi(buff);
						if(val < 0){
							val = 0;
							error = 1;
						}
					}
					else{
						error = 1;
						val = 0;
					}
					aindex.push_back(val);
				}
				else{
					nullflag++;
				}
				incnt = 0;
				brnum++;
			}
			else{
				*cc = *sp;
				cc++;
				incnt++;
				if(incnt > 30) return (-99);
			}
			brsw --;
			if(brsw < 0) return (-99);
			//////////////////////
			if(brsw == 0){
				if(ignorename){
					if(!futl_char_is_space(*sp)){
						(*pp++) = *sp;
					}
				}
			}
			//////////////////////
			break;
		default:
			if(brsw > 0){
				if(!futl_char_is_space(*sp)){
					(*cc++) = *sp;
					incnt++;
				}
				if(incnt > 30) return (-99);
			}
			//////////////////////
			else if(brsw == 0){
				if(ignorechar){
					for(const char* op=ignorechar;*op != '\0';op++){
						if(*op == *sp){
							each_loop = 1;
							break;
						}
					}
				}
				if(ignorename && !each_loop){
					if(!futl_char_is_space(*sp)){
						(*pp++) = *sp;
					}
				}
			}
			//////////////////////
		}
	}
	if(ignorename) *pp = 0x0;
	if(error > 0 || brsw != 0) return (-99);
	return brnum;
}
// Return value is an array dimension.
int	futl_get_array_index(
	const char	*name,
	string& ignorename,           // OUT:NO INDEX NAME
	vector<unsigned int>& aindex, // OUT:INDEX
	const char* ignorechar,       // IN:OPTION
	const bool allowPartNullIndex, // = false
	const unsigned int nullValue   // = UINT_MAX
)
{
	aindex.clear();
	if(!name) return 0;
	char* namep = (char*)malloc(strlen(name)+1);
	int rc = futl_get_array_index(name,aindex,namep,ignorechar,allowPartNullIndex,nullValue);
	if(rc < 0){
		if(namep) free(namep);
		return -1;
	}
	ignorename = namep;
	if(namep) free(namep);
	return rc;
}
/**************************************************/
// "[123]xyz[[[V]]][AB]]" -> "123","[[V]]","AB"
// Returned 3.
int futl_get_bracket_string(
	const string& text,
	vector<string>& contents,
	bool ignore_sp
)
{
	if(text.size() <= 0) return 0;
	int brsw = 0;
	recycle_string buff;
	contents.clear();
	for(const char *sp = text.c_str() ; *sp != '\0' ; sp++ ){
		switch(*sp){
		case '[':
			if(brsw > 0){
				if(ignore_sp && !futl_char_is_space(*sp)){
						buff.cat(*sp);
				}
				else buff.cat(*sp);
			}
			brsw++;
			break;
		case ']':
			if(brsw == 1){
				contents.push_back(buff.get());
				buff.reset();
			}
			else{
				if(ignore_sp && !futl_char_is_space(*sp)){
						buff.cat(*sp);
				}
				else buff.cat(*sp);
			}
			brsw--;
			break;
		default:
			if(brsw > 0){
				if(ignore_sp && !futl_char_is_space(*sp)){
						buff.cat(*sp);
				}
				else buff.cat(*sp);
			}
		}
	}
	return contents.size();
}
/**************************************************/

/* aa[].bb[][] -> aa[0].bb[][] etc.
 * In this example,
 * array_nest=3,array_indx[0]=0,array_indx[1]=0,array_indx[2]=1.
 */
int	futl_set_array_index_upto_struct(
	char	*name,		/* IN  */
	int	array_nest,	/* IN  */
	int	*array_indx,	/* IN  */
	int	struct_depth,	/* IN (>0)  */
	char	*new_name	/* OUT */
)
{
	int	anest,struct_count;
	int	indx_nest;
	int	brsw=0;
	char	*pp, brbuff[64];
	char	*np,*sp = new_name;

	if(struct_depth <= 0){
		strcpy(new_name,name);
		/* No operation. */
		return 0;
	}
	if(array_nest <= 0){
		strcpy(new_name,name);
		/* No operation. */
		return 0;
	}
	/* UDB_ARYBR[] = {'[',']','\0'} */
	*sp = UDB_ARYBR[2];
	anest = futl_count_all_array_bracket(name);
	if(anest != array_nest){
		/* Error. */
		return -1;
	}
	indx_nest = 0;
	struct_count = 1;
	for(np = name; *np != '\0' ;np++){
		/* Structure count */
		switch(*np){
		case '.':
			struct_count++;
			break;
		}
		if(struct_count > struct_depth){
			*sp = *np;
			sp++;
			*sp = UDB_ARYBR[2];
			continue;
		}
		/* Array count */
		switch(mc_isbracket( *np )){
		case 1: /* [ */
			brsw = 1;
			break;
		case 2: /* ] */
			/*****/
			if(array_indx[indx_nest] < 0){
				sprintf(brbuff,"[]");
			}
			else{
				sprintf(brbuff,"[%d]",array_indx[indx_nest]);
			}
			pp = brbuff;
			while(*pp != '\0'){
				*sp = *pp;
				sp++;
				pp++;
			}
			*sp = UDB_ARYBR[2];
			indx_nest++;
			brsw = 0;
			break;
		default:
			if(brsw <= 0){
				*sp = *np;
				sp++;
				*sp = UDB_ARYBR[2];
			}
		}
	}
	return indx_nest;
}

#define FUTIL_NO_INDEX ((unsigned int)(-1))

int	futl_set_array_index(
	const char	*name,		/* IN  */
	int adim,
	unsigned int* aindex,
	int	struct_depth,	/* IN (>0)  */
	recycle_string& new_name	/* OUT */
)
{
	int	struct_count;
	int	indx_nest;
	int	brsw=0;
	char  brbuff[32];
	const char	*np; // *sp = new_name

	if(struct_depth <= 0){
		new_name.set((char*)name);
		/* No operation. */
		return 0;
	}
	if(adim <= 0){
		new_name.set((char*)name);
		/* No operation. */
		return 0;
	}
	/* UDB_ARYBR[] = {'[',']','\0'} */
	new_name.reset();
	indx_nest = 0;
	struct_count = 1;
	for(np = name; *np != '\0' ;np++){
		/* Structure count */
		switch(*np){
		case '.':
			struct_count++;
			break;
		}
		if(struct_count > struct_depth){
			new_name.cat((char)*np);
			continue;
		}
		/* Array count */
		switch(mc_isbracket( *np )){
		case 1: /* [ */
			brsw = 1;
			break;
		case 2: /* ] */
			if(adim <= indx_nest){
				sprintf(brbuff,"[]");
			}
			else if(aindex[indx_nest] == FUTIL_NO_INDEX){
				sprintf(brbuff,"[]");
			}
			else{
				sprintf(brbuff,"[%d]",aindex[indx_nest]);
			}
			new_name.cat(brbuff);
			indx_nest++;
			brsw = 0;
			break;
		default:
			if(brsw <= 0){
				new_name.cat((char)*np);
			}
		}
	}
	return indx_nest;
}

int	futl_set_array_index(
	const char	*name,		/* IN  */
	vector<unsigned int>& aindex,
	int	struct_depth,	/* IN (>0)  */
	recycle_string& new_name	/* OUT */
)
{
	unsigned int* idx = 0;
	if(aindex.size() > 0){
		idx = &aindex[0];
	}
	return futl_set_array_index(name,aindex.size(),idx,struct_depth,new_name);
}

/* aa[].bb[][] -> aa[0].bb[0][1] etc.
 * In this example,
 * array_nest=3,array_indx[0]=0,array_indx[1]=0,array_indx[2]=1.
 */
int	futl_set_array_index(
	char	*name,		/* IN  */
	int	array_nest,	/* IN  */
	int	*array_indx,	/* IN  */
	char	*new_name	/* OUT */
)
{
	return futl_set_array_index_upto_struct(
		name,array_nest,array_indx,32760,new_name);
}

// Processing of a name divided by dots.
// Returns the name of parent's structure.
// If returned zero, there is no parent structure.
int futl_up_struct_name(
	const char	*name,
	char	*newupname
)
{
	int i,plen;
	int nlen = strlen(name);
	plen = 0;
	for(i=(nlen-1);i>=0;i--){
		switch(name[i]){
		case '.':
			plen = i;
			goto Next;
			break;
		}
	}
Next:
	if(plen <= 0){
		newupname[0] = 0x0;
		return 0;
	}
	else{
		strncpy(newupname,name,plen);
		newupname[plen] = 0x0;
		return plen;
	}
}

int futl_append_struct_name(
	const char	*name,
	const char	*child,
	char	*newupname
)
{
	const char *sp = child;
	futl_ignore_space_strcpy(newupname,name);
	while(*sp != '\0'){
		if(futl_char_is_space(*sp)){
			sp++;
		}
		else{
			break;
		}
	}
	if(*sp == '.'){
		futl_ignore_space_strcat(newupname,sp);
	}
	else if(*sp != '\0'){
		futl_ignore_space_strcat(newupname,".");
		futl_ignore_space_strcat(newupname,sp);
	}
	return strlen(newupname);
}

/* Ignoring array bracket name. */
/* aaa[].bbb.ccc[][].ddd.eee.fff */
/* aaa.bbb.ccc.ddd.eee.fff. */
int	futl_ignore_bracket_name(
	char	*name,
	char	*get_name
)
{
	int	i;
	int	strnn;
	int	brsw = 0;
	char	*cc;
	int	ret = -1;

	cc = get_name;
	*cc = '\0';
	strnn = strlen(name);
	for(i=0;i<strnn;i++){
		switch(name[i]){
		case '[':
			brsw = 1;
			break;
		case ']':
			if(brsw > 0){
				brsw = 0;
			}
			break;
		default:
			if(brsw <= 0){
				*cc = name[i];
				cc++;
				*cc = '\0';
				ret = 1;
			}
		}
	}
	return ret;
}

/* Getting struct main name. */
/* aaa[].bbb.ccc[][].ddd.eee.fff */
/* Main name = "aaa". */
int	futl_get_main_name(
	const char	*name,
	char	*get_name
)
{
	int	ret;
	int	indx = 0;
	indx = 0;
	ret = futl_get_struct_part_name((char *)name,indx,get_name);
	return ret;
}

/* Getting struct last name. */
/* aaa[].bbb.ccc[][].ddd.eee.fff[]          */
/* Last name = "fff".   Remark [] !!!!!!!!! */
int	futl_get_last_name(
	char	*name,
	char	*get_name
)
{
	int	ret;
	int	indx = 0;
	indx = futl_get_struct_level(name);
	ret = futl_get_struct_part_name(name,indx,get_name);
	return ret;
}

/* /users/home/udf.def -> udf.def */
int	futl_get_filename(
	char	*path,	/* IN */
	char	*name	/* OUT */
)
{
	int	i, nlen, startn = 0;
	nlen = strlen(path);
	for(i=nlen-1 ; i>=0 ; i--){
		if(path[i] == futl_get_path_delimiter()){
			startn = i+1;
			break;
		}
	}
	strcpy(name,&(path[startn]));
	return (int)strlen(name);
}

/* .->_ */
/* aa -> 0            */
/* a.b[].c.d[][] -> 3 */
int	futl_dot2bar(char *name)
{
	int	i = 0;
	char	*p = name;
	while(*p != '\0'){
		if(*p == '.'){
			*p = '_';
			i++;
		}
		p++;
	}
	return i;
}

/* Search struct number with last array. */
/* a.b[].c.d[][] -> 4 */
/* a.b[].c[][].d -> 3 */
/* a.b[].c.d -> 2 */
/* a.b.c.d -> 0 */
int	futl_search_last_array_struct(char *name)
{
	//int	i = 0;
	int	strcount = 1;
	int	arydepth = 0;
	int	arycount = 0;
	int	ary_str_level = 0;
	char	*p = name;
	while(*p != '\0'){
		switch(*p){
		case '.':
			if(arycount > 0){
				ary_str_level = strcount;
				arycount = 0;
			}
			strcount++;
			break;
		case '[':
			if(arydepth == 0) arycount++;
			arydepth++;
			break;
		case ']':
			arydepth--;
			break;
		}
		p++;
	}
	if(arycount > 0){
		ary_str_level = strcount;
	}
	return ary_str_level;
}

//**************************************
//***         FILE OPERATION         ***
//**************************************
const char *futl_dirdelimit()
{
#ifdef WIN32
	return "\\";
#else
	return "/";
#endif
}

bool futl_isdirdelimit(const char chr)
{
	switch(chr){
	case '\\':
	case '/':
		return true;
	}
	return false;
}

void futl_append_dirdelimit( string& fullpath )
{
	if(fullpath.length() > 0){
		if(!futl_isdirdelimit(fullpath.at(fullpath.length()-1))){
			fullpath = fullpath + futl_dirdelimit();
		}
	}
}

bool futl_makepath(const char* path, const char* fname, string& fullpath)
{
	fullpath = path;
	if(fullpath.length() > 0){
		if(!futl_isdirdelimit(fullpath.at(fullpath.length()-1))){
			fullpath = fullpath + futl_dirdelimit();
		}
	}
	fullpath = fullpath + fname;
	return true;
}

static bool futl_base_splitpath(const char* path, string* pathpart, string* filepart)
{
	if(!path) return false;
	int	i, len, loop ;
	int	cnt_delimiter ;
	int	div_point = 0;
	
	len = strlen(path) ;
	for(i=len-1,loop=1,cnt_delimiter=0;i>=0 && loop > 0;i--){
		switch( path[i] ){
		case	'\0' :
		case	'\n' :
		case	'\\' :
		case	'/' :
			cnt_delimiter++ ;
		/* case	' '  : */
		/* case	'\t' : */
			if(cnt_delimiter >= 1){
				div_point = i ;
				loop = 0 ;
			}
			break ;
		default :
			cnt_delimiter++ ;
		}
	}
	if(div_point > 0){
		if(pathpart){
			(*pathpart) = path;
			pathpart->erase(div_point,(len-div_point));
		}
		if(filepart){
			(*filepart) = &(path[div_point+1]);
		}
	}
	else{
		if(pathpart){
			(*pathpart) = "";
		}
		if(filepart){
			(*filepart) = path;
		}
	}
	return true;
}

bool futl_splitpath(const char* path, string& pathpart, string& filepart)
{
	if(!path) return false;
	return futl_base_splitpath(path, &pathpart, &filepart);
}

bool futl_pathpart( const char* path, string& pathpart )
{
	if(!path) return false;
	return futl_base_splitpath(path, &pathpart, NULL);
}
bool futl_filepart( const char* path, string& filepart )
{
	if(!path) return false;
	return futl_base_splitpath(path, NULL, &filepart);
}

int futl_path_compare(const char* path1,const char* path2)
{
	return futl_string_compare(path1,path2,"\\/",false);
}

int futl_split_extension(const string& name, string& base, string& extension)
{
	string::size_type wdpas = name.rfind("\\");
	if(wdpas == name.npos) wdpas = 0;
	string::size_type udpas = name.rfind("/");
	if(udpas == name.npos) udpas = 0;
	//////////////////////////////////////////////
	string::size_type npos = name.rfind(".");
	if(npos == name.npos || npos <= 0 || npos < wdpas || npos < udpas){
		base = name;
		return -1;
	}
	else{
		base = name.substr(0,npos);
		extension = name.substr(npos+1,name.size()-npos);
		return (int)npos;
	}
}

int futl_replace_extension(const string& name, const string& extension, string& result)
{
	string base;
	string ext;
	int ext_pos = futl_split_extension(name, base, ext);
	if(extension.length() > 0){
		if(extension[0] == '.'){
			result = base + extension;
		} else {
			result = base + ".";
			result = result + extension;
		}
	}
	return ext_pos;
}

//*****************************************
//***         POINTER OPERATION         ***
//*****************************************

/* How many pointer(<>) expression in variable. */
int	futl_udf_pointer_express(
	char	*variable
)
{
	int	brin = 0;
	int	count = 0;
	char	*ss = variable;
	while(*ss != '\0'){
		switch(*ss){
		case '<':
			brin++;
			break;
		case '>':
			if(brin >= 1){
				brin--;
				count++;
				if(brin <= 0) return count;
			}
			break;
		}
		ss++;
	}
	return count;
}

int	futl_udf_simple_pointer_express(
	char	*variable
)
{
	char	*ss = variable;
	while(*ss == ' ' || *ss == '\t' || *ss == '\n' || *ss == '\r') ss++;
	if(*ss != '<') return 0;
	return futl_udf_pointer_express(ss);
}

/* "<bond[].atom1>.pos" -> "bond[].atom1" */
int	futl_udf_pointer_get_content(
	char	*variable,	/* IN  */
	char	*content	/* OUT */
)
{
	int	i;
	int	startp = -1, endp = -1, lenp;
	int	count = 0;

	if(futl_udf_pointer_express( variable ) <= 0){
		return 0;
	}
	/*** same as futl_udf_pointer_get_content ***/
	/* get first & most inside brackets */
	for(i=0;variable[i] != '\0';i++){
		switch(variable[i]){
		case '<':
			if(count <= 0){
				startp = i+1;
			}
			break;
		case '>':
			if(count <= 0 && startp > 0){
				endp = i-1;
				count = 1;
			}
			break;
		}
	}
	/************************************/
	if(startp >= 0 && endp >= 0){
		lenp = endp - startp + 1;
		if(lenp > 0){
			strncpy(content,&(variable[startp]),lenp);
			content[lenp] = '\0';
			count = 1;
		}
	}
	return count;
}

/* <atom[]>.pos + molecule.atom[] -> molecule.atom[].pos */
/* No pointer bracket.  */
int	futl_udf_pointer_replace_content(
	char	*variable,	/* IN/OUT  */
	char	*content	/* IN */
)
{
	int	i,j,nn;
	int	startp = -1, endp = -1;
	int	replaceflg = 0;
	int	count = 0;
	char	buff[UDB_MAX_BUFF_SIZE];

	/***
	if(futl_udf_pointer_express( variable ) <= 0){
		return 0;
	}
	***/
	/*** same as futl_udf_pointer_get_content ***/
	/* get first & most inside brackets */
	for(i=0;variable[i] != '\0';i++){
		switch(variable[i]){
		case '<':
			if(count <= 0){
				startp = i+1;
			}
			break;
		case '>':
			if(count <= 0 && startp > 0){
				endp = i-1;
				count = 1;
			}
			break;
		}
	}
	/************************************/
	if(startp > 0 && endp > 0){
		startp --;
		endp += 2;
	}
	if(count <= 0) return 0;
	count = 0;
	nn = 0;
	for(i=0;variable[i] != '\0';i++){
		if(i == startp) replaceflg = 1;
		else if(i == endp) replaceflg = 0;
		if(replaceflg <= 0){
			buff[nn] = variable[i];
			nn++;
		}
		else if(replaceflg == 1){
			count = 1;
			for(j=0;content[j] != '\0';j++){
				buff[nn] = content[j];
				nn++;
			}
			replaceflg = 2;
			startp = -1;
		}
	}
	buff[nn] = '\0';
	strcpy(variable,buff);
	return count;
}

/* <atom[]>.pos + molecule.atom[] -> <molecule.atom[]>.pos */
/* With <> bracket !!! */
int	futl_udf_pointer_br_replace_content(
	char	*variable,	/* IN/OUT  */
	char	*content	/* IN */
)
{
	int	i,j,nn;
	int	startp = -1, endp = -1;
	int	replaceflg = 0;
	int	count = 0;
	char	buff[UDB_MAX_BUFF_SIZE];

	/***
	if(futl_udf_pointer_express( variable ) <= 0){
		return 0;
	}
	***/
	/*** same as futl_udf_pointer_get_content ***/
	/* get first & most inside brackets */
	for(i=0;variable[i] != '\0';i++){
		switch(variable[i]){
		case '<':
			if(count <= 0){
				startp = i+1;
			}
			break;
		case '>':
			if(count <= 0 && startp > 0){
				endp = i-1;
				count = 1;
			}
			break;
		}
	}
	/************************************/
	if(startp > 0 && endp > 0){
		endp ++;
	}
	if(count <= 0) return 0;
	count = 0;
	nn = 0;
	for(i=0;variable[i] != '\0';i++){
		if(i == startp) replaceflg = 1;
		else if(i == endp) replaceflg = 0;
		if(replaceflg <= 0){
			buff[nn] = variable[i];
			nn++;
		}
		else if(replaceflg == 1){
			count = 1;
			for(j=0;content[j] != '\0';j++){
				buff[nn] = content[j];
				nn++;
			}
			replaceflg = 2;
			startp = -1;
		}
	}
	buff[nn] = '\0';
	strcpy(variable,buff);

	return count;
}

#ifdef STOPE_TEST_POINTER
main()
{
	int	rc;
	char	*pp;
	char	variable[256];
	char	content[256];	/* OUT */
	/***/
	strcpy(variable,"<<TotalMolecules.molecule[]>.bond[].atom1>.coord");
	strcpy(content,"TotalMolecules.molecule[0]");
	rc = futl_udf_pointer_br_replace_content( variable,content);
	printf("VARIABLE:[%s]\n",variable);
	strcpy(content,"TotalMolecules.molecule[0].atom[1]");
	rc = futl_udf_pointer_br_replace_content( variable,content);
	printf("VARIABLE:[%s]\n",variable);
	/***/
	strcpy(variable,"<<bond[0].xx>.atom1>.pos");
	strcpy(content,"molecule.atom[]");
	rc = futl_udf_pointer_get_content( variable,content);
	printf("CONTENT:[%s]\n",content);
	strcpy(content,"molecule.atom[1]");
	rc = futl_udf_pointer_get_content( variable,content);
	printf("CONTENT:[%s]\n",content);
	/***/
	strcpy(variable,"<<bond[0].xx>.atom1>.pos");
	strcpy(content,"molecule.atom[1]");
	rc = futl_udf_pointer_replace_content( variable, content);
	if(rc > 0){
		printf("[%s]\n",variable);
	}
	strcpy(content,"molecule.atom[1]");
	rc = futl_udf_pointer_replace_content( variable, content);
	printf("[%s]\n",variable);
	/***
	pp = futl_trimstrdup("  abc 123   ");
	pp = futl_trimstrdup("            ");
	pp = futl_trimstrdup("abc 123");
	***/
	rc = futl_get_struct_upto_part_name("aa.bb[].ccc[][].dd",5,variable);
	printf("[%s]\n",variable);
	/***/
}
#endif
#ifdef STOPE_TEST_EXAMPLES
main()
{
	int	rc;
	char	*name;
	int	array_nest;	/* IN  */
	int	array_indx[8];	/* IN  */
	int	struct_depth;	/* IN (>0)  */
	char	new_name[256];	/* OUT */

	name = "aaa.bb[].cc.dd[][].ee.f[]";
	array_nest = 4;
	array_indx[0] = 2;
	array_indx[1] = 3;
	array_indx[2] = 8;
	array_indx[3] = 9;
	struct_depth = 4;
	rc = futl_set_array_index_upto_struct(
		name,array_nest,array_indx,struct_depth,new_name);
	printf("RET[%d]:[%s]\n",rc,new_name);
	printf("**********\n");
	name = "aaa.bb[].cc.dd[][].ee.f[]";
	rc =  futl_search_last_array_struct(name);
	printf("RET[%d]:[%s]\n",rc,name);
	name = "a.b[].c.d[][]";
	rc =  futl_search_last_array_struct(name);
	printf("RET[%d]:[%s]\n",rc,name);
	name = "a.b[].c[][].d";
	rc =  futl_search_last_array_struct(name);
	printf("RET[%d]:[%s]\n",rc,name);
	name = "a.b[].c.d";
	rc =  futl_search_last_array_struct(name);
	printf("RET[%d]:[%s]\n",rc,name);
	name = "a.b.c.d";
	rc =  futl_search_last_array_struct(name);
	printf("RET[%d]:[%s]\n",rc,name);
}
#endif

