
#ifndef REREAD_DATA_LISTENER_H
#define REREAD_DATA_LISTENER_H

#include "binaryInput.h"
#include "FileStreamManager.h"
#include "StructureCacheOperation.h"
#include "CacheEventController.h"

/**
 *  Reread array data listener
 */
class RereadDataListener {
	FileStreamManager* fileMgr;
	StructureCacheOperation* operation;
	CacheEventController* controller;

public:
	/**
	 * @brief Constructor
	 */
	RereadDataListener():
		fileMgr(0),operation(0),
		controller(0){}
	/**
	 * @brief Set objects
	 * @param[in] (fileMgr_) FileStreamManager object
	 * @param[in] (operation_) StructureCacheOperation object
	 * @param[in] (controller_) CacheEventController object
	 */
	void setObjects(FileStreamManager* fileMgr_,
		StructureCacheOperation* operation_,
		CacheEventController* controller_){
		fileMgr = fileMgr_;
		operation = operation_;
		controller = controller_;
	}
	/**
	 * @brief Set FileStreamManager object
	 * @param[in] (fileMgr_) FileStreamManager object
	 */
	void setFileManager(FileStreamManager* fileMgr_){
		fileMgr = fileMgr_;
	}
	virtual ~RereadDataListener(){
		fileMgr = 0;
		operation = 0;
		controller = 0;
	}
	/**
	 * @brief Reread data
	 * @param[in] (id) data definition id
	 * @param[in] (section) 0:common, 1:record
	 * @return true for normal end
	 */
	virtual bool reread(int id, unsigned int section) {
		UINT recordNumber = 0;
		if (section > 0) {
			recordNumber = controller->getCurrentRec();
			// Get the original record number after editing records.
			CacheRecordOperation* recop = fileMgr->getRecordOperation();
			if (recop) {
				recordNumber = recop->getOriginalRecordSimpleWay(recordNumber);
			}
		}

		bool is_virtual = fileMgr->isOpenedVirtual();
		string reread_file;
		const char* open_file = fileMgr->getOpenedFile();
		if (open_file) {
			reread_file.assign(open_file);
		}
		else { /* Normally, this process is not entered. */
			if (!fileMgr->getReadFile(recordNumber, reread_file, is_virtual)) {
				return false;
			}
		}

		BinaryInput binRead(is_virtual); // Close in destructor.
		if(!binRead.open(reread_file.c_str(), is_virtual)){
			return false;
		}
		return binRead.rereadData(id, recordNumber, fileMgr, operation);
	}
};


#endif

