#ifndef UTIL_CORE_CACHE_LIMIT_ARRAY_SECTION_H
#define UTIL_CORE_CACHE_LIMIT_ARRAY_SECTION_H

#include "UtilCoreCacheSection.h"

namespace CacheUtilCore {

	class CoreLimitByteArraySection : public CoreByteArraySection {
	public:
		CoreLimitByteArraySection(){}
		~CoreLimitByteArraySection(){}
		
		virtual bool set(const uint index,const BYTE_TYPE value,const bool append = true)
		{
			if(isRereading()){
				return CoreByteArraySection::set(index, value, append);
			}
			if(sizeOfRead() == 0 || index < sizeOfRead()){
				return CoreByteArraySection::set(index, value, append);
			}
			if(reread()){
				return CoreByteArraySection::set(index, value, append);
			}
			return false;
		}
		
		virtual bool get(const uint index,BYTE_TYPE& value,bool* exist = 0x0)
		{
			if(sizeOfRead() == 0 || index < sizeOfRead()){
				return CoreByteArraySection::get(index, value, exist);
			}
			if(reread()){
				return CoreByteArraySection::get(index, value, exist);
			}
			value = 0;
			if(exist){
				*exist = true;
			}
			return ( index < sequenceSize() );
		}

		virtual bool insert(int arrayLevel, uint start, uint number, const bool locationFront)
		{
			if (sizeOfRead() > 0) {
				reread();
			}
			return CoreByteArraySection::insert(arrayLevel, start, number, locationFront);
		}

		virtual bool erase(int arrayLevel, uint start, uint number)
		{
			if (sizeOfRead() > 0) {
				if (0 == sequenceSize()) { //  Erased all!
					setSizeOfRead(0);
				}
				else {
					reread();
				}
			}
			return CoreByteArraySection::erase(arrayLevel, start, number);
		}
	};
	
	class CoreLimitShortArraySection : public CoreShortArraySection {
	public:
		CoreLimitShortArraySection(){}
		~CoreLimitShortArraySection(){}
		
		virtual bool set(const uint index,const SHORT_TYPE value,const bool append = true)
		{
			if(isRereading()){
				return CoreShortArraySection::set(index, value, append);
			}
			if(sizeOfRead() == 0 || index < sizeOfRead()){
				return CoreShortArraySection::set(index, value, append);
			}
			if(reread()){
				return CoreShortArraySection::set(index, value, append);
			}
			return false;
		}
		
		virtual bool get(const uint index,SHORT_TYPE& value,bool* exist = 0x0)
		{
			if(sizeOfRead() == 0 || index < sizeOfRead()){
				return CoreShortArraySection::get(index, value, exist);
			}
			if(reread()){
				return CoreShortArraySection::get(index, value, exist);
			}
			value = 0;
			if(exist){
				*exist = true;
			}
			return ( index < sequenceSize() );
		}

		virtual bool insert(int arrayLevel, uint start, uint number, const bool locationFront)
		{
			if (sizeOfRead() > 0) {
				reread();
			}
			return CoreShortArraySection::insert(arrayLevel, start, number, locationFront);
		}

		virtual bool erase(int arrayLevel, uint start, uint number)
		{
			if (sizeOfRead() > 0) {
				if (0 == sequenceSize()) { //  Erased all!
					setSizeOfRead(0);
				}
				else {
					reread();
				}
			}
			return CoreShortArraySection::erase(arrayLevel, start, number);
		}
	};
	
	class CoreLimitIntArraySection : public CoreIntArraySection {
	public:
		CoreLimitIntArraySection(){}
		~CoreLimitIntArraySection(){}
		
		virtual bool set(const uint index,const INT_TYPE value,const bool append = true)
		{
			if(isRereading()){
				return CoreIntArraySection::set(index, value, append);
			}
			if(sizeOfRead() == 0 || index < sizeOfRead()){
				return CoreIntArraySection::set(index, value, append);
			}
			if(reread()){
				return CoreIntArraySection::set(index, value, append);
			}
			return false;
		}
		
		virtual bool get(const uint index,INT_TYPE& value,bool* exist = 0x0)
		{
			if(sizeOfRead() == 0 || index < sizeOfRead()){
				return CoreIntArraySection::get(index, value, exist);
			}
			if(reread()){
				return CoreIntArraySection::get(index, value, exist);
			}
			value = 0;
			if(exist){
				*exist = true;
			}
			return ( index < sequenceSize() );
		}

		virtual bool insert(int arrayLevel, uint start, uint number, const bool locationFront)
		{
			if (sizeOfRead() > 0) {
				reread();
			}
			return CoreIntArraySection::insert(arrayLevel, start, number, locationFront);
		}

		virtual bool erase(int arrayLevel, uint start, uint number)
		{
			if (sizeOfRead() > 0) {
				if (0 == sequenceSize()) { //  Erased all!
					setSizeOfRead(0);
				}
				else {
					reread();
				}
			}
			return CoreIntArraySection::erase(arrayLevel, start, number);
		}
	};
	
	class CoreLimitDoubleArraySection : public CoreDoubleArraySection {
	public:
		CoreLimitDoubleArraySection(){}
		~CoreLimitDoubleArraySection(){}
	
		virtual bool set(const uint index,const DOUBLE_TYPE value,const bool append = true)
		{
			if(isRereading()){
				return CoreDoubleArraySection::set(index, value, append);
			}
			if(sizeOfRead() == 0 || index < sizeOfRead()){
				return CoreDoubleArraySection::set(index, value, append);
			}
			if(reread()){
				return CoreDoubleArraySection::set(index, value, append);
			}
			return false;
		}
		
		virtual bool get(const uint index,DOUBLE_TYPE& value,bool* exist = 0x0)
		{
			if(sizeOfRead() == 0 || index < sizeOfRead()){
				return CoreDoubleArraySection::get(index, value, exist);
			}
			if(reread()){
				return CoreDoubleArraySection::get(index, value, exist);
			}
			value = 0.0;
			if(exist){
				*exist = true;
			}
			return ( index < sequenceSize() );
		}

		virtual bool insert(int arrayLevel, uint start, uint number, const bool locationFront)
		{
			if (sizeOfRead() > 0) {
				reread();
			}
			return CoreDoubleArraySection::insert(arrayLevel, start, number, locationFront);
		}

		virtual bool erase(int arrayLevel, uint start, uint number)
		{
			if (sizeOfRead() > 0) {
				if (0 == sequenceSize()) { //  Erased all!
					setSizeOfRead(0);
				}
				else {
					reread();
				}
			}
			return CoreDoubleArraySection::erase(arrayLevel, start, number);
		}
	};
	
	class CoreLimitStringArraySection : public CoreStringArraySection {
	public:
		CoreLimitStringArraySection(){}
		virtual ~CoreLimitStringArraySection(){}
	
		virtual bool set(const uint index,const char* value,const bool append = true){
			if(isRereading()){
				return CoreStringArraySection::set(index, value, append);
			}
			if(sizeOfRead() == 0 || index < sizeOfRead()){
				return CoreStringArraySection::set(index, value, append);
			}
			if(reread()){
				return CoreStringArraySection::set(index, value, append);
			}
			return false;
		}
		
		virtual bool get(const uint index,char** value,bool* exist = 0x0)
		{
			if(sizeOfRead() == 0 || index < sizeOfRead()){
				return CoreStringArraySection::get(index, value, exist);
			}
			if(reread()){
				return CoreStringArraySection::get(index, value, exist);
			}
			*value = "";
			if(exist){
				*exist = true;
			}
			return ( index < sequenceSize() );
		}

		virtual bool insert(int arrayLevel, uint start, uint number, const bool locationFront)
		{
			if (sizeOfRead() > 0) {
				reread();
			}
			return CoreStringArraySection::insert(arrayLevel, start, number, locationFront);
		}

		virtual bool erase(int arrayLevel, uint start, uint number)
		{
			if (sizeOfRead() > 0) {
				if (0 == sequenceSize()) { //  Erased all!
					setSizeOfRead(0);
				}
				else {
					reread();
				}
			}
			return CoreStringArraySection::erase(arrayLevel, start, number);
		}
	};

} // End of namespace CacheUtilCore
#endif

