
/*
 *  Material Modeling Platform Package
 *  Copyright(c) 2002, The Japan Research Institute, Ltd.
 *  All rights reserved.
 *  
 *  $Id: binaryInput.h,v 1.1.1.1 2004/11/29 12:19:13 nishimoto Exp $
 */

#ifndef BINARY_INPUT_H
#define BINARY_INPUT_H
#include "UtilHugeFile.h"
#include "file_stream_record.h"
#include <string>

class FileStreamManager;
class StructureCacheOperation;
class AbstractReadUDFRecord;
class TextUdfOutput;
class BinaryOutput;
class BinaryDivideRecordFile;

struct BinaryFileInfo {
	std::string       fileName;
	unsigned int      recordNumber;
	REC_POSITION_TYPE recordPosition;
	bool              is_virtual_record; // Edited or appended record.
	BinaryFileInfo() :recordNumber(0), recordPosition(0), is_virtual_record(false) {}
};

struct BinaryFileOutput {
	BinaryFileOutput():outputBinaryHeader(false){}
	bool outputBinaryHeader;
};

class BinaryInput {
private:
	bool existBinaryCommon;
	bool _get_virtual_info;
	bool _binaryOpendIsVirtual;
	FILE* _binaryOpend;
	std::string _binaryOpendFile;
	unsigned int _searchedRecord; // 0:not found.
	AbstractReadUDFRecord* _read_recorder;
	TextUdfOutput* _textOutput;
	longint_t binary_RecordNumber;
	BinaryDivideRecordFile* divide_record_data;
	// Method.
	void checkBinaryDivide();
public:
	BinaryInput(bool get_virtual = true):
		existBinaryCommon(false),
		_get_virtual_info(get_virtual),
		_binaryOpendIsVirtual(false),
		_binaryOpend(0),
		_searchedRecord(0),
		_read_recorder(0),
		_textOutput(0),
		divide_record_data(0){}
	virtual ~BinaryInput();
	bool isBinaryCommon();
	// File operation.
	FILE* open( const char* file, bool open_virtual = false );
	void close();
	bool seek( longint_t pos );
	static bool seek( FILE* fp, longint_t pos );
	longint_t seekEnd();
	longint_t tell();
	bool eof();
	// Optional operation.
	void setRecorder( AbstractReadUDFRecord* recorder );
	AbstractReadUDFRecord* getRecorder(){ return _read_recorder; }
	// Text output.
	bool createTextOutput( const char* file, bool appendMode );
	void destroyTextOutput();
	// Binary output.
	//bool createBinaryOutput( const char* file, int divide_type );
	//void destroyBinaryOutput();
	// Divide record
	void setDirectory(const char* dirname);
	const char* getDirectory();
	void setRecordName(const char* recname);
	const char* getRecordName();
	void setRecordPath(const char* path);
	const char* getRecordPath();
	void setRecordNumber(longint_t num);
	longint_t getRecordNumber();
	// Initial reading.
	bool initialReadProcess(
		FileStreamManager* fmgr,
		StructureCacheOperation* operation,
		int&         section,
		std::string& label,
		longint_t&   datasize,
		int*         shiftFineTuning,
		FILE*        binaryOpened
	);
	bool convertBinaryData(
		StructureCacheOperation* operation,
		int&         section,
		std::string& label,
		longint_t&   datasize,
		int*         shiftFineTuning,
		FILE*        binaryOpened
	);
	bool readDevideBinary(
		int&         section,
		std::string& label,
		longint_t&   datasize,
		int*         shiftFineTuning,
		FileStreamManager*       fmgr,
		StructureCacheOperation* operation
	);
	// Read.
	int readBinaryInitial(
		unsigned int             lastRecord,
		FileStreamManager*       fileMgr,
		StructureCacheOperation* operation,
		unsigned int&            totalRecordNumber,
		const char*              serchLabel = 0,
		bool                     readOnlyHeader = false
	);
	bool readBinaryRecord(
		BinaryFileInfo&          recordInfo,
		FileStreamManager*       fileMgr,
		StructureCacheOperation* operation
	);
	bool readBinaryRecord(
		unsigned int             recordNumber,
		FileStreamManager*       fileMgr,
		StructureCacheOperation* operation
	);
	bool rereadData(
		int		id,
		unsigned int             recordNumber,
		FileStreamManager*       fileMgr,
		StructureCacheOperation* operation
	);
	bool rereadAllData(
		unsigned int             recordNumber,
		FileStreamManager*       fileMgr,
		StructureCacheOperation* operation
	);
	// record number
	unsigned int getSearchedRecord(){
		return _searchedRecord;
	}
	int skipDividedRecords(
		FileStreamManager*       fileMgr,
		StructureCacheOperation* operation,
		unsigned int&            totalRecordNumber
	);
};


#endif
