
/*
 *
 *  Material Modeling Platform Package
 *  
 *  Copyright(c) 2000, The Japan Research Institute, Ltd.
 *  All rights reserved.
 *
 */

#ifdef WIN32
#pragma warning (disable:4786)
#endif

/**
 **     The Main Routine of UDF Analizer
 **/
#include <stdio.h>
#include "yylib.h"
#include "udf_operations.h"
#include "udf_functions.h"
#include "yut_fileope.h"
#include "cache_configure.h"
#include "mhdl_iostream.h"
using namespace std;
#include "UDFParser.h"
#include "MasterCacheMaker.h"
#include "CacheEventController.h"
#include "CacheUnitInterface.h"
#include "udfParserClass.h"
#include "CacheExportManager.h"
#include "binaryInput.h"


UDFParser::UDFParser()
{
	_classkey = "UDFParser";
	_maker = 0;
	_recorder = 0;
}

UDFParser::UDFParser( MasterCacheMaker* maker )
{
	_classkey = "UDFParser";
	_maker = maker;
	_recorder = 0;
}

UDFParser::~UDFParser()
{
	_maker = 0;
}

int
UDFParser::read_udf_file(
	UINT   rec,
	char** message,
	int    reload_force_load_option // 0:none, 1:reload, 2:force_load.
)
{
	E_PARSER_ERROR rc = ePRS_NORMAL;
	*message = 0x0;
	if(reload_force_load_option <= 0){
		CacheEventController* controler = _maker->getController();
		if(controler){
			//// no need for reading of a current record. /////
			long LastReadRec = controler->getLastReadRec();
			if(LastReadRec >= 0){
				if(LastReadRec == ((long)rec) || rec == 0){
					controler->setCurrentRec(rec);
					return 0;
				}
			}
		}
	}
	if(rec <= 0){ // First reading.
		rc = read_udf_spec( ePARSE_READY, rec, 0x0, message );
	}
	else{
		_maker->clearCache(1);  // Clear Data.
		switch(reload_force_load_option){
		case 1: // RELOAD.
			rc = read_udf_spec(ePARSE_READY_REC,rec,0x0,message);
			break;
		case 2: // FORCE TO LOAD.
		default:
			rc = read_udf_spec(ePARSE_RECORD,rec,0x0,message);
		}
	}
	if(rc != ePRS_NORMAL) return -1;
	return 0;
}

int
UDFParser::read_udf_file(
	const char*  label,
	char** message,
	int    reload
)
{
	E_PARSER_ERROR rc = ePRS_NORMAL;
	*message = 0x0;
	
	if(!reload){
		FileStreamManager* fs = _maker->getFileStream();
		if(fs){
			bool rb;
			UINT rec;
			UINT line;				// OUT
			REC_POSITION_TYPE pos;  // OUT
			rb = fs->getRecInfo(label,&rec,&line,&pos);
			if(!rb){
				return -1;
			}
			//// no need for reading of a current record. /////
			CacheEventController* controler = _maker->getController();
			if(controler){
				controler->setRecordInfo(rec, line, pos);
				long LastReadRec = controler->getLastReadRec();
				if(LastReadRec >= 0){
					if(LastReadRec == ((long)rec) || rec == 0){
						controler->setCurrentRec(rec);
						return 0;
					}
				}
			}
		}
	}
	_maker->clearCache(1); // Clear Data.
	if(reload){
		rc = read_udf_spec(ePARSE_READY_REC,0,(char*)label,message);
	}
	else{
		rc = read_udf_spec(ePARSE_RECORD,0,(char*)label,message);
	}
	if(rc != ePRS_NORMAL) return -1;
	return 0;
}

int
UDFParser::read_udf_file(
	AbstractReadUDFRecord* recorder,
	char** message
)
{
	*message = 0x0;
	if(!recorder) return -1;
	_recorder = recorder;
	E_PARSER_ERROR rc = read_udf_spec( ePARSE_CONTINUE, 0, 0x0, message );
	if(rc != ePRS_NORMAL) return -1;
	return 0;
}

int
UDFParser::read_udf_file(
	EPARSE_TYPE parse,
	char** message
)
{
	*message = 0x0;
	E_PARSER_ERROR rc = read_udf_spec( parse, 0, 0x0, message );
	if(rc != ePRS_NORMAL) return -1;
	return 0;
}

int UDFParser::read_record_directory(
	const char* record_directory,
	char** message
)
{
	*message = 0x0;
	udfParserClass parser(_maker, ePARSE_RECORD, false);
	E_PARSER_ERROR rc = parser.readBinaryRecordDirectory(record_directory);
	const char* str = parser.get_error().getMessage();
	if (str) {
		*message = strdup(str);
	}
	if (rc != ePRS_NORMAL) return -1;
	return 0;
}

E_PARSER_ERROR
UDFParser::file_check()
{
	FileStreamManager* fsmgr = _maker->getFileStream();
	if(!fsmgr) return ePRS_ERR_SYSTEM;
	return ePRS_NORMAL;
}


E_PARSER_ERROR
UDFParser::read_udf_spec(
	EPARSE_TYPE prstype,
	UINT  record_no,
	char* rec_label,
	char** message
)
{
	E_PARSER_ERROR err = file_check();
	if(err != ePRS_NORMAL) return err;

	/* Initialize */	
	int	ret = 0;
	char* str = 0x0;
	char* msg = 0x0;
	UINT reading_record = record_no;
	string main_message;
	REC_POSITION_DATA* recPosition = 0;
	*message = msg;
	if(rec_label) record_no = 999;

	/* Input stream. */
	FileStreamManager* fsmgr = _maker->getFileStream();
	// Substitution for the Edit/Append record files.
	bool rb = false;
	const char* vrecfile = 0x0; /* Virtual record file: edited or appended record */
	if(prstype == ePARSE_RECORD){
		CacheRecordOperation* recop = fsmgr->getRecordOperation();
		if(recop->isChanged()){
			if(rec_label){ // By record label.
				unsigned int tmp_record = 0;
				rb = recop->findRecordFromLabel( rec_label, tmp_record );
				if(rb){
					if(recop->isEditedRecord(tmp_record)){
						vrecfile = recop->getEditedRecordFilePath( tmp_record );
						if(vrecfile){
							prstype = ePARSE_NOCOND; // Read a edited record file.
							reading_record = record_no = tmp_record;
						}
					}
					if(!vrecfile){
						if(recop->isOriginal( tmp_record )){
							unsigned int original = recop->getOriginalRecord( tmp_record );
							if(original != CACHE_RECORD_NO_RECORD){
								reading_record = original;
								record_no = tmp_record;
							}
						}
					}
				}
			}
			else{ // By record number.
				if(recop->isEditedRecord(record_no)){ // Original & edited, inserted, appended.
					vrecfile = recop->getEditedRecordFilePath( record_no );
					if(vrecfile){
						prstype = ePARSE_NOCOND; // Read a edited record file.
					}
				}
				if(!vrecfile){
					if(recop->isOriginal( record_no )){
						unsigned int original = recop->getOriginalRecord( record_no );
						if(original != CACHE_RECORD_NO_RECORD){
							reading_record = original;
						}
					}
					else{
						// INSERTED & NO DATA RECORD.
						CacheEventController* controler = _maker->getController();
						controler->setCurrentRec( record_no );
						controler->setLastReadRec( record_no );
						*message = strdup("");
						return err;
					}
				}
			}
		}
		if(vrecfile){
			if(fsmgr->exfile(vrecfile)){
				if(fsmgr->isBinaryRecord( reading_record )){
					BinaryFileInfo recordInfo;
					recordInfo.fileName = vrecfile;
					recordInfo.recordNumber = reading_record;
					recordInfo.recordPosition = 0;
					recordInfo.is_virtual_record = true;
					if(parseBinaryRecord(recordInfo,main_message)){
						CacheEventController* controler = _maker->getController();
						controler->setCurrentRec( record_no );
						controler->setLastReadRec( record_no );
						return err;
					}
					else{
						err = ePRS_ERR_PARSE;
						msg = strdup("binary read error.\n");
						*message = msg;
						return (err);
					}
				}
				else{ /* Text udf file. */
					if(fsmgr->openTempFile(vrecfile) <= 0){
						*message = strdup("Can not open.");
						return ePRS_ERR_FILE;
					}
				}
			}
			else{
				vrecfile = 0x0;
			}
		}
	}
	//-TemporaryCacheRecord* tmprec = 0x0;
	//-if(prstype == ePARSE_RECORD){
	//-	CacheVirtualRecord* vrec = fsmgr->getVirtualRecord();
	//-	if(rec_label){
	//-		rb = vrec->getVRec(eVIRTUAL_ACCESS_BOTH,rec_label,&tmprec);
	//-		if(rb){
	//-			vrecfile = tmprec->getFile();
	//-			prstype = ePARSE_NOCOND;
	//-			rec_no = tmprec->getRecordNo();
	//-		}
	//-	}
	//-	else{
	//-		rb = vrec->getVRec(eVIRTUAL_ACCESS_BOTH,rec_no,&tmprec);
	//-		if(rb){
	//-			vrecfile = tmprec->getFile();
	//-			prstype = ePARSE_NOCOND;
	//-		}
	//-	}
	//-	if(vrecfile){
	//-		if(fsmgr->exfile(vrecfile)){
	//-			if(fsmgr->isBinaryRecord(tmprec->getRecordNo())){
	//-				BinaryFileInfo recordInfo;
	//-				recordInfo.fileName = tmprec->getFile();
	//-				recordInfo.recordNumber = tmprec->getRecordNo();
	//-				recordInfo.recordPosition = 0;
	//-				if(parseBinaryRecord(recordInfo,main_message)){
	//-					CacheEventController* controler = _maker->getController();
	//-					controler->setCurrentRec( tmprec->getRecordNo() );
	//-					controler->setLastReadRec( tmprec->getRecordNo() );
	//-					return (err);
	//-				}
	//-				else{
	//-					err = ePRS_ERR_PARSE;
	//-					msg = strdup("binary read error.\n");
	//-					*message = msg;
	//-					return (err);
	//-				}
	//-			}
	//-			else{
	//-				if(fsmgr->openTempFile(vrecfile) <= 0){
	//-					*message = strdup("Can not open.");
	//-					return ePRS_ERR_FILE;
	//-				}
	//-			}
	//-		}
	//-		else{
	//-			vrecfile = 0x0;
	//-		}
	//-	}
	//-}

	// Comparison of file correction times.
	if(!vrecfile){
		// Compare FileTime.
		if(fsmgr->timeCompare() > 0){
			// Since it is corrected, it is re-loading.
			fsmgr->setFileTime();
			_maker->reloadClear(reading_record); // 999?
			prstype = ePARSE_READY_REC;
		}
	}
	// record position.
	if(prstype == ePARSE_RECORD){
		if(rec_label){
			if(fsmgr->getRecInfo(rec_label,&record_no,&_recPosition.line,&_recPosition.pos)){
				recPosition = &_recPosition;
				reading_record = record_no;
			}
			else{
				err = ePRS_ERR_STEP;
			}
		}
		else {
			if(fsmgr->getRecInfo(reading_record,&_recPosition.line,&_recPosition.pos)){
				recPosition = &_recPosition;
			}
			else{
				if(fsmgr->isBinaryRecord(reading_record) && fsmgr->isDividedRecord()){
					CacheEventController* controler = _maker->getController();
					controler->setCurrentRec(record_no);
					controler->setLastReadRec(record_no);
					return err; /* Skip reading a empty record file. */
				}
				else {
					err = ePRS_ERR_STEP;
				}
			}
		}
		if(err != ePRS_NORMAL){
			main_message += "Step designation error.\n";
			// Creating message string.
			msg = strdup(main_message.c_str());
			*message = msg;
			return err;
		}
		if(fsmgr->isBinaryRecord(reading_record)){
			BinaryFileInfo recordInfo;
			if(fsmgr->isDividedRecord()){
				string record_file;
				bool flag = false;
				if(fsmgr->getRecInfo(reading_record,0,0,0,&record_file,&flag)){
					recordInfo.fileName = record_file;
				}
			} else {
				recordInfo.fileName = fsmgr->getFile();
			}
			recordInfo.recordNumber = reading_record;
			recordInfo.recordPosition = recPosition->pos;
			if(parseBinaryRecord(recordInfo,main_message)){
				CacheEventController* controler = _maker->getController();
				controler->setCurrentRec( record_no );
				controler->setLastReadRec( record_no );
				return (err);
			}
			else{
				err = ePRS_ERR_PARSE;
				msg = strdup("binary read error.\n");
				*message = msg;
				return (err);
			}
		}
	}
	// Normal Openning file.
	if(!vrecfile){
		if(fsmgr->openFile() <= 0){
			*message = strdup("Can not open.");
			return ePRS_ERR_FILE;
		}
	}
	/* Parser */
	udfParserClass parser(_maker,prstype);
	parser.udfparser( this );
	switch(prstype){
	case ePARSE_READY: // Reading header,definitions,common data,...
	case ePARSE_READY_REC: // Reloading.
		if(rec_label){
			if(parser.setOptions( rec_label,recPosition )){
				*message = strdup("No needed object.");
				err = ePRS_ERR_SYSTEM;
				goto Error;
			}
		}
		else{
			if(parser.setOptions( reading_record,recPosition )){
				*message = strdup("No needed object.");
				err =  ePRS_ERR_SYSTEM;
				goto Error;
			}
		}
		// Create headers & unit data.
		parser._header->createHeaderInAttr();
		_maker->getUnit()->parseStartUnit();
		break;
	case ePARSE_RECORD: // Reading only one record.
		///////////////////////////////////////////////////
		//if(rec_label){
		//	if(parser.setOptions( rec_label,recPosition )){
		//		*message = strdup("No needed object.");
		//		err = ePRS_ERR_SYSTEM;
		//		goto Error;
		//	}
		//}
		//else{
		///////////////////////////////////////////////////
		if(parser.setOptions( reading_record,recPosition )){
			*message = strdup("Record number error.");
			err =  ePRS_ERR_SYSTEM;
			goto Error;
		}
		break;
	case ePARSE_CONTINUE:
		parser.setReadUDFRecord(_recorder);
		if(parser.setOptions( reading_record,recPosition )){
			*message = strdup("Record number error.");
			err =  ePRS_ERR_SYSTEM;
			goto Error;
		}
		// Create headers & unit data.
		parser._header->createHeaderInAttr();
		_maker->getUnit()->parseStartUnit();
		break;
	case ePARSE_READY_BINARY:
		if(parser.setOptions( (unsigned int)0,0 )){
			*message = strdup("No needed object.");
			err =  ePRS_ERR_SYSTEM;
			goto Error;
		}
		// Create headers & unit data.
		parser._header->createHeaderInAttr();
		_maker->getUnit()->parseStartUnit();
		break;
	case ePARSE_NOCOND:  // Appended or edited records are read.
		if(parser.setOptions( reading_record,recPosition )){
			*message = strdup("Record number error.");
			err =  ePRS_ERR_SYSTEM;
			goto Error;
		}
		// WITHOUT NEXT LINE, VIRTUAL DIVIDED RECORD FILE READS TO COMMON.
		parser.lexer->set_step_counter(reading_record);
		break;
	default:
		err =  ePRS_ERR_SYSTEM;
		goto Error;
	}
	// Starting to parse. //
	ret = parser.yyparse();
	// Processing at the time of normal termination.
	if(!ret){
		// Divided record file.(ver 4.1.0)
		if(parser.isDividedRecordFile()){
			//if(!fsmgr->compareRecFile( reading_record, parser.getDividedRecordFile() )){
			//	printf("Different record file at %d\n",reading_record);
			//}
		}
		// Rearrangement processing.
		switch(prstype){
		case ePARSE_READY:
		case ePARSE_READY_REC:
		case ePARSE_CONTINUE:
			if(!_maker->getUnit()->parseEndUnit(main_message)){
				main_message += "Unit conversion service cannot be offered.\n";
			}
			break;
		case ePARSE_RECORD:
		case ePARSE_NOCOND:
		default:
			break;
		}
		// 
		// Proper registered record number.
		//
		UINT ReadRec = parser.lexer->getTargetRecordID();
		REC_POSITION_TYPE LastPos = 0;
		REC_LINENUMBER_TYPE  ReadLine = 0;
		switch(prstype){
		case ePARSE_READY:     // Reading to the last of a file,investigating all record positions.
		case ePARSE_READY_REC: // And specification record data are read too.
			LastPos = parser.getLastPosition(); // always returns the last file position.
			ReadLine = parser.getLine(); // the last text line number (from a lexer).
			if(parser.useBoyerMoore()){  // 2004/12/24.
				ReadLine = parser.getBoyerMooreLine(); // (from the Boyer-Moore searching).
			}
			fsmgr->setLastPosition(LastPos);
			fsmgr->setLastLineNo(ReadLine);   // Set all line number.
			break;
		default:
			//-ReadRec = tmprec->getRecordNo();
			ReadRec = record_no;
		}
		//////////////////////////////////////////////////////////
		CacheEventController* controler = _maker->getController();
		controler->setCurrentRec( ReadRec );
		controler->setLastReadRec( ReadRec );
	}
	// Error message operation //
	if(ret > 0){
		// Error messages.
		str = (char*)parser.get_error().getMessage();
		main_message += str;
		err = parser.get_error().lastError();
		if(err == ePRS_NORMAL){
			main_message += "\nAn unrecognized error occurred in the UDF file.\n";
			err = ePRS_ERR_PARSE;
		}
		UINT ReadRec = parser.lexer->getTargetRecordID();
		CacheEventController* controler = _maker->getController();
		controler->setCurrentRec( ReadRec );
		controler->setLastReadRec( ReadRec );
	}
	else if(parser.get_error().status()){
		// Warning messages.
		str = (char*)parser.get_error().getMessage();
		main_message += str;
		err = ePRS_ERR_WARN;
	}
	else if(parser.parser_record_operation_status){
		err = ePRS_ERR_STEP;
		main_message += "Step designation error.\n";
	}
	else{
		// Normal end.
		if(parser.get_error().getSize() > 0){
			str = (char*)parser.get_error().getMessage();
			main_message += str;
		}
	}
	// Creating message string.
	msg = strdup(main_message.c_str());
	*message = msg;
Error:
	fsmgr->closeFile();
	return (err);
}

bool UDFParser::parseBinaryRecord(
	BinaryFileInfo& recordInfo,
	string& message
)
{
	FileStreamManager* fileMgr = _maker->getFileStream();
	StructureCacheOperation* operation = _maker->getCache();
	BinaryInput binRead(recordInfo.is_virtual_record);
	if(!binRead.open(recordInfo.fileName.c_str(), recordInfo.is_virtual_record)) return false;
	// begin record. //////////////////
	_maker->beginParseUnconditional();
	///////////////////////////////////
	bool rb = binRead.readBinaryRecord(recordInfo,fileMgr,operation);
	if(rb){
		// end record. //////////////////
		if(!_maker->endParseUnconditional(message)){
			message += "\n";
		}
		/////////////////////////////////
	}
	return rb;
}
