
/*
 *
 *  Material Modeling Platform Package
 *  
 *  Copyright(c) 2000, The Japan Research Institute, Ltd.
 *  All rights reserved.
 *
 */
#ifdef WIN32
#pragma warning (disable:4786)
#endif
#include <stdlib.h>
#include <string.h>
#include "yut_fileope.h"
#include "cache_util_character.h"
#include "udfParserClass.h"
#include "udfparseyy.cpp.h"
#include "binaryInput.h"
#include "StructureCacheOperation.h"
#include "CacheEventController.h"
// Search algorithm:
#include "simple_bm.h"
#include "ugz_file.h"
#include "udf_bm_comment.h"
///////////////////////////
#ifdef WIN32
#if !defined(snprintf)
#define snprintf _snprintf
#endif
#endif
///////////////////////////
#include <vector>
#include <set>
using namespace std;
using namespace util_stack;

static char null_string[2] = { '\0', '\0' };

int
udfParserClass::setOptions(
	UINT	record_number,
	REC_POSITION_DATA* recPosition
)
{
	_record = record_number;
	_searchRecLabel = 0;
	option();
	return lexer_initialize(recPosition);
}

int
udfParserClass::setOptions(
	char* record_label,
	REC_POSITION_DATA* recPosition
)
{
	_searchRecLabel = (const char*)record_label;
	option();
	return lexer_initialize( recPosition, record_label );
}

void
udfParserClass::option()
{
	switch(parse_type){
	case ePARSE_READY:
		record_operation_option = PARSER_RECORD_ALL;
		break;
	case ePARSE_READY_REC:
		record_operation_option = PARSER_RECORD_ALL;
		break;
	case  ePARSE_RECORD:
		record_operation_option = PARSER_RECORD_AT;
		break;
	case ePARSE_CONTINUE:
		record_operation_option = PARSER_RECORD_EACH;
		break;
	case ePARSE_READY_BINARY: // DEFINITION & BINARY HEADER!
		record_operation_option = PARSER_RECORD_UNTILBIN;
		break;
	case ePARSE_NOCOND:
		record_operation_option = PARSER_RECORD_ALL;
		break;
	}
}

udfParserClass::udfParserClass(
	MasterCacheMaker* maker,
	EPARSE_TYPE prstype,
	bool open_file
)
{
	recordParser = 0;
	yyyynerrs = 0;

	_structRoot.setObj(maker,this,&parser_record_operation_status );
	_key_ref = new udfSearchDefs(maker->getDefine());

	//////////
	_header = new udfParserHeader(maker);
	_attr = new udfParserAttr(maker);

	// Parser init. //
	_searchRecLabel = 0;
	_continuous_written_recorder = -9999;
	_continuous_recorder = 0;
	_create_file_name = 0x0;
	////////////////////////////////
	initDefPreset();
	yut_memtext_init(&_attrtext);
	// _attr_list.resize(0);
	_attr_list.clearObjects();
	err_ret = ePRS_NORMAL;
	parse_type = prstype;
	record_counter_before = -1;
	record_counter_after = 0;
	parser_record_operation_status = 0;
	record_operation_option = PARSER_RECORD_ALL;
	InitializeParse();

	// String.
	errorbuffer.step(64);
	wholename.step(64);
	tdefname.step(64);
	tdefback.step(64);
	tdefindx.step(64);
	tdataval.step(64);
	// Attribute.
	adefname.step(64);
	adefback.step(64);
	adataval.step(64);

	// MasterCacheMaker
	_maker = maker;
	_record = 0;
	_arg_yyout = 0x0;
	_arg_yyin = 0;
	_yyin = 0;
	lexer = 0;
	
	// use 'Boyer-Moore' record search.
	_use_boyer_moore = false;
	_boyer_moore_line = 0;

	if (!open_file) {
		return;
	}

	//////// _maker->getFileStream()->openFile(); ////////
	_arg_yyin = (ifstream*)_maker->getFileStream()->getStream();

	// Creating Lexer.
	_yyin = _arg_yyin;
	lexer = new udfFlexLexer( _arg_yyin, _arg_yyout );
}


udfParserClass::~udfParserClass()
{
	if(recordParser) delete recordParser;

	delete _header;
	delete _attr;
	delete _key_ref;
	if(_create_file_name) free(_create_file_name);
	clearEveryFamily();
	cleanDefPreset();
	yut_memtext_clear(&_attrtext);
	
	delete lexer;
}

char*
udfParserClass::getFile()
{
	char* path = getPath();
	if(!path) return 0x0;
	char* dir_part = (char*)malloc(strlen(path)+1);
	if(_create_file_name) free(_create_file_name);
	_create_file_name = (char*)malloc(strlen(path)+1);
	yut_divide_dir_file(path,dir_part,_create_file_name);
	return _create_file_name;
}

char*
udfParserClass::getPath()
{
	FileStreamManager* fsm = _maker->getFileStream();
	if(!fsm) return 0x0;
	return fsm->getFile();
}

int
udfParserClass::yyparse(void)
{
	/***/
	if(parse_type == ePARSE_RECORD && recordParser != NULL){
		if(recordParser->parseUDF()) return 0; // normal
		else                         return 1; // error.
	}
	/***/
	
	return uyyparse((void *)this);
}


int
udfParserClass::yyyylex (YYYYSTYPE *yylval)
{
	int	ret;
	ret = lexer->yylex();
	if(ret == 0){
		if(lexer->restart_out() > 0){
			return INCFILE;
		}
	}
	return ret;
}

int
udfParserClass::lexer_initialize( REC_POSITION_DATA* recPosition, char* record_label )
{
	/***/
	if(parse_type == ePARSE_RECORD){
		bool binaryb = _maker->getFileStream()->isBinaryRecord(_record);
		//const char* file = _maker->getFileStream()->getOpenedFile();
		if(!binaryb && recPosition != NULL){
			recordParser = new UDFRecordParser( this );
			//if(!recordParser->open( file ) ) return -1;
			//if(!recordParser->seek( recPosition->pos )) return -1;
			//recordParser->setShiftLineNo( recPosition->line );
		}
	}
	/***/

	lexer->set_file(_maker->getFileStream()->getFile());
	lexer->set_path(_maker->getFileStream()->getFile());

	return
	lexer->setOptions(
		&_parser_error,
		_maker->getFileStream(),
		parse_type,
		_record,
		recPosition,
		record_label
	);
}

unsigned int
udfParserClass::getLine()
{
	return lexer->get_line_no();
}
REC_POSITION_TYPE
udfParserClass::getLastPosition()
{
	return lexer->get_last_position();
}

// Divided record file.
bool udfParserClass::isDividedRecordFile()
{
	if(recordParser){
		return recordParser->isRecordFile();
	}
	return false;
}
const char* udfParserClass::getDividedRecordFile()
{
	if(recordParser){
		return recordParser->getRecordFile();
	}
	return NULL;
}

void
udfParserClass::InitializeParse()
{
	pointer_brct = 0;
	parserDefCount = 0;
	type_flag = eUDF_PARSE_ATTR_NON;
	wholename.reset();
	tdefname.reset();
	tdefback.reset();
	tdefindx.reset();
	tdataval.reset();
	message_buff[0] = 0x0;
	namep = 0x0;
	typep = 0x0;
	parserOrdering = 0;
	retval = 0;
	header = 0;
	expnd = 0;
	definiter = 0;
	attributer = 0;
	classer = 0;
	current_etype = eTYPE_NON;
}

void
udfParserClass::clearEveryFamily()
{
	type_stack.clear();
	unit_stack.clear();
	structure_stack.clear();
}
void
udfParserClass::clearEveryData()
{
	adefhelp.clear();
	unit_stack.clear();
}
int
udfParserClass::sizeStack()
{
	return type_stack.size();
}
char* 
udfParserClass::getNameTopStack()
{
	udfParseDataType* ss = type_stack.top();
	if(ss){
		return (char*)ss->name.c_str();
	}
	return 0x0;
}
udfParseDataType* 
udfParserClass::getTypeTopStack()
{
	return  type_stack.top();
}

int 
udfParserClass::builtinType( char* stype )
{
	E_DATA_TYPE etype;
	etype = fcmgr_type_name2etype( stype );
	current_etype = etype;
	switch(etype){
	case eTYPE_NON:
	case eTYPE_OID:
	//case eTYPE_STRUCT: // REMARK!! Included the builtin-types.
	case eTYPE_ARRAY:
		return -1; // error.
	case eTYPE_USER:
		return 0;
	default: // builtin type.
		if(checkClassPointing(etype,stype) > 0){
			return 1;
		}
		else{
			return -1;
		}
	}
}

int
udfParserClass::checkClassPointing( E_DATA_TYPE etype, char* stype )
{
	int ret = 1;
	char* pointing_name = 0x0;
	switch(etype){
	case eTYPE_REFINDEX:
	case eTYPE_REFKEY:
		pointing_name = (char*)malloc(strlen(stype)+1);
		if(fcmgr_get_keyindex_class(stype, pointing_name) > 0){
			DefineCache* defcch = _maker->getDefine();
			DefineTable* dtbl;
			int dtbln = defcch->get_def(pointing_name,&dtbl,eDEFINE_GET_REAL);
			if(dtbln > 0){
				if(dtbl->etype == eTYPE_STRUCT){
					ret = 1;	// OK.
				}
			}
			else{
				ret = (-1);	// The class name was not found.
				yyyyerror("class name was not found");
			}
		}
		else{
			ret = (-1);	// Probably it is impossible.
		}
	}
	if(pointing_name) free(pointing_name);
	return ret;
}


void
udfParserClass::error_mem_add( char *err_msg )
{
	_parser_error.setError(ePRS_ERR_PARSE,err_msg);
}
void
udfParserClass::warn_mem_add( char *warn_msg )
{
	_parser_error.setError(ePRS_ERR_WARN,warn_msg);
}

void udfParserClass::clearMessage()
{
	_parser_error.clear();
}

ParserErrorCache&
udfParserClass::get_error( )
{
	return _parser_error;
}

int
udfParserClass::step_initialize()
{
	int	ret = 0;
	switch(record_operation_option){
	case PARSER_RECORD_ALL:	/* ALL */
		if(parse_type == ePARSE_READY){
			if(searchBoyerMoore() == ePRS_NORMAL){
				// Nothind.
			}
			else{ // Error.
				err_ret = ePRS_ERR_PARSE;
				unsigned long linenum = getBoyerMooreLine();
				//sprintf(message_buff,"Unexpected error occurred in searching records,line %d",linenum);
				sprintf(message_buff,"Unexpected error occurred in searching records,line %lu",linenum);
				yyyyAddError( message_buff );
			}
			ret = 1; // termination.
		}
		break;
	case PARSER_RECORD_UNTIL: /* UNTIL */
		break;
	case PARSER_RECORD_EACH:
		break;
	case PARSER_RECORD_AT:	/* AT */
		if(lexer->get_read_counter() > 0){
			ret = 1; // termination.
		}
		break;
	}
	record_counter_before++;
	return ret;
}

int
udfParserClass::step_finalize()
{
	int	ret = 0;
	switch(record_operation_option){
	case PARSER_RECORD_ALL:	/* ALL */
		break;
	case PARSER_RECORD_UNTIL: /* UNTIL */
		if(lexer->get_read_counter() > 0){
			ret = 1;
		}
		break;
	case PARSER_RECORD_EACH:
		ret = readContinuous();
		break;
	case PARSER_RECORD_AT:	/* AT */
		if(lexer->get_read_counter() > 0){
			ret = 1;
		}
		break;
	}
	record_counter_after++;
	return ret;
}

/*
 * Adding data name & type to the structure data array.
 */
void
udfParserClass::initDefPreset()
{
	_def_ptbl.clear();
}
void
udfParserClass::cleanDefPreset()
{
	_def_ptbl.clear();
}

// Preset definitions in one family.
int	
udfParserClass::addDefPreset(
	int          ordering,
	const char*  fullname,
	char*        datatype,
	char*        user_type_name,
	int          defsection,
	int          development,
	int          class_flag,
	const vector<string>& helpmsg,
	const vector<string>& unitarray,
	short def_location,
	CacheArrayIndexSize* arraySpec
)
{
	int rc = 0;
	int	i;
	char* buff = 0x0;
	DefineSetTable *pre;
	vector<DefineSetTable>* ptbl = &_def_ptbl;

	if(datatype == NULL) return 0;
	//// Get top data-name. ////
	buff = (char*)malloc(strlen(fullname)+1);
	fudf_get_main_name(fullname,buff);
	//// Create preset table. ////
	int nn = ptbl->size();
	ptbl->resize(nn+1);
	pre = &((*ptbl)[nn]);
	//////////////////////////////
	pre->iattr.resize(2);   // [0]:develop;[1]:group;
	pre->selcond.resize(2); // [0]:class_flag;[1]:global
	pre->top = buff;
	pre->ordering = ordering;
	pre->name = fullname;
	pre->etype = fcmgr_type_name2etype(datatype);
	pre->stype = datatype;
	if(user_type_name) pre->utype = user_type_name;
	else               pre->utype = "";
	pre->location = def_location;
	// Help.
	if(helpmsg.size() > 0){
		pre->help.resize(helpmsg.size());
		for(i=0;i<helpmsg.size();i++){
			pre->help[i] = helpmsg[i];
		}
	}
	// Unit.
	if(unitarray.size() > 0){
		pre->unit.resize(unitarray.size());
		for(i=0;i<unitarray.size();i++){
			pre->unit[i] = unitarray[i];
		}
	}
	// Development 
	pre->iattr[0] = development;
	// Definition section.
	pre->iattr[1] = defsection;
	/* CLASS (No instance) DECLARATION. */
	pre->selcond[0] = class_flag;
	// Global definition flag(switch).
	pre->selcond[1] = 0;
	if(defsection == DUDF_GLOBAL_DEF_SECTION){
		pre->selcond[1] = 1;
	}
	//////////////////////////
	if(arraySpec){
		if(arraySpec->effectual()){
			unsigned int arraydim = arraySpec->getDimension();
			if(arraydim > 0){
				unsigned int nempty = arraySpec->sizeEmptyArrayDimension(); // size of [].
				if(arraydim == nempty || nempty == 0){}
				else{
					rc = RET_ARRAYSPEC_ERR;
				}
			}
			pre->AsizeSpec.resize(arraydim);
			for(i=0;i<arraySpec->getDimension();i++){
				switch(arraySpec->getType(i)){
				case INDEX_SIZE_NUMBER:
					pre->AsizeSpec[i].setNumber(arraySpec->getNumber(i));
					break;
				case INDEX_SIZE_STRING:
					pre->AsizeSpec[i].setName(arraySpec->getString(i));
					break;
				default:
					pre->AsizeSpec[i].setNone();
				}
			}
		}
		arraySpec->cleaning();
	}
	// ptbl->push_back(pre);
	/************************/
	if(buff) free(buff);
	if(rc < 0) return rc;
	return ptbl->size();
}
int	
udfParserClass::findDefPreset(
	const char	*fullname
)
{
	vector<DefineSetTable>* ptbl = &_def_ptbl;
	for(unsigned int i=0;i<ptbl->size();i++){
		if(strcmp((*ptbl)[i].name.c_str(),fullname) == 0){
			return 1;
		}
	}
	return 0;
}

// Return false at the time of a syntax error.
bool 
udfParserClass::setDefUnit(const char* unit_txt)
{
	if(!unit_txt) return true;
	//  is it already set.
	set<string>::iterator sp = def_unit_set.find(unit_txt);
	if(sp != def_unit_set.end()) return true;
	def_unit_set.insert(unit_txt);
	// it includes operator "*/^+-".
	if(!futl_find_character(unit_txt,"*/^+-")) return true;
	// set.(error as syntax error.)
	E_PARSER_ERROR retv = _attr->setUnitDef(unit_txt,string_message);
	if(retv == ePRS_NORMAL) return true;
	return false;
}

/**
 * @brief Read binary records from an all-in-one file or divided record files. 
 * @param[in] (binRead) BinaryInput object
 * @param[in] (rec_no) target record number
 * @param[in] (fileMgr) FileStreamManager object
 * @param[in] (operation) StructureCacheOperation object
 * @return true for normal end
 */
bool readBinaryRecord(
	BinaryInput& binRead,
	unsigned int rec_no,
	FileStreamManager* fileMgr,
	StructureCacheOperation* operation
)
{
	BinaryInput *binput = 0;
	BinaryInput binDivRecRead;
	if(fileMgr->isDividedRecord()){
		string record_file;
		bool binary_record_file = false;
		if(fileMgr->getRecInfo(rec_no, 0, 0, 0, &record_file, &binary_record_file)){
			if(binDivRecRead.open(record_file.c_str())){
				binput = &binDivRecRead;
			}
		}
	} else {
		binput = &binRead;
	}
	if(binput){
		return binput->readBinaryRecord(rec_no,fileMgr,operation);
	}
	return false;
}


// Binary read.
E_PARSER_ERROR udfParserClass::readBinary()
{
	// Does not execute for the previous errors.
	if(error_status()){
		return ePRS_ERR_FATAL;
	}
	// treatStream(); // process of the left stream.
	FileStreamManager* fileMgr = _maker->getFileStream();
	StructureCacheOperation* operation = _maker->getCache();
	BinaryInput binRead;
	if(_continuous_recorder){
		binRead.setRecorder(_continuous_recorder);
	}
	if(!binRead.open(fileMgr->getFile())){
		return ePRS_ERR_FATAL;
	}
	unsigned int recNumber = lexer->getRecordCounter();
	unsigned int totalRecNumber = 0;
	bool readOnlyHeader = false;

	// The following specifies whether binary data is read.
	switch(parse_type){
	case ePARSE_READY_BINARY:
		readOnlyHeader = true;
		break;
	default:
		break;
	}

	if(!binRead.eof()){
		int status = binRead.readBinaryInitial(
				recNumber,fileMgr,operation,
				totalRecNumber,
				_searchRecLabel,readOnlyHeader);
		switch(status){
		case 0: // Normal.
			break;
		case 1:
		{
			string wkmsg = "file or directory is read-only.\n";
			yyyywarning((char*)wkmsg.c_str());
		}
		break;
		case 10:
		{
			string wkmsg = "Unexpected binary data error.\n";
			yyyyAddError((char*)wkmsg.c_str());
		}
		return ePRS_ERR_FATAL;
		//break;
		case 11: // Some record has error.
			{
				char buffer[32];
				string wkmsg = "Binary data after the ";
				wkmsg += IntToA(totalRecNumber,buffer);
				wkmsg += "-th record are collapsed.\n";
				yyyymessage((char*)wkmsg.c_str());
			}
			return ePRS_ERR_FATAL;
			//break;
		case 20:
		{
			string wkmsg = "Record directory exists but it's a normal file.\n";
			yyyymessage((char*)wkmsg.c_str());
		}
		return ePRS_ERR_FATAL;
		case 21:
		{
			string wkmsg = "Record directory does not exist.\n";
			setSingleWarn((char*)wkmsg.c_str());
		}
		return ePRS_NORMAL;
		default:
			if(operation->getBinaryReadError()){
				yyyyAddError(operation->getBinaryReadError());
			}
			return ePRS_ERR_FATAL;
		}
	}

	if(binRead.getRecorder()){
		// CONVRTER END.
		binRead.getRecorder()->terminate(); /* or _continuous_recorder */
		return ePRS_NORMAL;
	}

	// Reading of the binary header is completed here.
	if(readOnlyHeader){
		return ePRS_NORMAL;
	}

	// Reading of the binary data.
	if(binRead.isBinaryCommon()){
		if(!readBinaryRecord(binRead, 0U, fileMgr, operation)){
			if(operation->getBinaryReadError()){
				yyyyAddError(operation->getBinaryReadError());
			}
			return ePRS_ERR_FATAL;
		}
	}
	// Read ready & record.
	switch(parse_type){
	case ePARSE_READY_REC:
		if(_searchRecLabel){
			unsigned int searchRec = binRead.getSearchedRecord();
			if(searchRec <= 0) return ePRS_ERR_STEP;
			if(!readBinaryRecord(binRead, searchRec, fileMgr, operation)){
				return ePRS_ERR_FATAL;
			}
		}
		else{
			if(fileMgr->isBinaryRecord(_record)){
				if(!readBinaryRecord(binRead, _record, fileMgr, operation)){
					return ePRS_ERR_FATAL;
				}
			}
		}
		break;
	}
	return ePRS_NORMAL; // ePRS_ERR_FATAL.
}

E_PARSER_ERROR udfParserClass::readBinaryRecordDirectory(const char* record_directory)
{
	unsigned int totalRecordNumber = 0;
	FileStreamManager* fileMgr = _maker->getFileStream();
	StructureCacheOperation* operation = _maker->getCache();
	operation->setDivideDirectory(record_directory);

	BinaryInput binRead;
	int rec_status = binRead.skipDividedRecords(fileMgr, operation, totalRecordNumber);
	switch (rec_status) {
	case 0: /* Normal */
	case 1: /* Read only */
		break;
	default:
		return ePRS_ERR_FATAL;
	}

	// Fix the number of records in UDF.
	fileMgr->callAtEndOfFirstParse();

	if (binRead.isBinaryCommon()) {
		if (!readBinaryRecord(binRead, 0U, fileMgr, operation)) {
			if (operation->getBinaryReadError()) {
				yyyyAddError(operation->getBinaryReadError());
			}
			return ePRS_ERR_FATAL;
		}
	}
	CacheEventController* controller = _maker->getController();
	controller->forceBinaryDivideEdited();

	return ePRS_NORMAL;
}

// Read an text UDF.
int udfParserClass::readContinuous()
{
	int ret = 0;
	if(!_continuous_recorder->alive()) return ret;
	int record_count = lexer->get_step_counter();
	if(_continuous_written_recorder == record_count) return ret;
	if(record_count < 0 && !_continuous_recorder->needlessText()){
		// Output text type definitions & initial data.
		if(!_continuous_recorder->operationDefine()){
			ret = 1;
		}
		// Common.
		if(!_continuous_recorder->operationCommon()){
			ret = 1;
		}
	}
	else if(record_count <= 0){
		// Definitions & initial data & record.
		if(!_continuous_recorder->operationDefine()){
			ret = 1;
		}
		// Common.
		if(!_continuous_recorder->operationCommon()){
			ret = 1;
		}
		// Record.
		if(record_count < 0){ // needlessText() is true.
			if(lexer->getDataBlock() > 0){
				_continuous_recorder->operationNonrecordText();
			}
		}
		else{ // Record. -> Output binary record.
			if(!_continuous_recorder->operationRecord((record_count+1))){
				ret = 1;
			}
		}
		// Next.
		if(!_continuous_recorder->operationNextRecord()) ret = 1;
	}
	else{
		// Record. -> Output binary record.
		if(!_continuous_recorder->operationRecord((record_count+1))){
			ret = 1;
		}
		// Next.
		if(!_continuous_recorder->operationNextRecord()) ret = 1;
	}
	_continuous_written_recorder = record_count;
	return ret;
}

// Search "\begin{record}" using the algorithm(Boyer-Moore).
E_PARSER_ERROR udfParserClass::searchBoyerMoore()
{
	FileStreamManager* fileMgr = _maker->getFileStream();
	UGZ_FILE ufile(fileMgr->getFile(),"rb");
	if(!ufile.good()) return ePRS_ERR_FATAL;

	// [0] start.
	_use_boyer_moore = true;

	// [1] make pattern.
	DBMPattern bmpat("\\begin"); // (1) searched pattern.
	bmpat.makeTable();           // (2) create table.

	// [2] udf comment format.
	UdfBMComment* ubm = new UdfBMComment(); // find "{record}".
	BMSearch bms(ubm);
	UdfBMRecord bmrec;

	// get Lexer infomation.
	const char* reclabel = 0;
	unsigned long linenum = lexer->getLastEndLinenumber();
	TBM_FILE_REC_TYPE recpos = lexer->getLastEndPosition();
	if(linenum > 0) linenum--;

	// [3] search.
	int iRet;
	unsigned int rec_no = 0;
	bool binaryMode = false;
	
	while( bms.BMMatch(&ufile, bmpat, recpos) ){ // found "\\begin".
		recpos = bmpat.FoundPosition + bmpat.patlen;
		iRet = bmrec.searchRecord( &ufile, recpos );
		if(iRet >= 0){
			reclabel = bmrec.getKeyLabel();
			if(reclabel){
				rec_no++;
				_boyer_moore_line = linenum + ubm->getLineNumber();
				iRet = bmrec.searchLink( &ufile, recpos );
				if(iRet == CUDF_BM_YEN_LINK){
					const char* rec_file = bmrec.getRecFile();
					fileMgr->setRecInfo( rec_no, reclabel,
	                      _boyer_moore_line, bmpat.FoundPosition, rec_file); // Text!
					fileMgr->setDividedRecord( true ); // Text divide udf!
				}
				else{
					fileMgr->setRecInfo( rec_no, reclabel,
	                      _boyer_moore_line, bmpat.FoundPosition ); // Text!
#if defined(USING_NONDIVIDED_RECORD)
					fileMgr->setNondividedRecord();
#endif
				}
			}
		}
		else{
			if(bmrec.isBinary()){ // found "\\begin{binary}" -> binary mode.
				binaryMode = true;
				break;
			}
		}
		recpos = CONST_BM_NEXT;
	}

	// [4] termination.
	_boyer_moore_line = linenum + ubm->getLineNumber();
	ufile.close();

	// Binary mode reading.
	if(binaryMode){
		//  v             v
		// \\begin{binary}0DATA
		//  +---------------+
		recpos = bmpat.FoundPosition + bmpat.patlen + bmrec.getBufferPosition() + 1L;
		fileMgr->beginBinary(recpos);
		fileMgr->setBinaryStartRecord(rec_no+1); // binary records starts at the next record.
		lexer->set_step_counter(rec_no); // read until now.
		return readBinary();
	}
	return ePRS_NORMAL; // ePRS_ERR_FATAL;
}

// Array specification.
bool udfParserClass::setArrayNon()
{
	indexSpecStack.data().addType(INDEX_SIZE_NON);
	return true;
}
bool udfParserClass::setArrayNumber( long number )
{
	//if(!numberString) return false;
	//int number = converter.convertInteger(numberString);
	if(number < 0) number = 0;
	indexSpecStack.data().addNumber((unsigned int)number);
	return true;
}
bool udfParserClass::setArrayString( const char* nameString )
{
	// If an array is included, an error occurred.
	if(!nameString) return false;
	if(futl_count_all_array_bracket(nameString) > 0) return false;
	indexSpecStack.data().addString(nameString);
	return true;
}
// Used in parser.
char* udfParserClass::makeErrorMessage( const char* format, const char* content )
{
	if(format == NULL) return null_string;
	if(content == NULL){
		//snprintf(message_buff,(TXT_MAX_BUFF-1),format);
		snprintf(message_buff,(TXT_MAX_BUFF-1),"%s",format);
	}
	else{
		snprintf(message_buff,(TXT_MAX_BUFF-1),format,content);
	}
	message_buff[(TXT_MAX_BUFF-1)] = '\0';
	return message_buff;
}
